/// <reference types="webxr" />
interface Props {
}

declare const REVISION: string;

// https://developer.mozilla.org/en-US/docs/Web/API/MouseEvent.button
enum MOUSE {
    LEFT = 0,
    MIDDLE = 1,
    RIGHT = 2,
    ROTATE = 0,
    DOLLY = 1,
    PAN = 2,
}

enum TOUCH {
    ROTATE,
    PAN,
    DOLLY_PAN,
    DOLLY_ROTATE,
}

// GL STATE CONSTANTS
enum CullFace {}
declare const CullFaceNone: CullFace;
declare const CullFaceBack: CullFace;
declare const CullFaceFront: CullFace;
declare const CullFaceFrontBack: CullFace;

// Shadowing Type
enum ShadowMapType {}
declare const BasicShadowMap: ShadowMapType;
declare const PCFShadowMap: ShadowMapType;
declare const PCFSoftShadowMap: ShadowMapType;
declare const VSMShadowMap: ShadowMapType;

// MATERIAL CONSTANTS

// side
enum Side {}
declare const FrontSide: Side;
declare const BackSide: Side;
declare const DoubleSide: Side;

// shading
enum Shading {}
declare const FlatShading: Shading;
declare const SmoothShading: Shading;

// blending modes
enum Blending {}
declare const NoBlending: Blending;
declare const NormalBlending: Blending;
declare const AdditiveBlending: Blending;
declare const SubtractiveBlending: Blending;
declare const MultiplyBlending: Blending;
declare const CustomBlending: Blending;

// custom blending equations
// (numbers start from 100 not to clash with other
// mappings to OpenGL constants defined in Texture.js)
enum BlendingEquation {}
declare const AddEquation: BlendingEquation;
declare const SubtractEquation: BlendingEquation;
declare const ReverseSubtractEquation: BlendingEquation;
declare const MinEquation: BlendingEquation;
declare const MaxEquation: BlendingEquation;

// custom blending destination factors
enum BlendingDstFactor {}
declare const ZeroFactor: BlendingDstFactor;
declare const OneFactor: BlendingDstFactor;
declare const SrcColorFactor: BlendingDstFactor;
declare const OneMinusSrcColorFactor: BlendingDstFactor;
declare const SrcAlphaFactor: BlendingDstFactor;
declare const OneMinusSrcAlphaFactor: BlendingDstFactor;
declare const DstAlphaFactor: BlendingDstFactor;
declare const OneMinusDstAlphaFactor: BlendingDstFactor;
declare const DstColorFactor: BlendingDstFactor;
declare const OneMinusDstColorFactor: BlendingDstFactor;

// custom blending src factors
enum BlendingSrcFactor {}
declare const SrcAlphaSaturateFactor: BlendingSrcFactor;

// depth modes
enum DepthModes {}
declare const NeverDepth: DepthModes;
declare const AlwaysDepth: DepthModes;
declare const LessDepth: DepthModes;
declare const LessEqualDepth: DepthModes;
declare const EqualDepth: DepthModes;
declare const GreaterEqualDepth: DepthModes;
declare const GreaterDepth: DepthModes;
declare const NotEqualDepth: DepthModes;

// TEXTURE CONSTANTS
// Operations
enum Combine {}
declare const MultiplyOperation: Combine;
declare const MixOperation: Combine;
declare const AddOperation: Combine;

// Tone Mapping modes
enum ToneMapping {}
declare const NoToneMapping: ToneMapping;
declare const LinearToneMapping: ToneMapping;
declare const ReinhardToneMapping: ToneMapping;
declare const CineonToneMapping: ToneMapping;
declare const ACESFilmicToneMapping: ToneMapping;
declare const CustomToneMapping: ToneMapping;

// Mapping modes
enum Mapping {}
declare const UVMapping: Mapping;
declare const CubeReflectionMapping: Mapping;
declare const CubeRefractionMapping: Mapping;
declare const EquirectangularReflectionMapping: Mapping;
declare const EquirectangularRefractionMapping: Mapping;
declare const CubeUVReflectionMapping: Mapping;

// Wrapping modes
enum Wrapping {}
declare const RepeatWrapping: Wrapping;
declare const ClampToEdgeWrapping: Wrapping;
declare const MirroredRepeatWrapping: Wrapping;

// Filters
enum TextureFilter {}
declare const NearestFilter: TextureFilter;
declare const NearestMipmapNearestFilter: TextureFilter;
declare const NearestMipMapNearestFilter: TextureFilter;
declare const NearestMipmapLinearFilter: TextureFilter;
declare const NearestMipMapLinearFilter: TextureFilter;
declare const LinearFilter: TextureFilter;
declare const LinearMipmapNearestFilter: TextureFilter;
declare const LinearMipMapNearestFilter: TextureFilter;
declare const LinearMipmapLinearFilter: TextureFilter;
declare const LinearMipMapLinearFilter: TextureFilter;

// Data types
enum TextureDataType {}
declare const UnsignedByteType: TextureDataType;
declare const ByteType: TextureDataType;
declare const ShortType: TextureDataType;
declare const UnsignedShortType: TextureDataType;
declare const IntType: TextureDataType;
declare const UnsignedIntType: TextureDataType;
declare const FloatType: TextureDataType;
declare const HalfFloatType: TextureDataType;
declare const UnsignedShort4444Type: TextureDataType;
declare const UnsignedShort5551Type: TextureDataType;
declare const UnsignedInt248Type: TextureDataType;

// Pixel formats
enum PixelFormat {}
declare const AlphaFormat: PixelFormat;
declare const RGBFormat: PixelFormat;
declare const RGBAFormat: PixelFormat;
declare const LuminanceFormat: PixelFormat;
declare const LuminanceAlphaFormat: PixelFormat;
declare const DepthFormat: PixelFormat;
declare const DepthStencilFormat: PixelFormat;
declare const RedFormat: PixelFormat;
declare const RedIntegerFormat: PixelFormat;
declare const RGFormat: PixelFormat;
declare const RGIntegerFormat: PixelFormat;
declare const RGBAIntegerFormat: PixelFormat;
declare const _SRGBFormat: PixelFormat; // fallback for WebGL 1
declare const _SRGBAFormat: PixelFormat; // fallback for WebGL 1

// Internal Pixel Formats
type PixelFormatGPU =
    | 'ALPHA'
    | 'RGB'
    | 'RGBA'
    | 'LUMINANCE'
    | 'LUMINANCE_ALPHA'
    | 'RED_INTEGER'
    | 'R8'
    | 'R8_SNORM'
    | 'R8I'
    | 'R8UI'
    | 'R16I'
    | 'R16UI'
    | 'R16F'
    | 'R32I'
    | 'R32UI'
    | 'R32F'
    | 'RG8'
    | 'RG8_SNORM'
    | 'RG8I'
    | 'RG8UI'
    | 'RG16I'
    | 'RG16UI'
    | 'RG16F'
    | 'RG32I'
    | 'RG32UI'
    | 'RG32F'
    | 'RGB565'
    | 'RGB8'
    | 'RGB8_SNORM'
    | 'RGB8I'
    | 'RGB8UI'
    | 'RGB16I'
    | 'RGB16UI'
    | 'RGB16F'
    | 'RGB32I'
    | 'RGB32UI'
    | 'RGB32F'
    | 'RGB9_E5'
    | 'SRGB8'
    | 'R11F_G11F_B10F'
    | 'RGBA4'
    | 'RGBA8'
    | 'RGBA8_SNORM'
    | 'RGBA8I'
    | 'RGBA8UI'
    | 'RGBA16I'
    | 'RGBA16UI'
    | 'RGBA16F'
    | 'RGBA32I'
    | 'RGBA32UI'
    | 'RGBA32F'
    | 'RGB5_A1'
    | 'RGB10_A2'
    | 'RGB10_A2UI'
    | 'SRGB8_ALPHA8'
    | 'SRGB8'
    | 'DEPTH_COMPONENT16'
    | 'DEPTH_COMPONENT24'
    | 'DEPTH_COMPONENT32F'
    | 'DEPTH24_STENCIL8'
    | 'DEPTH32F_STENCIL8';

// Compressed texture formats
// DDS / ST3C Compressed texture formats
enum CompressedPixelFormat {}
declare const RGB_S3TC_DXT1_Format: CompressedPixelFormat;
declare const RGBA_S3TC_DXT1_Format: CompressedPixelFormat;
declare const RGBA_S3TC_DXT3_Format: CompressedPixelFormat;
declare const RGBA_S3TC_DXT5_Format: CompressedPixelFormat;

// PVRTC compressed './texture formats
declare const RGB_PVRTC_4BPPV1_Format: CompressedPixelFormat;
declare const RGB_PVRTC_2BPPV1_Format: CompressedPixelFormat;
declare const RGBA_PVRTC_4BPPV1_Format: CompressedPixelFormat;
declare const RGBA_PVRTC_2BPPV1_Format: CompressedPixelFormat;

// ETC compressed texture formats
declare const RGB_ETC1_Format: CompressedPixelFormat;
declare const RGB_ETC2_Format: CompressedPixelFormat;
declare const RGBA_ETC2_EAC_Format: CompressedPixelFormat;

// ASTC compressed texture formats
declare const RGBA_ASTC_4x4_Format: CompressedPixelFormat;
declare const RGBA_ASTC_5x4_Format: CompressedPixelFormat;
declare const RGBA_ASTC_5x5_Format: CompressedPixelFormat;
declare const RGBA_ASTC_6x5_Format: CompressedPixelFormat;
declare const RGBA_ASTC_6x6_Format: CompressedPixelFormat;
declare const RGBA_ASTC_8x5_Format: CompressedPixelFormat;
declare const RGBA_ASTC_8x6_Format: CompressedPixelFormat;
declare const RGBA_ASTC_8x8_Format: CompressedPixelFormat;
declare const RGBA_ASTC_10x5_Format: CompressedPixelFormat;
declare const RGBA_ASTC_10x6_Format: CompressedPixelFormat;
declare const RGBA_ASTC_10x8_Format: CompressedPixelFormat;
declare const RGBA_ASTC_10x10_Format: CompressedPixelFormat;
declare const RGBA_ASTC_12x10_Format: CompressedPixelFormat;
declare const RGBA_ASTC_12x12_Format: CompressedPixelFormat;

// BPTC compressed texture formats
declare const RGBA_BPTC_Format: CompressedPixelFormat;

// Loop styles for AnimationAction
enum AnimationActionLoopStyles {}
declare const LoopOnce: AnimationActionLoopStyles;
declare const LoopRepeat: AnimationActionLoopStyles;
declare const LoopPingPong: AnimationActionLoopStyles;

// Interpolation
enum InterpolationModes {}
declare const InterpolateDiscrete: InterpolationModes;
declare const InterpolateLinear: InterpolationModes;
declare const InterpolateSmooth: InterpolationModes;

// Interpolant ending modes
enum InterpolationEndingModes {}
declare const ZeroCurvatureEnding: InterpolationEndingModes;
declare const ZeroSlopeEnding: InterpolationEndingModes;
declare const WrapAroundEnding: InterpolationEndingModes;

// Animation blending modes
enum AnimationBlendMode {}
declare const NormalAnimationBlendMode: AnimationBlendMode;
declare const AdditiveAnimationBlendMode: AnimationBlendMode;

// Triangle Draw modes
enum TrianglesDrawModes {}
declare const TrianglesDrawMode: TrianglesDrawModes;
declare const TriangleStripDrawMode: TrianglesDrawModes;
declare const TriangleFanDrawMode: TrianglesDrawModes;

// Texture Encodings
enum TextureEncoding {}
declare const LinearEncoding: TextureEncoding;
declare const sRGBEncoding: TextureEncoding;

// Depth packing strategies
enum DepthPackingStrategies {}
declare const BasicDepthPacking: DepthPackingStrategies;
declare const RGBADepthPacking: DepthPackingStrategies;

// Normal Map types
enum NormalMapTypes {}
declare const TangentSpaceNormalMap: NormalMapTypes;
declare const ObjectSpaceNormalMap: NormalMapTypes;

type ColorSpace = NoColorSpace | SRGBColorSpace | LinearSRGBColorSpace;
type NoColorSpace = '';
type SRGBColorSpace = 'srgb';
type LinearSRGBColorSpace = 'srgb-linear';

// Stencil Op types
enum StencilOp {}
declare const ZeroStencilOp: StencilOp;
declare const KeepStencilOp: StencilOp;
declare const ReplaceStencilOp: StencilOp;
declare const IncrementStencilOp: StencilOp;
declare const DecrementStencilOp: StencilOp;
declare const IncrementWrapStencilOp: StencilOp;
declare const DecrementWrapStencilOp: StencilOp;
declare const InvertStencilOp: StencilOp;

// Stencil Func types
enum StencilFunc {}
declare const NeverStencilFunc: StencilFunc;
declare const LessStencilFunc: StencilFunc;
declare const EqualStencilFunc: StencilFunc;
declare const LessEqualStencilFunc: StencilFunc;
declare const GreaterStencilFunc: StencilFunc;
declare const NotEqualStencilFunc: StencilFunc;
declare const GreaterEqualStencilFunc: StencilFunc;
declare const AlwaysStencilFunc: StencilFunc;

// usage types
enum Usage {}
declare const StaticDrawUsage: Usage;
declare const DynamicDrawUsage: Usage;
declare const StreamDrawUsage: Usage;
declare const StaticReadUsage: Usage;
declare const DynamicReadUsage: Usage;
declare const StreamReadUsage: Usage;
declare const StaticCopyUsage: Usage;
declare const DynamicCopyUsage: Usage;
declare const StreamCopyUsage: Usage;

enum GLSLVersion {}
declare const GLSL1: GLSLVersion;
declare const GLSL3: GLSLVersion;

type BuiltinShaderAttributeName =
    | 'position'
    | 'normal'
    | 'uv'
    | 'color'
    | 'skinIndex'
    | 'skinWeight'
    | 'instanceMatrix'
    | 'morphTarget0'
    | 'morphTarget1'
    | 'morphTarget2'
    | 'morphTarget3'
    | 'morphTarget4'
    | 'morphTarget5'
    | 'morphTarget6'
    | 'morphTarget7'
    | 'morphNormal0'
    | 'morphNormal1'
    | 'morphNormal2'
    | 'morphNormal3';

/**
 * Implementation of a quaternion. This is used for rotating things without incurring in the dreaded gimbal lock issue, amongst other advantages.
 *
 * @example
 * const quaternion = new THREE.Quaternion();
 * quaternion.setFromAxisAngle( new THREE.Vector3( 0, 1, 0 ), Math.PI / 2 );
 * const vector = new THREE.Vector3( 1, 0, 0 );
 * vector.applyQuaternion( quaternion );
 */
declare class Quaternion {
    /**
     * @param x x coordinate
     * @param y y coordinate
     * @param z z coordinate
     * @param w w coordinate
     */
    constructor(x?: number, y?: number, z?: number, w?: number);

    /**
     * @default 0
     */
    x: number;

    /**
     * @default 0
     */
    y: number;

    /**
     * @default 0
     */
    z: number;

    /**
     * @default 1
     */
    w: number;
    readonly isQuaternion: true;

    /**
     * Sets values of this quaternion.
     */
    set(x: number, y: number, z: number, w: number): Quaternion;

    /**
     * Clones this quaternion.
     */
    clone(): this;

    /**
     * Copies values of q to this quaternion.
     */
    copy(q: Quaternion): this;

    /**
     * Sets this quaternion from rotation specified by Euler angles.
     */
    setFromEuler(euler: Euler, update?: boolean): Quaternion;

    /**
     * Sets this quaternion from rotation specified by axis and angle.
     * Adapted from http://www.euclideanspace.com/maths/geometry/rotations/conversions/angleToQuaternion/index.htm.
     * Axis have to be normalized, angle is in radians.
     */
    setFromAxisAngle(axis: Vector3, angle: number): Quaternion;

    /**
     * Sets this quaternion from rotation component of m. Adapted from http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/index.htm.
     */
    setFromRotationMatrix(m: Matrix4): Quaternion;
    setFromUnitVectors(vFrom: Vector3, vTo: Vector3): Quaternion;
    angleTo(q: Quaternion): number;
    rotateTowards(q: Quaternion, step: number): Quaternion;

    identity(): Quaternion;

    /**
     * Inverts this quaternion.
     */
    invert(): Quaternion;

    conjugate(): Quaternion;
    dot(v: Quaternion): number;
    lengthSq(): number;

    /**
     * Computes length of this quaternion.
     */
    length(): number;

    /**
     * Normalizes this quaternion.
     */
    normalize(): Quaternion;

    /**
     * Multiplies this quaternion by b.
     */
    multiply(q: Quaternion): Quaternion;
    premultiply(q: Quaternion): Quaternion;

    /**
     * Sets this quaternion to a x b
     * Adapted from http://www.euclideanspace.com/maths/algebra/realNormedAlgebra/quaternions/code/index.htm.
     */
    multiplyQuaternions(a: Quaternion, b: Quaternion): Quaternion;

    slerp(qb: Quaternion, t: number): Quaternion;
    slerpQuaternions(qa: Quaternion, qb: Quaternion, t: number): Quaternion;
    equals(v: Quaternion): boolean;

    /**
     * Sets this quaternion's x, y, z and w value from the provided array or array-like.
     * @param array the source array or array-like.
     * @param offset (optional) offset into the array. Default is 0.
     */
    fromArray(array: number[] | ArrayLike<number>, offset?: number): this;

    /**
     * Returns an array [x, y, z, w], or copies x, y, z and w into the provided array.
     * @param array (optional) array to store the quaternion to. If this is not provided, a new array will be created.
     * @param offset (optional) optional offset into the array.
     * @return The created or provided array.
     */
    toArray(array?: number[], offset?: number): number[];

    /**
     * Copies x, y, z and w into the provided array-like.
     * @param array array-like to store the quaternion to.
     * @param offset (optional) optional offset into the array.
     * @return The provided array-like.
     */
    toArray(array: ArrayLike<number>, offset?: number): ArrayLike<number>;

    _onChange(callback: () => void): Quaternion;
    _onChangeCallback: () => void;

    static slerpFlat(
        dst: number[],
        dstOffset: number,
        src0: number[],
        srcOffset: number,
        src1: number[],
        stcOffset1: number,
        t: number,
    ): Quaternion;

    static multiplyQuaternionsFlat(
        dst: number[],
        dstOffset: number,
        src0: number[],
        srcOffset: number,
        src1: number[],
        stcOffset1: number,
    ): number[];

    /**
     * @deprecated Use qm.slerpQuaternions( qa, qb, t ) instead..
     */
    static slerp(qa: Quaternion, qb: Quaternion, qm: Quaternion, t: number): number;

    /**
     * @deprecated Use {@link Vector#applyQuaternion vector.applyQuaternion( quaternion )} instead.
     */
    multiplyVector3(v: any): any;

    /**
     * @deprecated Use {@link Quaternion#invert .invert()} instead.
     */
    inverse(): Quaternion;

    random(): Quaternion;

    [Symbol.iterator](): Generator<number, void>;
}

declare class Euler {
    constructor(x?: number, y?: number, z?: number, order?: string);

    /**
     * @default 0
     */
    x: number;

    /**
     * @default 0
     */
    y: number;

    /**
     * @default 0
     */
    z: number;

    /**
     * @default THREE.Euler.DefaultOrder
     */
    order: string;
    readonly isEuler: true;

    _onChangeCallback: () => void;

    set(x: number, y: number, z: number, order?: string): Euler;
    clone(): this;
    copy(euler: Euler): this;
    setFromRotationMatrix(m: Matrix4, order?: string, update?: boolean): Euler;
    setFromQuaternion(q: Quaternion, order?: string, update?: boolean): Euler;
    setFromVector3(v: Vector3, order?: string): Euler;
    reorder(newOrder: string): Euler;
    equals(euler: Euler): boolean;
    fromArray(xyzo: any[]): Euler;
    toArray(array?: number[], offset?: number): number[];
    _onChange(callback: () => void): this;

    static RotationOrders: string[];
    static DefaultOrder: string;

    [Symbol.iterator](): Generator<string | number, void>;
}

declare class Layers {
    constructor();

    /**
     * @default 1 | 0
     */
    mask: number;

    set(channel: number): void;
    enable(channel: number): void;
    enableAll(): void;
    toggle(channel: number): void;
    disable(channel: number): void;
    disableAll(): void;
    test(layers: Layers): boolean;
    isEnabled(channel: number): boolean;
}

interface FogBase {
    name: string;
    color: Color;
    clone(): FogBase;
    toJSON(): any;
}

/**
 * This class contains the parameters that define linear fog, i.e., that grows linearly denser with the distance.
 */
declare class Fog implements FogBase {
    constructor(color: ColorRepresentation, near?: number, far?: number);

    /**
     * @default ''
     */
    name: string;

    /**
     * Fog color.
     */
    color: Color;

    /**
     * The minimum distance to start applying fog. Objects that are less than 'near' units from the active camera won't be affected by fog.
     * @default 1
     */
    near: number;

    /**
     * The maximum distance at which fog stops being calculated and applied. Objects that are more than 'far' units away from the active camera won't be affected by fog.
     * @default 1000
     */
    far: number;

    readonly isFog: true;

    clone(): Fog;
    toJSON(): any;
}

type Vector2Tuple = [number, number];

/**
 * ( interface Vector<T> )
 *
 * Abstract interface of {@link https://github.com/mrdoob/three.js/blob/master/src/math/Vector2.js|Vector2},
 * {@link https://github.com/mrdoob/three.js/blob/master/src/math/Vector3.js|Vector3}
 * and {@link https://github.com/mrdoob/three.js/blob/master/src/math/Vector4.js|Vector4}.
 *
 * Currently the members of Vector is NOT type safe because it accepts different typed vectors.
 *
 * Those definitions will be changed when TypeScript innovates Generics to be type safe.
 *
 * @example
 * const v:THREE.Vector = new THREE.Vector3();
 * v.addVectors(new THREE.Vector2(0, 1), new THREE.Vector2(2, 3)); // invalid but compiled successfully
 */
interface Vector {
    setComponent(index: number, value: number): this;

    getComponent(index: number): number;

    set(...args: number[]): this;

    setScalar(scalar: number): this;

    /**
     * copy(v:T):T;
     */
    copy(v: Vector): this;

    /**
     * NOTE: The second argument is deprecated.
     *
     * add(v:T):T;
     */
    add(v: Vector): this;

    /**
     * addVectors(a:T, b:T):T;
     */
    addVectors(a: Vector, b: Vector): this;

    addScaledVector(vector: Vector, scale: number): this;

    /**
     * Adds the scalar value s to this vector's values.
     */
    addScalar(scalar: number): this;

    /**
     * sub(v:T):T;
     */
    sub(v: Vector): this;

    /**
     * subVectors(a:T, b:T):T;
     */
    subVectors(a: Vector, b: Vector): this;

    /**
     * multiplyScalar(s:number):T;
     */
    multiplyScalar(s: number): this;

    /**
     * divideScalar(s:number):T;
     */
    divideScalar(s: number): this;

    /**
     * negate():T;
     */
    negate(): this;

    /**
     * dot(v:T):T;
     */
    dot(v: Vector): number;

    /**
     * lengthSq():number;
     */
    lengthSq(): number;

    /**
     * length():number;
     */
    length(): number;

    /**
     * normalize():T;
     */
    normalize(): this;

    /**
     * NOTE: Vector4 doesn't have the property.
     *
     * distanceTo(v:T):number;
     */
    distanceTo?(v: Vector): number;

    /**
     * NOTE: Vector4 doesn't have the property.
     *
     * distanceToSquared(v:T):number;
     */
    distanceToSquared?(v: Vector): number;

    /**
     * setLength(l:number):T;
     */
    setLength(l: number): this;

    /**
     * lerp(v:T, alpha:number):T;
     */
    lerp(v: Vector, alpha: number): this;

    /**
     * equals(v:T):boolean;
     */
    equals(v: Vector): boolean;

    /**
     * clone():T;
     */
    clone(): Vector;
}

/**
 * 2D vector.
 *
 * ( class Vector2 implements Vector<Vector2> )
 */
declare class Vector2 implements Vector {
    constructor(x?: number, y?: number);

    /**
     * @default 0
     */
    x: number;

    /**
     * @default 0
     */
    y: number;
    width: number;
    height: number;
    readonly isVector2: true;

    /**
     * Sets value of this vector.
     */
    set(x: number, y: number): this;

    /**
     * Sets the x and y values of this vector both equal to scalar.
     */
    setScalar(scalar: number): this;

    /**
     * Sets X component of this vector.
     */
    setX(x: number): this;

    /**
     * Sets Y component of this vector.
     */
    setY(y: number): this;

    /**
     * Sets a component of this vector.
     */
    setComponent(index: number, value: number): this;

    /**
     * Gets a component of this vector.
     */
    getComponent(index: number): number;

    /**
     * Returns a new Vector2 instance with the same `x` and `y` values.
     */
    clone(): this;

    /**
     * Copies value of v to this vector.
     */
    copy(v: Vector2): this;

    /**
     * Adds v to this vector.
     */
    add(v: Vector2, w?: Vector2): this;

    /**
     * Adds the scalar value s to this vector's x and y values.
     */
    addScalar(s: number): this;

    /**
     * Sets this vector to a + b.
     */
    addVectors(a: Vector2, b: Vector2): this;

    /**
     * Adds the multiple of v and s to this vector.
     */
    addScaledVector(v: Vector2, s: number): this;

    /**
     * Subtracts v from this vector.
     */
    sub(v: Vector2): this;

    /**
     * Subtracts s from this vector's x and y components.
     */
    subScalar(s: number): this;

    /**
     * Sets this vector to a - b.
     */
    subVectors(a: Vector2, b: Vector2): this;

    /**
     * Multiplies this vector by v.
     */
    multiply(v: Vector2): this;

    /**
     * Multiplies this vector by scalar s.
     */
    multiplyScalar(scalar: number): this;

    /**
     * Divides this vector by v.
     */
    divide(v: Vector2): this;

    /**
     * Divides this vector by scalar s.
     * Set vector to ( 0, 0 ) if s == 0.
     */
    divideScalar(s: number): this;

    /**
     * Multiplies this vector (with an implicit 1 as the 3rd component) by m.
     */
    applyMatrix3(m: Matrix3): this;

    /**
     * If this vector's x or y value is greater than v's x or y value, replace that value with the corresponding min value.
     */
    min(v: Vector2): this;

    /**
     * If this vector's x or y value is less than v's x or y value, replace that value with the corresponding max value.
     */
    max(v: Vector2): this;

    /**
     * If this vector's x or y value is greater than the max vector's x or y value, it is replaced by the corresponding value.
     * If this vector's x or y value is less than the min vector's x or y value, it is replaced by the corresponding value.
     * @param min the minimum x and y values.
     * @param max the maximum x and y values in the desired range.
     */
    clamp(min: Vector2, max: Vector2): this;

    /**
     * If this vector's x or y values are greater than the max value, they are replaced by the max value.
     * If this vector's x or y values are less than the min value, they are replaced by the min value.
     * @param min the minimum value the components will be clamped to.
     * @param max the maximum value the components will be clamped to.
     */
    clampScalar(min: number, max: number): this;

    /**
     * If this vector's length is greater than the max value, it is replaced by the max value.
     * If this vector's length is less than the min value, it is replaced by the min value.
     * @param min the minimum value the length will be clamped to.
     * @param max the maximum value the length will be clamped to.
     */
    clampLength(min: number, max: number): this;

    /**
     * The components of the vector are rounded down to the nearest integer value.
     */
    floor(): this;

    /**
     * The x and y components of the vector are rounded up to the nearest integer value.
     */
    ceil(): this;

    /**
     * The components of the vector are rounded to the nearest integer value.
     */
    round(): this;

    /**
     * The components of the vector are rounded towards zero (up if negative, down if positive) to an integer value.
     */
    roundToZero(): this;

    /**
     * Inverts this vector.
     */
    negate(): this;

    /**
     * Computes dot product of this vector and v.
     */
    dot(v: Vector2): number;

    /**
     * Computes cross product of this vector and v.
     */
    cross(v: Vector2): number;

    /**
     * Computes squared length of this vector.
     */
    lengthSq(): number;

    /**
     * Computes length of this vector.
     */
    length(): number;

    /**
     * @deprecated Use {@link Vector2#manhattanLength .manhattanLength()} instead.
     */
    lengthManhattan(): number;

    /**
     * Computes the Manhattan length of this vector.
     *
     * see {@link http://en.wikipedia.org/wiki/Taxicab_geometry|Wikipedia: Taxicab Geometry}
     */
    manhattanLength(): number;

    /**
     * Normalizes this vector.
     */
    normalize(): this;

    /**
     * computes the angle in radians with respect to the positive x-axis
     */
    angle(): number;

    /**
     * Computes distance of this vector to v.
     */
    distanceTo(v: Vector2): number;

    /**
     * Computes squared distance of this vector to v.
     */
    distanceToSquared(v: Vector2): number;

    /**
     * @deprecated Use {@link Vector2#manhattanDistanceTo .manhattanDistanceTo()} instead.
     */
    distanceToManhattan(v: Vector2): number;

    /**
     * Computes the Manhattan length (distance) from this vector to the given vector v
     *
     * see {@link http://en.wikipedia.org/wiki/Taxicab_geometry|Wikipedia: Taxicab Geometry}
     */
    manhattanDistanceTo(v: Vector2): number;

    /**
     * Normalizes this vector and multiplies it by l.
     */
    setLength(length: number): this;

    /**
     * Linearly interpolates between this vector and v, where alpha is the distance along the line - alpha = 0 will be this vector, and alpha = 1 will be v.
     * @param v vector to interpolate towards.
     * @param alpha interpolation factor in the closed interval [0, 1].
     */
    lerp(v: Vector2, alpha: number): this;

    /**
     * Sets this vector to be the vector linearly interpolated between v1 and v2 where alpha is the distance along the line connecting the two vectors - alpha = 0 will be v1, and alpha = 1 will be v2.
     * @param v1 the starting vector.
     * @param v2 vector to interpolate towards.
     * @param alpha interpolation factor in the closed interval [0, 1].
     */
    lerpVectors(v1: Vector2, v2: Vector2, alpha: number): this;

    /**
     * Checks for strict equality of this vector and v.
     */
    equals(v: Vector2): boolean;

    /**
     * Sets this vector's x and y value from the provided array or array-like.
     * @param array the source array or array-like.
     * @param offset (optional) offset into the array. Default is 0.
     */
    fromArray(array: number[] | ArrayLike<number>, offset?: number): this;

    /**
     * Returns an array [x, y], or copies x and y into the provided array.
     * @param array (optional) array to store the vector to. If this is not provided, a new array will be created.
     * @param offset (optional) optional offset into the array.
     * @return The created or provided array.
     */
    toArray(array?: number[], offset?: number): number[];
    toArray(array?: Vector2Tuple, offset?: 0): Vector2Tuple;

    /**
     * Copies x and y into the provided array-like.
     * @param array array-like to store the vector to.
     * @param offset (optional) optional offset into the array.
     * @return The provided array-like.
     */
    toArray(array: ArrayLike<number>, offset?: number): ArrayLike<number>;

    /**
     * Sets this vector's x and y values from the attribute.
     * @param attribute the source attribute.
     * @param index index in the attribute.
     */
    fromBufferAttribute(attribute: BufferAttribute, index: number): this;

    /**
     * Rotates the vector around center by angle radians.
     * @param center the point around which to rotate.
     * @param angle the angle to rotate, in radians.
     */
    rotateAround(center: Vector2, angle: number): this;

    /**
     * Sets this vector's x and y from Math.random
     */
    random(): this;
}

/**
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/core/InterleavedBuffer.js|src/core/InterleavedBuffer.js}
 */
declare class InterleavedBuffer {
    constructor(array: ArrayLike<number>, stride: number);

    array: ArrayLike<number>;
    stride: number;

    /**
     * @default THREE.StaticDrawUsage
     */
    usage: Usage;

    /**
     * @default { offset: number; count: number }
     */
    updateRange: { offset: number; count: number };

    /**
     * @default 0
     */
    version: number;

    length: number;

    /**
     * @default 0
     */
    count: number;
    needsUpdate: boolean;
    uuid: string;

    setUsage(usage: Usage): InterleavedBuffer;
    clone(data: object): InterleavedBuffer;
    copy(source: InterleavedBuffer): this;
    copyAt(index1: number, attribute: InterleavedBufferAttribute, index2: number): InterleavedBuffer;
    set(value: ArrayLike<number>, index: number): InterleavedBuffer;
    toJSON(data: object): {
        uuid: string;
        buffer: string;
        type: string;
        stride: number;
    };
}

/**
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/core/InterleavedBufferAttribute.js|src/core/InterleavedBufferAttribute.js}
 */
declare class InterleavedBufferAttribute {
    constructor(interleavedBuffer: InterleavedBuffer, itemSize: number, offset: number, normalized?: boolean);

    /**
     * @default ''
     */
    name: string;
    data: InterleavedBuffer;
    itemSize: number;
    offset: number;

    /**
     * @default false
     */
    normalized: boolean;

    get count(): number;
    get array(): ArrayLike<number>;
    set needsUpdate(value: boolean);

    readonly isInterleavedBufferAttribute: true;

    applyMatrix4(m: Matrix4): this;
    clone(data?: object): BufferAttribute;
    getX(index: number): number;
    setX(index: number, x: number): this;
    getY(index: number): number;
    setY(index: number, y: number): this;
    getZ(index: number): number;
    setZ(index: number, z: number): this;
    getW(index: number): number;
    setW(index: number, z: number): this;
    setXY(index: number, x: number, y: number): this;
    setXYZ(index: number, x: number, y: number, z: number): this;
    setXYZW(index: number, x: number, y: number, z: number, w: number): this;
    toJSON(data?: object): {
        isInterleavedBufferAttribute: true;
        itemSize: number;
        data: string;
        offset: number;
        normalized: boolean;
    };
    applyNormalMatrix(matrix: Matrix): this;
    transformDirection(matrix: Matrix): this;
}

declare class Triangle {
    constructor(a?: Vector3, b?: Vector3, c?: Vector3);

    /**
     * @default new THREE.Vector3()
     */
    a: Vector3;

    /**
     * @default new THREE.Vector3()
     */
    b: Vector3;

    /**
     * @default new THREE.Vector3()
     */
    c: Vector3;

    set(a: Vector3, b: Vector3, c: Vector3): Triangle;
    setFromPointsAndIndices(points: Vector3[], i0: number, i1: number, i2: number): this;
    setFromAttributeAndIndices(
        attribute: BufferAttribute | InterleavedBufferAttribute,
        i0: number,
        i1: number,
        i2: number,
    ): this;
    clone(): this;
    copy(triangle: Triangle): this;
    getArea(): number;
    getMidpoint(target: Vector3): Vector3;
    getNormal(target: Vector3): Vector3;
    getPlane(target: Plane): Plane;
    getBarycoord(point: Vector3, target: Vector3): Vector3;
    getUV(point: Vector3, uv1: Vector2, uv2: Vector2, uv3: Vector2, target: Vector2): Vector2;
    containsPoint(point: Vector3): boolean;
    intersectsBox(box: Box3): boolean;
    isFrontFacing(direction: Vector3): boolean;
    closestPointToPoint(point: Vector3, target: Vector3): Vector3;
    equals(triangle: Triangle): boolean;

    static getNormal(a: Vector3, b: Vector3, c: Vector3, target: Vector3): Vector3;
    static getBarycoord(point: Vector3, a: Vector3, b: Vector3, c: Vector3, target: Vector3): Vector3;
    static containsPoint(point: Vector3, a: Vector3, b: Vector3, c: Vector3): boolean;
    static getUV(
        point: Vector3,
        p1: Vector3,
        p2: Vector3,
        p3: Vector3,
        uv1: Vector2,
        uv2: Vector2,
        uv3: Vector2,
        target: Vector2,
    ): Vector2;
    static isFrontFacing(a: Vector3, b: Vector3, c: Vector3, direction: Vector3): boolean;
}

declare class Box3 {
    constructor(min?: Vector3, max?: Vector3);

    /**
     * @default new THREE.Vector3( + Infinity, + Infinity, + Infinity )
     */
    min: Vector3;

    /**
     * @default new THREE.Vector3( - Infinity, - Infinity, - Infinity )
     */
    max: Vector3;
    readonly isBox3: true;

    set(min: Vector3, max: Vector3): this;
    setFromArray(array: ArrayLike<number>): this;
    setFromBufferAttribute(bufferAttribute: BufferAttribute): this;
    setFromPoints(points: Vector3[]): this;
    setFromCenterAndSize(center: Vector3, size: Vector3): this;
    setFromObject(object: Object3D, precise?: boolean): this;
    clone(): this;
    copy(box: Box3): this;
    makeEmpty(): this;
    isEmpty(): boolean;
    getCenter(target: Vector3): Vector3;
    getSize(target: Vector3): Vector3;
    expandByPoint(point: Vector3): this;
    expandByVector(vector: Vector3): this;
    expandByScalar(scalar: number): this;
    expandByObject(object: Object3D, precise?: boolean): this;
    containsPoint(point: Vector3): boolean;
    containsBox(box: Box3): boolean;
    getParameter(point: Vector3, target: Vector3): Vector3;
    intersectsBox(box: Box3): boolean;
    intersectsSphere(sphere: Sphere): boolean;
    intersectsPlane(plane: Plane): boolean;
    intersectsTriangle(triangle: Triangle): boolean;
    clampPoint(point: Vector3, target: Vector3): Vector3;
    distanceToPoint(point: Vector3): number;
    getBoundingSphere(target: Sphere): Sphere;
    intersect(box: Box3): this;
    union(box: Box3): this;
    applyMatrix4(matrix: Matrix4): this;
    translate(offset: Vector3): this;
    equals(box: Box3): boolean;
    /**
     * @deprecated Use {@link Box3#isEmpty .isEmpty()} instead.
     */
    empty(): any;
    /**
     * @deprecated Use {@link Box3#intersectsBox .intersectsBox()} instead.
     */
    isIntersectionBox(b: any): any;
    /**
     * @deprecated Use {@link Box3#intersectsSphere .intersectsSphere()} instead.
     */
    isIntersectionSphere(s: any): any;
}

declare class Sphere {
    constructor(center?: Vector3, radius?: number);

    /**
     * @default new Vector3()
     */
    center: Vector3;

    /**
     * @default 1
     */
    radius: number;

    set(center: Vector3, radius: number): Sphere;
    setFromPoints(points: Vector3[], optionalCenter?: Vector3): Sphere;
    clone(): this;
    copy(sphere: Sphere): this;
    expandByPoint(point: Vector3): this;
    isEmpty(): boolean;
    makeEmpty(): this;
    containsPoint(point: Vector3): boolean;
    distanceToPoint(point: Vector3): number;
    intersectsSphere(sphere: Sphere): boolean;
    intersectsBox(box: Box3): boolean;
    intersectsPlane(plane: Plane): boolean;
    clampPoint(point: Vector3, target: Vector3): Vector3;
    getBoundingBox(target: Box3): Box3;
    applyMatrix4(matrix: Matrix4): Sphere;
    translate(offset: Vector3): Sphere;
    equals(sphere: Sphere): boolean;
    union(sphere: Sphere): this;

    /**
     * @deprecated Use {@link Sphere#isEmpty .isEmpty()} instead.
     */
    empty(): any;
}

declare class Line3 {
    constructor(start?: Vector3, end?: Vector3);

    /**
     * @default new THREE.Vector3()
     */
    start: Vector3;

    /**
     * @default new THREE.Vector3()
     */
    end: Vector3;

    set(start?: Vector3, end?: Vector3): Line3;
    clone(): this;
    copy(line: Line3): this;
    getCenter(target: Vector3): Vector3;
    delta(target: Vector3): Vector3;
    distanceSq(): number;
    distance(): number;
    at(t: number, target: Vector3): Vector3;
    closestPointToPointParameter(point: Vector3, clampToLine?: boolean): number;
    closestPointToPoint(point: Vector3, clampToLine: boolean, target: Vector3): Vector3;
    applyMatrix4(matrix: Matrix4): Line3;
    equals(line: Line3): boolean;
}

declare class Plane {
    constructor(normal?: Vector3, constant?: number);

    /**
     * @default new THREE.Vector3( 1, 0, 0 )
     */
    normal: Vector3;

    /**
     * @default 0
     */
    constant: number;

    readonly isPlane: true;

    set(normal: Vector3, constant: number): Plane;
    setComponents(x: number, y: number, z: number, w: number): Plane;
    setFromNormalAndCoplanarPoint(normal: Vector3, point: Vector3): Plane;
    setFromCoplanarPoints(a: Vector3, b: Vector3, c: Vector3): Plane;
    clone(): this;
    copy(plane: Plane): this;
    normalize(): Plane;
    negate(): Plane;
    distanceToPoint(point: Vector3): number;
    distanceToSphere(sphere: Sphere): number;
    projectPoint(point: Vector3, target: Vector3): Vector3;
    intersectLine(line: Line3, target: Vector3): Vector3 | null;
    intersectsLine(line: Line3): boolean;
    intersectsBox(box: Box3): boolean;
    intersectsSphere(sphere: Sphere): boolean;
    coplanarPoint(target: Vector3): Vector3;
    applyMatrix4(matrix: Matrix4, optionalNormalMatrix?: Matrix3): Plane;
    translate(offset: Vector3): Plane;
    equals(plane: Plane): boolean;

    /**
     * @deprecated Use {@link Plane#intersectsLine .intersectsLine()} instead.
     */
    isIntersectionLine(l: any): any;
}

interface BaseEvent {
    type: string;
}

/**
 * Event object.
 */
interface Event$1 extends BaseEvent {
    target?: any;
    [attachment: string]: any;
}
type EventListener<E, T, U> = (event: E & { type: T } & { target: U }) => void;

/**
 * JavaScript events for custom objects
 *
 * @source src/core/EventDispatcher.js
 */
declare class EventDispatcher<E extends BaseEvent = Event$1> {
    /**
     * Creates eventDispatcher object. It needs to be call with '.call' to add the functionality to an object.
     */
    constructor();

    /**
     * Adds a listener to an event type.
     * @param type The type of event to listen to.
     * @param listener The function that gets called when the event is fired.
     */
    addEventListener<T extends E['type']>(type: T, listener: EventListener<E, T, this>): void;

    /**
     * Checks if listener is added to an event type.
     * @param type The type of event to listen to.
     * @param listener The function that gets called when the event is fired.
     */
    hasEventListener<T extends E['type']>(type: T, listener: EventListener<E, T, this>): boolean;

    /**
     * Removes a listener from an event type.
     * @param type The type of the listener that gets removed.
     * @param listener The listener function that gets removed.
     */
    removeEventListener<T extends E['type']>(type: T, listener: EventListener<E, T, this>): void;

    /**
     * Fire an event type.
     * @param type The type of event that gets fired.
     */
    dispatchEvent(event: E): void;
}

// tslint:disable-next-line:interface-name
interface IUniform<TValue = any> {
    value: TValue;
}

declare let UniformsLib: {
    common: {
        diffuse: IUniform;
        opacity: IUniform;
        map: IUniform;
        uvTransform: IUniform;
        uv2Transform: IUniform;
        alphaMap: IUniform;
    };
    specularmap: {
        specularMap: IUniform;
    };
    envmap: {
        envMap: IUniform;
        flipEnvMap: IUniform;
        reflectivity: IUniform;
        refractionRatio: IUniform;
        maxMipLevel: IUniform;
    };
    aomap: {
        aoMap: IUniform;
        aoMapIntensity: IUniform;
    };
    lightmap: {
        lightMap: IUniform;
        lightMapIntensity: IUniform;
    };
    emissivemap: {
        emissiveMap: IUniform;
    };
    bumpmap: {
        bumpMap: IUniform;
        bumpScale: IUniform;
    };
    normalmap: {
        normalMap: IUniform;
        normalScale: IUniform;
    };
    displacementmap: {
        displacementMap: IUniform;
        displacementScale: IUniform;
        displacementBias: IUniform;
    };
    roughnessmap: {
        roughnessMap: IUniform;
    };
    metalnessmap: {
        metalnessMap: IUniform;
    };
    gradientmap: {
        gradientMap: IUniform;
    };
    fog: {
        fogDensity: IUniform;
        fogNear: IUniform;
        fogFar: IUniform;
        fogColor: IUniform;
    };
    lights: {
        ambientLightColor: IUniform;
        directionalLights: {
            value: any[];
            properties: {
                direction: {};
                color: {};
            };
        };
        directionalLightShadows: {
            value: any[];
            properties: {
                shadowBias: {};
                shadowNormalBias: {};
                shadowRadius: {};
                shadowMapSize: {};
            };
        };
        directionalShadowMap: IUniform;
        directionalShadowMatrix: IUniform;
        spotLights: {
            value: any[];
            properties: {
                color: {};
                position: {};
                direction: {};
                distance: {};
                coneCos: {};
                penumbraCos: {};
                decay: {};
            };
        };
        spotLightShadows: {
            value: any[];
            properties: {
                shadowBias: {};
                shadowNormalBias: {};
                shadowRadius: {};
                shadowMapSize: {};
            };
        };
        spotShadowMap: IUniform;
        spotShadowMatrix: IUniform;
        pointLights: {
            value: any[];
            properties: {
                color: {};
                position: {};
                decay: {};
                distance: {};
            };
        };
        pointLightShadows: {
            value: any[];
            properties: {
                shadowBias: {};
                shadowNormalBias: {};
                shadowRadius: {};
                shadowMapSize: {};
            };
        };
        pointShadowMap: IUniform;
        pointShadowMatrix: IUniform;
        hemisphereLights: {
            value: any[];
            properties: {
                direction: {};
                skycolor: {};
                groundColor: {};
            };
        };
        rectAreaLights: {
            value: any[];
            properties: {
                color: {};
                position: {};
                width: {};
                height: {};
            };
        };
    };
    points: {
        diffuse: IUniform;
        opacity: IUniform;
        size: IUniform;
        scale: IUniform;
        map: IUniform;
        uvTransform: IUniform;
    };
};

interface Shader {
    uniforms: { [uniform: string]: IUniform };
    vertexShader: string;
    fragmentShader: string;
}

declare let ShaderLib: {
    [name: string]: Shader;
    basic: Shader;
    lambert: Shader;
    phong: Shader;
    standard: Shader;
    matcap: Shader;
    points: Shader;
    dashed: Shader;
    depth: Shader;
    normal: Shader;
    sprite: Shader;
    background: Shader;
    cube: Shader;
    equirect: Shader;
    distanceRGBA: Shader;
    shadow: Shader;
    physical: Shader;
};

interface MaterialParameters {
    alphaTest?: number | undefined;
    alphaToCoverage?: boolean | undefined;
    blendDst?: BlendingDstFactor | undefined;
    blendDstAlpha?: number | undefined;
    blendEquation?: BlendingEquation | undefined;
    blendEquationAlpha?: number | undefined;
    blending?: Blending | undefined;
    blendSrc?: BlendingSrcFactor | BlendingDstFactor | undefined;
    blendSrcAlpha?: number | undefined;
    clipIntersection?: boolean | undefined;
    clippingPlanes?: Plane[] | undefined;
    clipShadows?: boolean | undefined;
    colorWrite?: boolean | undefined;
    defines?: any;
    depthFunc?: DepthModes | undefined;
    depthTest?: boolean | undefined;
    depthWrite?: boolean | undefined;
    name?: string | undefined;
    opacity?: number | undefined;
    polygonOffset?: boolean | undefined;
    polygonOffsetFactor?: number | undefined;
    polygonOffsetUnits?: number | undefined;
    precision?: 'highp' | 'mediump' | 'lowp' | null | undefined;
    premultipliedAlpha?: boolean | undefined;
    dithering?: boolean | undefined;
    side?: Side | undefined;
    shadowSide?: Side | undefined;
    toneMapped?: boolean | undefined;
    transparent?: boolean | undefined;
    vertexColors?: boolean | undefined;
    visible?: boolean | undefined;
    format?: PixelFormat | undefined;
    stencilWrite?: boolean | undefined;
    stencilFunc?: StencilFunc | undefined;
    stencilRef?: number | undefined;
    stencilWriteMask?: number | undefined;
    stencilFuncMask?: number | undefined;
    stencilFail?: StencilOp | undefined;
    stencilZFail?: StencilOp | undefined;
    stencilZPass?: StencilOp | undefined;
    userData?: any;
}

/**
 * Materials describe the appearance of objects. They are defined in a (mostly) renderer-independent way, so you don't have to rewrite materials if you decide to use a different renderer.
 */
declare class Material extends EventDispatcher {
    constructor();

    /**
     * Sets the alpha value to be used when running an alpha test. Default is 0.
     * @default 0
     */
    alphaTest: number;

    /**
     * Enables alpha to coverage. Can only be used with MSAA-enabled rendering contexts.
     * @default false
     */
    alphaToCoverage: boolean;

    /**
     * Blending destination. It's one of the blending mode constants defined in Three.js. Default is {@link OneMinusSrcAlphaFactor}.
     * @default THREE.OneMinusSrcAlphaFactor
     */
    blendDst: BlendingDstFactor;

    /**
     * The tranparency of the .blendDst. Default is null.
     * @default null
     */
    blendDstAlpha: number | null;

    /**
     * Blending equation to use when applying blending. It's one of the constants defined in Three.js. Default is {@link AddEquation}.
     * @default THREE.AddEquation
     */
    blendEquation: BlendingEquation;

    /**
     * The tranparency of the .blendEquation. Default is null.
     * @default null
     */
    blendEquationAlpha: number | null;

    /**
     * Which blending to use when displaying objects with this material. Default is {@link NormalBlending}.
     * @default THREE.NormalBlending
     */
    blending: Blending;

    /**
     * Blending source. It's one of the blending mode constants defined in Three.js. Default is {@link SrcAlphaFactor}.
     * @default THREE.SrcAlphaFactor
     */
    blendSrc: BlendingSrcFactor | BlendingDstFactor;

    /**
     * The tranparency of the .blendSrc. Default is null.
     * @default null
     */
    blendSrcAlpha: number | null;

    /**
     * Changes the behavior of clipping planes so that only their intersection is clipped, rather than their union. Default is false.
     * @default false
     */
    clipIntersection: boolean;

    /**
     * User-defined clipping planes specified as THREE.Plane objects in world space.
     * These planes apply to the objects this material is attached to.
     * Points in space whose signed distance to the plane is negative are clipped (not rendered).
     * See the WebGL / clipping /intersection example. Default is null.
     * @default null
     */
    clippingPlanes: any;

    /**
     * Defines whether to clip shadows according to the clipping planes specified on this material. Default is false.
     * @default false
     */
    clipShadows: boolean;

    /**
     * Whether to render the material's color. This can be used in conjunction with a mesh's .renderOrder property to create invisible objects that occlude other objects. Default is true.
     * @default true
     */
    colorWrite: boolean;

    /**
     * Custom defines to be injected into the shader. These are passed in form of an object literal, with key/value pairs. { MY_CUSTOM_DEFINE: '' , PI2: Math.PI * 2 }.
     * The pairs are defined in both vertex and fragment shaders. Default is undefined.
     * @default undefined
     */
    defines: undefined | { [key: string]: any };

    /**
     * Which depth function to use. Default is {@link LessEqualDepth}. See the depth mode constants for all possible values.
     * @default THREE.LessEqualDepth
     */
    depthFunc: DepthModes;

    /**
     * Whether to have depth test enabled when rendering this material. Default is true.
     * @default true
     */
    depthTest: boolean;

    /**
     * Whether rendering this material has any effect on the depth buffer. Default is true.
     * When drawing 2D overlays it can be useful to disable the depth writing in order to layer several things together without creating z-index artifacts.
     * @default true
     */
    depthWrite: boolean;

    /**
     * Unique number of this material instance.
     */
    id: number;

    /**
     * Whether rendering this material has any effect on the stencil buffer. Default is *false*.
     * @default false
     */
    stencilWrite: boolean;

    /**
     * The stencil comparison function to use. Default is {@link AlwaysStencilFunc}. See stencil operation constants for all possible values.
     * @default THREE.AlwaysStencilFunc
     */
    stencilFunc: StencilFunc;

    /**
     * The value to use when performing stencil comparisons or stencil operations. Default is *0*.
     * @default 0
     */
    stencilRef: number;

    /**
     * The bit mask to use when writing to the stencil buffer. Default is *0xFF*.
     * @default 0xff
     */
    stencilWriteMask: number;

    /**
     * The bit mask to use when comparing against the stencil buffer. Default is *0xFF*.
     * @default 0xff
     */
    stencilFuncMask: number;

    /**
     * Which stencil operation to perform when the comparison function returns false. Default is {@link KeepStencilOp}. See the stencil operation constants for all possible values.
     * @default THREE.KeepStencilOp
     */
    stencilFail: StencilOp;

    /**
     * Which stencil operation to perform when the comparison function returns true but the depth test fails.
     * Default is {@link KeepStencilOp}.
     * See the stencil operation constants for all possible values.
     * @default THREE.KeepStencilOp
     */
    stencilZFail: StencilOp;

    /**
     * Which stencil operation to perform when the comparison function returns true and the depth test passes.
     * Default is {@link KeepStencilOp}.
     * See the stencil operation constants for all possible values.
     * @default THREE.KeepStencilOp
     */
    stencilZPass: StencilOp;

    /**
     * Used to check whether this or derived classes are materials. Default is true.
     * You should not change this, as it used internally for optimisation.
     */
    readonly isMaterial: true;

    /**
     * Material name. Default is an empty string.
     * @default ''
     */
    name: string;

    /**
     * Specifies that the material needs to be updated, WebGL wise. Set it to true if you made changes that need to be reflected in WebGL.
     * This property is automatically set to true when instancing a new material.
     * @default false
     */
    needsUpdate: boolean;

    /**
     * Opacity. Default is 1.
     * @default 1
     */
    opacity: number;

    /**
     * Whether to use polygon offset. Default is false. This corresponds to the POLYGON_OFFSET_FILL WebGL feature.
     * @default false
     */
    polygonOffset: boolean;

    /**
     * Sets the polygon offset factor. Default is 0.
     * @default 0
     */
    polygonOffsetFactor: number;

    /**
     * Sets the polygon offset units. Default is 0.
     * @default 0
     */
    polygonOffsetUnits: number;

    /**
     * Override the renderer's default precision for this material. Can be "highp", "mediump" or "lowp". Defaults is null.
     * @default null
     */
    precision: 'highp' | 'mediump' | 'lowp' | null;

    /**
     * Whether to premultiply the alpha (transparency) value. See WebGL / Materials / Transparency for an example of the difference. Default is false.
     * @default false
     */
    premultipliedAlpha: boolean;

    /**
     * Whether to apply dithering to the color to remove the appearance of banding. Default is false.
     * @default false
     */
    dithering: boolean;

    /**
     * Defines which of the face sides will be rendered - front, back or both.
     * Default is THREE.FrontSide. Other options are THREE.BackSide and THREE.DoubleSide.
     * @default THREE.FrontSide
     */
    side: Side;

    /**
     * Defines which of the face sides will cast shadows. Default is *null*.
     * If *null*, the value is opposite that of side, above.
     * @default null
     */
    shadowSide: Side | null;

    /**
     * Defines whether this material is tone mapped according to the renderer's toneMapping setting.
     * Default is true.
     * @default true
     */
    toneMapped: boolean;

    /**
     * Defines whether this material is transparent. This has an effect on rendering as transparent objects need special treatment and are rendered after non-transparent objects.
     * When set to true, the extent to which the material is transparent is controlled by setting it's .opacity property.
     * Default is false.
     * @default false
     */
    transparent: boolean;

    /**
     * Value is the string 'Material'. This shouldn't be changed, and can be used to find all objects of this type in a scene.
     * @default 'Material'
     */
    type: string;

    /**
     * UUID of this material instance. This gets automatically assigned, so this shouldn't be edited.
     */
    uuid: string;

    /**
     * Defines whether vertex coloring is used. Default is false.
     * @default false
     */
    vertexColors: boolean;

    /**
     * Defines whether this material is visible. Default is true.
     * @default true
     */
    visible: boolean;

    /**
     * An object that can be used to store custom data about the Material. It should not hold references to functions as these will not be cloned.
     * @default {}
     */
    userData: any;

    /**
     * This starts at 0 and counts how many times .needsUpdate is set to true.
     * @default 0
     */
    version: number;

    /**
     * Return a new material with the same parameters as this material.
     */
    clone(): this;

    /**
     * Copy the parameters from the passed material into this material.
     * @param material
     */
    copy(material: Material): this;

    /**
     * This disposes the material. Textures of a material don't get disposed. These needs to be disposed by {@link Texture}.
     */
    dispose(): void;

    /**
     * An optional callback that is executed immediately before the shader program is compiled.
     * This function is called with the shader source code as a parameter.
     * Useful for the modification of built-in materials.
     * @param shader Source code of the shader
     * @param renderer WebGLRenderer Context that is initializing the material
     */
    onBeforeCompile(shader: Shader, renderer: WebGLRenderer): void;

    /**
     * In case onBeforeCompile is used, this callback can be used to identify values of settings used in onBeforeCompile, so three.js can reuse a cached shader or recompile the shader as needed.
     */
    customProgramCacheKey(): string;

    /**
     * Sets the properties based on the values.
     * @param values A container with parameters.
     */
    setValues(values: MaterialParameters): void;

    /**
     * Convert the material to three.js JSON format.
     * @param meta Object containing metadata such as textures or images for the material.
     */
    toJSON(meta?: any): any;
}

/**
 * Represents the data source of a texture.
 */
declare class Source {
    /**
     * @param [data] The data definition of a texture. default is **null**.
     */
    constructor(data: any);

    /**
     * The actual data of a texture. The type of this property depends on the texture that uses this instance.
     */
    data: any;

    /**
     * Set this to **true** to trigger a data upload to the GPU next time the source is used.
     */
    set needsUpdate(value: boolean);

    /**
     * [UUID](http://en.wikipedia.org/wiki/Universally_unique_identifier) of this object instance.
     * This gets automatically assigned, so this shouldn't be edited.
     */
    uuid: string;

    /**
     * This starts at **0** and counts how many times [property:Boolean needsUpdate] is set to **true**.
     */
    version: number;

    /**
     * Convert the data source to three.js [JSON Object/Scene format](https://github.com/mrdoob/three.js/wiki/JSON-Object-Scene-format-4).
     *
     * @param [meta] optional object containing metadata.
     */
    toJSON(meta: any): any;

    readonly isTexture: true;
}

declare class Texture extends EventDispatcher {
    /**
     * @param [image]
     * @param [mapping=THREE.Texture.DEFAULT_MAPPING]
     * @param [wrapS=THREE.ClampToEdgeWrapping]
     * @param [wrapT=THREE.ClampToEdgeWrapping]
     * @param [magFilter=THREE.LinearFilter]
     * @param [minFilter=THREE.LinearMipmapLinearFilter]
     * @param [format=THREE.RGBAFormat]
     * @param [type=THREE.UnsignedByteType]
     * @param [anisotropy=1]
     * @param [encoding=THREE.LinearEncoding]
     */
    constructor(
        image?: HTMLImageElement | HTMLCanvasElement | HTMLVideoElement,
        mapping?: Mapping,
        wrapS?: Wrapping,
        wrapT?: Wrapping,
        magFilter?: TextureFilter,
        minFilter?: TextureFilter,
        format?: PixelFormat,
        type?: TextureDataType,
        anisotropy?: number,
        encoding?: TextureEncoding,
    );

    id: number;
    uuid: string;

    /**
     * @default ''
     */
    name: string;
    sourceFile: string;

    /**
     * The data definition of a texture. A reference to the data source can be shared across textures.
     * This is often useful in context of spritesheets where multiple textures render the same data but with different texture transformations.
     */
    source: Source;

    /**
     * An image object, typically created using the {@link TextureLoader.load} method.
     * This can be any image (e.g., PNG, JPG, GIF, DDS) or video (e.g., MP4, OGG/OGV) type supported by three.js.
     *
     * To use video as a texture you need to have a playing HTML5
     * video element as a source for your texture image and continuously update this texture
     * as long as video is playing - the {@link VideoTexture} class handles this automatically.
     */
    get image(): any;

    /**
     * An image object, typically created using the {@link TextureLoader.load} method.
     * This can be any image (e.g., PNG, JPG, GIF, DDS) or video (e.g., MP4, OGG/OGV) type supported by three.js.
     *
     * To use video as a texture you need to have a playing HTML5
     * video element as a source for your texture image and continuously update this texture
     * as long as video is playing - the {@link VideoTexture} class handles this automatically.
     */
    set image(data: any);

    /**
     * @default []
     */
    mipmaps: any[]; // ImageData[] for 2D textures and CubeTexture[] for cube textures;

    /**
     * @default THREE.Texture.DEFAULT_MAPPING
     */
    mapping: Mapping;

    /**
     * @default THREE.ClampToEdgeWrapping
     */
    wrapS: Wrapping;

    /**
     * @default THREE.ClampToEdgeWrapping
     */
    wrapT: Wrapping;

    /**
     * @default THREE.LinearFilter
     */
    magFilter: TextureFilter;

    /**
     * @default THREE.LinearMipmapLinearFilter
     */
    minFilter: TextureFilter;

    /**
     * @default 1
     */
    anisotropy: number;

    /**
     * @default THREE.RGBAFormat
     */
    format: PixelFormat;

    internalFormat: PixelFormatGPU | null;

    /**
     * @default THREE.UnsignedByteType
     */
    type: TextureDataType;

    /**
     * @default new THREE.Matrix3()
     */
    matrix: Matrix3;

    /**
     * @default true
     */
    matrixAutoUpdate: boolean;

    /**
     * @default new THREE.Vector2( 0, 0 )
     */
    offset: Vector2;

    /**
     * @default new THREE.Vector2( 1, 1 )
     */
    repeat: Vector2;

    /**
     * @default new THREE.Vector2( 0, 0 )
     */
    center: Vector2;

    /**
     * @default 0
     */
    rotation: number;

    /**
     * @default true
     */
    generateMipmaps: boolean;

    /**
     * @default false
     */
    premultiplyAlpha: boolean;

    /**
     * @default true
     */
    flipY: boolean;

    /**
     * @default 4
     */
    unpackAlignment: number;

    /**
     * @default THREE.LinearEncoding
     */
    encoding: TextureEncoding;

    /**
     * @default false
     */
    isRenderTargetTexture: boolean;

    /**
     * @default false
     */
    needsPMREMUpdate: boolean;

    /**
     * An object that can be used to store custom data about the Material. It should not hold references to functions as these will not be cloned.
     * @default {}
     */
    userData: any;

    /**
     * @default 0
     */
    version: number;
    set needsUpdate(value: boolean);
    readonly isTexture: true;

    onUpdate: () => void;
    static DEFAULT_IMAGE: any;
    static DEFAULT_MAPPING: any;

    clone(): this;
    copy(source: Texture): this;
    toJSON(meta: any): any;
    dispose(): void;
    transformUv(uv: Vector2): Vector2;
    updateMatrix(): void;
}

// Scenes /////////////////////////////////////////////////////////////////////

/**
 * Scenes allow you to set up what and where is to be rendered by three.js. This is where you place objects, lights and cameras.
 */
declare class Scene$1 extends Object3D {
    constructor();

    type: 'Scene';

    /**
     * A fog instance defining the type of fog that affects everything rendered in the scene. Default is null.
     * @default null
     */
    fog: FogBase | null;

    /**
     * If not null, it will force everything in the scene to be rendered with that material. Default is null.
     * @default null
     */
    overrideMaterial: Material | null;

    /**
     * @default true
     */
    autoUpdate: boolean;

    /**
     * @default null
     */
    background: null | Color | Texture;

    /**
     * @default null
     */
    environment: null | Texture;

    readonly isScene: true;

    toJSON(meta?: any): any;
}

interface WebGLCapabilitiesParameters {
    precision?: string | undefined;
    logarithmicDepthBuffer?: boolean | undefined;
}

declare class WebGLCapabilities {
    constructor(gl: WebGLRenderingContext, extensions: any, parameters: WebGLCapabilitiesParameters);

    readonly isWebGL2: boolean;
    precision: string;
    logarithmicDepthBuffer: boolean;
    maxTextures: number;
    maxVertexTextures: number;
    maxTextureSize: number;
    maxCubemapSize: number;
    maxAttributes: number;
    maxVertexUniforms: number;
    maxVaryings: number;
    maxFragmentUniforms: number;
    vertexTextures: boolean;
    floatFragmentTextures: boolean;
    floatVertexTextures: boolean;

    getMaxAnisotropy(): number;
    getMaxPrecision(precision: string): string;
}

declare class WebGLExtensions {
    constructor(gl: WebGLRenderingContext);

    has(name: string): boolean;
    init(capabilities: WebGLCapabilities): void;
    get(name: string): any;
}

declare function WebGLShader(gl: WebGLRenderingContext, type: string, string: string): WebGLShader;

type Vector4Tuple = [number, number, number, number];

/**
 * 4D vector.
 *
 * ( class Vector4 implements Vector<Vector4> )
 */
declare class Vector4 implements Vector {
    constructor(x?: number, y?: number, z?: number, w?: number);

    /**
     * @default 0
     */
    x: number;

    /**
     * @default 0
     */
    y: number;

    /**
     * @default 0
     */
    z: number;

    /**
     * @default 0
     */
    w: number;

    width: number;
    height: number;
    readonly isVector4: true;

    /**
     * Sets value of this vector.
     */
    set(x: number, y: number, z: number, w: number): this;

    /**
     * Sets all values of this vector.
     */
    setScalar(scalar: number): this;

    /**
     * Sets X component of this vector.
     */
    setX(x: number): this;

    /**
     * Sets Y component of this vector.
     */
    setY(y: number): this;

    /**
     * Sets Z component of this vector.
     */
    setZ(z: number): this;

    /**
     * Sets w component of this vector.
     */
    setW(w: number): this;

    setComponent(index: number, value: number): this;

    getComponent(index: number): number;

    /**
     * Clones this vector.
     */
    clone(): this;

    /**
     * Copies value of v to this vector.
     */
    copy(v: Vector4): this;

    /**
     * Adds v to this vector.
     */
    add(v: Vector4): this;

    addScalar(scalar: number): this;

    /**
     * Sets this vector to a + b.
     */
    addVectors(a: Vector4, b: Vector4): this;

    addScaledVector(v: Vector4, s: number): this;
    /**
     * Subtracts v from this vector.
     */
    sub(v: Vector4): this;

    subScalar(s: number): this;

    /**
     * Sets this vector to a - b.
     */
    subVectors(a: Vector4, b: Vector4): this;

    multiply(v: Vector4): this;

    /**
     * Multiplies this vector by scalar s.
     */
    multiplyScalar(s: number): this;

    applyMatrix4(m: Matrix4): this;

    /**
     * Divides this vector by scalar s.
     * Set vector to ( 0, 0, 0 ) if s == 0.
     */
    divideScalar(s: number): this;

    /**
     * http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToAngle/index.htm
     * @param q is assumed to be normalized
     */
    setAxisAngleFromQuaternion(q: Quaternion): this;

    /**
     * http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToAngle/index.htm
     * @param m assumes the upper 3x3 of m is a pure rotation matrix (i.e, unscaled)
     */
    setAxisAngleFromRotationMatrix(m: Matrix4): this;

    min(v: Vector4): this;
    max(v: Vector4): this;
    clamp(min: Vector4, max: Vector4): this;
    clampScalar(min: number, max: number): this;
    floor(): this;
    ceil(): this;
    round(): this;
    roundToZero(): this;

    /**
     * Inverts this vector.
     */
    negate(): this;

    /**
     * Computes dot product of this vector and v.
     */
    dot(v: Vector4): number;

    /**
     * Computes squared length of this vector.
     */
    lengthSq(): number;

    /**
     * Computes length of this vector.
     */
    length(): number;

    /**
     * Computes the Manhattan length of this vector.
     *
     * see {@link http://en.wikipedia.org/wiki/Taxicab_geometry|Wikipedia: Taxicab Geometry}
     */
    manhattanLength(): number;

    /**
     * Normalizes this vector.
     */
    normalize(): this;
    /**
     * Normalizes this vector and multiplies it by l.
     */
    setLength(length: number): this;

    /**
     * Linearly interpolate between this vector and v with alpha factor.
     */
    lerp(v: Vector4, alpha: number): this;

    lerpVectors(v1: Vector4, v2: Vector4, alpha: number): this;

    /**
     * Checks for strict equality of this vector and v.
     */
    equals(v: Vector4): boolean;

    /**
     * Sets this vector's x, y, z and w value from the provided array or array-like.
     * @param array the source array or array-like.
     * @param offset (optional) offset into the array. Default is 0.
     */
    fromArray(array: number[] | ArrayLike<number>, offset?: number): this;

    /**
     * Returns an array [x, y, z, w], or copies x, y, z and w into the provided array.
     * @param array (optional) array to store the vector to. If this is not provided, a new array will be created.
     * @param offset (optional) optional offset into the array.
     * @return The created or provided array.
     */
    toArray(array?: number[], offset?: number): number[];
    toArray(array?: Vector4Tuple, offset?: 0): Vector4Tuple;

    /**
     * Copies x, y, z and w into the provided array-like.
     * @param array array-like to store the vector to.
     * @param offset (optional) optional offset into the array-like.
     * @return The provided array-like.
     */
    toArray(array: ArrayLike<number>, offset?: number): ArrayLike<number>;

    fromBufferAttribute(attribute: BufferAttribute, index: number): this;

    /**
     * Sets this vector's x, y, z and w from Math.random
     */
    random(): this;
}

declare class WebGLColorBuffer {
    constructor();

    setMask(colorMask: boolean): void;
    setLocked(lock: boolean): void;
    setClear(r: number, g: number, b: number, a: number, premultipliedAlpha: boolean): void;
    reset(): void;
}

declare class WebGLDepthBuffer {
    constructor();

    setTest(depthTest: boolean): void;
    setMask(depthMask: boolean): void;
    setFunc(depthFunc: DepthModes): void;
    setLocked(lock: boolean): void;
    setClear(depth: number): void;
    reset(): void;
}

declare class WebGLStencilBuffer {
    constructor();

    setTest(stencilTest: boolean): void;
    setMask(stencilMask: number): void;
    setFunc(stencilFunc: number, stencilRef: number, stencilMask: number): void;
    setOp(stencilFail: number, stencilZFail: number, stencilZPass: number): void;
    setLocked(lock: boolean): void;
    setClear(stencil: number): void;
    reset(): void;
}

declare class WebGLState {
    constructor(gl: WebGLRenderingContext, extensions: WebGLExtensions, capabilities: WebGLCapabilities);

    buffers: {
        color: WebGLColorBuffer;
        depth: WebGLDepthBuffer;
        stencil: WebGLStencilBuffer;
    };

    initAttributes(): void;
    enableAttribute(attribute: number): void;
    enableAttributeAndDivisor(attribute: number, meshPerAttribute: number): void;
    disableUnusedAttributes(): void;
    vertexAttribPointer(
        index: number,
        size: number,
        type: number,
        normalized: boolean,
        stride: number,
        offset: number,
    ): void;
    enable(id: number): void;
    disable(id: number): void;
    bindFramebuffer(target: number, framebuffer: WebGLFramebuffer | null): void;
    bindXRFramebuffer(framebuffer: WebGLFramebuffer | null): void;
    useProgram(program: any): boolean;
    setBlending(
        blending: Blending,
        blendEquation?: BlendingEquation,
        blendSrc?: BlendingSrcFactor,
        blendDst?: BlendingDstFactor,
        blendEquationAlpha?: BlendingEquation,
        blendSrcAlpha?: BlendingSrcFactor,
        blendDstAlpha?: BlendingDstFactor,
        premultiplyAlpha?: boolean,
    ): void;
    setMaterial(material: Material, frontFaceCW: boolean): void;
    setFlipSided(flipSided: boolean): void;
    setCullFace(cullFace: CullFace): void;
    setLineWidth(width: number): void;
    setPolygonOffset(polygonoffset: boolean, factor?: number, units?: number): void;
    setScissorTest(scissorTest: boolean): void;
    activeTexture(webglSlot: number): void;
    bindTexture(webglType: number, webglTexture: any): void;
    unbindTexture(): void;
    // Same interface as https://developer.mozilla.org/en-US/docs/Web/API/WebGLRenderingContext/compressedTexImage2D
    compressedTexImage2D(
        target: number,
        level: number,
        internalformat: number,
        width: number,
        height: number,
        border: number,
        data: ArrayBufferView,
    ): void;
    // Same interface as https://developer.mozilla.org/en-US/docs/Web/API/WebGLRenderingContext/texImage2D
    texImage2D(
        target: number,
        level: number,
        internalformat: number,
        width: number,
        height: number,
        border: number,
        format: number,
        type: number,
        pixels: ArrayBufferView | null,
    ): void;
    texImage2D(target: number, level: number, internalformat: number, format: number, type: number, source: any): void;
    texImage3D(
        target: number,
        level: number,
        internalformat: number,
        width: number,
        height: number,
        depth: number,
        border: number,
        format: number,
        type: number,
        pixels: any,
    ): void;
    scissor(scissor: Vector4): void;
    viewport(viewport: Vector4): void;
    reset(): void;
}

declare class WebGLProperties {
    constructor();

    get(object: any): any;
    remove(object: any): void;
    update(object: any, key: any, value: any): any;
    dispose(): void;
}

declare class WebGLUtils {
    constructor(gl: WebGLRenderingContext | WebGL2RenderingContext, extensions: any, capabilities: any);

    convert(p: PixelFormat | CompressedPixelFormat | TextureDataType, encoding?: TextureEncoding | null): number | null;
}

declare class WebGLTextures {
    constructor(
        gl: WebGLRenderingContext,
        extensions: WebGLExtensions,
        state: WebGLState,
        properties: WebGLProperties,
        capabilities: WebGLCapabilities,
        utils: WebGLUtils,
        info: WebGLInfo,
    );

    allocateTextureUnit(): void;
    resetTextureUnits(): void;
    setTexture2D(texture: any, slot: number): void;
    setTexture2DArray(texture: any, slot: number): void;
    setTexture3D(texture: any, slot: number): void;
    setTextureCube(texture: any, slot: number): void;
    setupRenderTarget(renderTarget: any): void;
    updateRenderTargetMipmap(renderTarget: any): void;
    updateMultisampleRenderTarget(renderTarget: any): void;
    safeSetTexture2D(texture: any, slot: number): void;
    safeSetTextureCube(texture: any, slot: number): void;
}

declare class WebGLUniforms {
    constructor(gl: WebGLRenderingContext, program: WebGLProgram);

    setValue(gl: WebGLRenderingContext, name: string, value: any, textures: WebGLTextures): void;
    setOptional(gl: WebGLRenderingContext, object: any, name: string): void;

    static upload(gl: WebGLRenderingContext, seq: any, values: any[], textures: WebGLTextures): void;
    static seqWithValue(seq: any, values: any[]): any[];
}

declare class WebGLProgram {
    constructor(renderer: WebGLRenderer, cacheKey: string, parameters: object);

    name: string;
    id: number;
    cacheKey: string; // unique identifier for this program, used for looking up compiled programs from cache.

    /**
     * @default 1
     */
    usedTimes: number;
    program: any;
    vertexShader: WebGLShader;
    fragmentShader: WebGLShader;
    /**
     * @deprecated Use {@link WebGLProgram#getUniforms getUniforms()} instead.
     */
    uniforms: any;
    /**
     * @deprecated Use {@link WebGLProgram#getAttributes getAttributes()} instead.
     */
    attributes: any;

    getUniforms(): WebGLUniforms;
    getAttributes(): any;
    destroy(): void;
}

/**
 * An object with a series of statistical information about the graphics board memory and the rendering process.
 */
declare class WebGLInfo {
    constructor(gl: WebGLRenderingContext);

    /**
     * @default true
     */
    autoReset: boolean;

    /**
     * @default { geometries: 0, textures: 0 }
     */
    memory: {
        geometries: number;
        textures: number;
    };

    /**
     * @default null
     */
    programs: WebGLProgram[] | null;

    /**
     * @default { frame: 0, calls: 0, triangles: 0, points: 0, lines: 0 }
     */
    render: {
        calls: number;
        frame: number;
        lines: number;
        points: number;
        triangles: number;
    };
    update(count: number, mode: number, instanceCount: number): void;
    reset(): void;
}

declare class WebGLObjects {
    constructor(gl: WebGLRenderingContext, geometries: any, attributes: any, info: any);

    update(object: any): any;
    dispose(): void;
}

declare class DepthTexture extends Texture {
    /**
     * @param width
     * @param height
     * @param type
     * @param [mapping=THREE.Texture.DEFAULT_MAPPING]
     * @param [wrapS=THREE.ClampToEdgeWrapping]
     * @param [wrapT=THREE.ClampToEdgeWrapping]
     * @param [magFilter=THREE.NearestFilter]
     * @param [minFilter=THREE.NearestFilter]
     * @param [anisotropy=1]
     */
    constructor(
        width: number,
        height: number,
        type?: TextureDataType,
        mapping?: Mapping,
        wrapS?: Wrapping,
        wrapT?: Wrapping,
        magFilter?: TextureFilter,
        minFilter?: TextureFilter,
        anisotropy?: number,
    );

    get image(): { width: number; height: number };
    set image(value: { width: number; height: number });

    /**
     * @default false
     */
    flipY: boolean;

    /**
     * @default false
     */
    generateMipmaps: boolean;

    readonly isDepthTexture: true;
}

interface WebGLRenderTargetOptions {
    wrapS?: Wrapping | undefined;
    wrapT?: Wrapping | undefined;
    magFilter?: TextureFilter | undefined;
    minFilter?: TextureFilter | undefined;
    format?: number | undefined; // RGBAFormat;
    type?: TextureDataType | undefined; // UnsignedByteType;
    anisotropy?: number | undefined; // 1;
    depthBuffer?: boolean | undefined; // true;
    stencilBuffer?: boolean | undefined; // false;
    generateMipmaps?: boolean | undefined; // true;
    depthTexture?: DepthTexture | undefined;
    encoding?: TextureEncoding | undefined;

    /**
     * Defines the count of MSAA samples. Can only be used with WebGL 2. Default is **0**.
     * @default 0
     */
    samples?: number;
}

declare class WebGLRenderTarget extends EventDispatcher {
    constructor(width: number, height: number, options?: WebGLRenderTargetOptions);

    width: number;
    height: number;
    depth: number;

    scissor: Vector4;
    /**
     * @default false
     */
    scissorTest: boolean;
    viewport: Vector4;
    texture: Texture;

    /**
     * @default true
     */
    depthBuffer: boolean;

    /**
     * @default true
     */
    stencilBuffer: boolean;

    /**
     * @default null
     */
    depthTexture: DepthTexture;

    /**
     * Defines the count of MSAA samples. Can only be used with WebGL 2. Default is **0**.
     * @default 0
     */
    samples: number;

    readonly isWebGLRenderTarget: true;

    /**
     * @deprecated Use {@link Texture#wrapS texture.wrapS} instead.
     */
    wrapS: any;
    /**
     * @deprecated Use {@link Texture#wrapT texture.wrapT} instead.
     */
    wrapT: any;
    /**
     * @deprecated Use {@link Texture#magFilter texture.magFilter} instead.
     */
    magFilter: any;
    /**
     * @deprecated Use {@link Texture#minFilter texture.minFilter} instead.
     */
    minFilter: any;
    /**
     * @deprecated Use {@link Texture#anisotropy texture.anisotropy} instead.
     */
    anisotropy: any;
    /**
     * @deprecated Use {@link Texture#offset texture.offset} instead.
     */
    offset: any;
    /**
     * @deprecated Use {@link Texture#repeat texture.repeat} instead.
     */
    repeat: any;
    /**
     * @deprecated Use {@link Texture#format texture.format} instead.
     */
    format: any;
    /**
     * @deprecated Use {@link Texture#type texture.type} instead.
     */
    type: any;
    /**
     * @deprecated Use {@link Texture#generateMipmaps texture.generateMipmaps} instead.
     */
    generateMipmaps: any;

    setSize(width: number, height: number, depth?: number): void;
    clone(): this;
    copy(source: WebGLRenderTarget): this;
    dispose(): void;
}

declare class LightShadow {
    constructor(camera: Camera);

    camera: Camera;

    /**
     * @default 0
     */
    bias: number;

    /**
     * @default 0
     */
    normalBias: number;

    /**
     * @default 1
     */
    radius: number;

    /**
     * @default 8
     */
    blurSamples: number;

    /**
     * @default new THREE.Vector2( 512, 512 )
     */
    mapSize: Vector2;

    /**
     * @default null
     */
    map: WebGLRenderTarget;

    /**
     * @default null
     */
    mapPass: WebGLRenderTarget;

    /**
     * @default new THREE.Matrix4()
     */
    matrix: Matrix4;

    /**
     * @default true
     */
    autoUpdate: boolean;

    /**
     * @default false
     */
    needsUpdate: boolean;

    copy(source: LightShadow): this;
    clone(recursive?: boolean): this;
    toJSON(): any;
    getFrustum(): number;
    updateMatrices(light: Light, viewportIndex?: number): void;
    getViewport(viewportIndex: number): Vector4;
    getFrameExtents(): Vector2;
    dispose(): void;
}

// Lights //////////////////////////////////////////////////////////////////////////////////

/**
 * Abstract base class for lights.
 */
declare class Light extends Object3D {
    constructor(hex?: number | string, intensity?: number);

    /**
     * @default 'Light'
     */
    type: string;

    color: Color;

    /**
     * @default 1
     */
    intensity: number;
    readonly isLight: true;

    shadow: LightShadow;
    /**
     * @deprecated Use shadow.camera.fov instead.
     */
    shadowCameraFov: any;
    /**
     * @deprecated Use shadow.camera.left instead.
     */
    shadowCameraLeft: any;
    /**
     * @deprecated Use shadow.camera.right instead.
     */
    shadowCameraRight: any;
    /**
     * @deprecated Use shadow.camera.top instead.
     */
    shadowCameraTop: any;
    /**
     * @deprecated Use shadow.camera.bottom instead.
     */
    shadowCameraBottom: any;
    /**
     * @deprecated Use shadow.camera.near instead.
     */
    shadowCameraNear: any;
    /**
     * @deprecated Use shadow.camera.far instead.
     */
    shadowCameraFar: any;
    /**
     * @deprecated Use shadow.bias instead.
     */
    shadowBias: any;
    /**
     * @deprecated Use shadow.mapSize.width instead.
     */
    shadowMapWidth: any;
    /**
     * @deprecated Use shadow.mapSize.height instead.
     */
    shadowMapHeight: any;

    dispose(): void;
}

declare class WebGLShadowMap {
    constructor(_renderer: WebGLRenderer, _objects: WebGLObjects, _capabilities: WebGLCapabilities);

    /**
     * @default false
     */
    enabled: boolean;

    /**
     * @default true
     */
    autoUpdate: boolean;

    /**
     * @default false
     */
    needsUpdate: boolean;

    /**
     * @default THREE.PCFShadowMap
     */
    type: ShadowMapType;

    render(shadowsArray: Light[], scene: Scene$1, camera: Camera): void;

    /**
     * @deprecated Use {@link Material#shadowSide} instead.
     */
    cullFace: any;
}

declare class Group extends Object3D {
    constructor();
    type: 'Group';
    readonly isGroup: true;
}

/**
 * This is a superefficent class for geometries because it saves all data in buffers.
 * It reduces memory costs and cpu cycles. But it is not as easy to work with because of all the necessary buffer calculations.
 * It is mainly interesting when working with static objects.
 *
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/core/BufferGeometry.js|src/core/BufferGeometry.js}
 */
declare class BufferGeometry extends EventDispatcher {
    /**
     * This creates a new BufferGeometry. It also sets several properties to an default value.
     */
    constructor();

    /**
     * Unique number of this buffergeometry instance
     */
    id: number;
    uuid: string;

    /**
     * @default ''
     */
    name: string;

    /**
     * @default 'BufferGeometry'
     */
    type: string;

    /**
     * @default null
     */
    index: BufferAttribute | null;

    /**
     * @default {}
     */
    attributes: {
        [name: string]: BufferAttribute | InterleavedBufferAttribute;
    };

    /**
     * @default {}
     */
    morphAttributes: {
        [name: string]: Array<BufferAttribute | InterleavedBufferAttribute>;
    };

    /**
     * @default false
     */
    morphTargetsRelative: boolean;

    /**
     * @default []
     */
    groups: Array<{ start: number; count: number; materialIndex?: number | undefined }>;

    /**
     * @default null
     */
    boundingBox: Box3 | null;

    /**
     * @default null
     */
    boundingSphere: Sphere | null;

    /**
     * @default { start: 0, count: Infinity }
     */
    drawRange: { start: number; count: number };

    /**
     * @default {}
     */
    userData: { [key: string]: any };
    readonly isBufferGeometry: true;

    getIndex(): BufferAttribute | null;
    setIndex(index: BufferAttribute | number[] | null): BufferGeometry;

    setAttribute(
        name: BuiltinShaderAttributeName | (string & {}),
        attribute: BufferAttribute | InterleavedBufferAttribute,
    ): BufferGeometry;
    getAttribute(name: BuiltinShaderAttributeName | (string & {})): BufferAttribute | InterleavedBufferAttribute;
    deleteAttribute(name: BuiltinShaderAttributeName | (string & {})): BufferGeometry;
    hasAttribute(name: BuiltinShaderAttributeName | (string & {})): boolean;

    addGroup(start: number, count: number, materialIndex?: number): void;
    clearGroups(): void;

    setDrawRange(start: number, count: number): void;

    /**
     * Bakes matrix transform directly into vertex coordinates.
     */
    applyMatrix4(matrix: Matrix4): BufferGeometry;
    applyQuaternion(q: Quaternion): BufferGeometry;

    rotateX(angle: number): BufferGeometry;
    rotateY(angle: number): BufferGeometry;
    rotateZ(angle: number): BufferGeometry;
    translate(x: number, y: number, z: number): BufferGeometry;
    scale(x: number, y: number, z: number): BufferGeometry;
    lookAt(v: Vector3): void;

    center(): BufferGeometry;

    setFromPoints(points: Vector3[] | Vector2[]): BufferGeometry;

    /**
     * Computes bounding box of the geometry, updating Geometry.boundingBox attribute.
     * Bounding boxes aren't computed by default. They need to be explicitly computed, otherwise they are null.
     */
    computeBoundingBox(): void;

    /**
     * Computes bounding sphere of the geometry, updating Geometry.boundingSphere attribute.
     * Bounding spheres aren't' computed by default. They need to be explicitly computed, otherwise they are null.
     */
    computeBoundingSphere(): void;

    /**
     * Computes and adds tangent attribute to this geometry.
     */
    computeTangents(): void;

    /**
     * Computes vertex normals by averaging face normals.
     */
    computeVertexNormals(): void;

    merge(geometry: BufferGeometry, offset?: number): BufferGeometry;
    normalizeNormals(): void;

    toNonIndexed(): BufferGeometry;

    toJSON(): any;
    clone(): BufferGeometry;
    copy(source: BufferGeometry): this;

    /**
     * Disposes the object from memory.
     * You need to call this when you want the bufferGeometry removed while the application is running.
     */
    dispose(): void;

    /**
     * @deprecated Use {@link BufferGeometry#groups .groups} instead.
     */
    drawcalls: any;

    /**
     * @deprecated Use {@link BufferGeometry#groups .groups} instead.
     */
    offsets: any;

    /**
     * @deprecated Use {@link BufferGeometry#setIndex .setIndex()} instead.
     */
    addIndex(index: any): void;

    /**
     * @deprecated Use {@link BufferGeometry#addGroup .addGroup()} instead.
     */
    addDrawCall(start: any, count: any, indexOffset?: any): void;

    /**
     * @deprecated Use {@link BufferGeometry#clearGroups .clearGroups()} instead.
     */
    clearDrawCalls(): void;

    /**
     * @deprecated Use {@link BufferGeometry#setAttribute .setAttribute()} instead.
     */
    addAttribute(name: string, attribute: BufferAttribute | InterleavedBufferAttribute): BufferGeometry;
    addAttribute(name: any, array: any, itemSize: any): any;

    /**
     * @deprecated Use {@link BufferGeometry#deleteAttribute .deleteAttribute()} instead.
     */
    removeAttribute(name: string): BufferGeometry;
}

interface RenderItem {
    id: number;
    object: Object3D;
    geometry: BufferGeometry | null;
    material: Material;
    program: WebGLProgram;
    groupOrder: number;
    renderOrder: number;
    z: number;
    group: Group | null;
}

declare class WebGLRenderList {
    constructor(properties: WebGLProperties);

    /**
     * @default []
     */
    opaque: RenderItem[];

    /**
     * @default []
     */
    transparent: RenderItem[];

    /**
     * @default []
     */
    transmissive: RenderItem[];

    init(): void;
    push(
        object: Object3D,
        geometry: BufferGeometry | null,
        material: Material,
        groupOrder: number,
        z: number,
        group: Group | null,
    ): void;
    unshift(
        object: Object3D,
        geometry: BufferGeometry | null,
        material: Material,
        groupOrder: number,
        z: number,
        group: Group | null,
    ): void;
    sort(opaqueSort: (a: any, b: any) => number, transparentSort: (a: any, b: any) => number): void;
    finish(): void;
}

declare class WebGLRenderLists {
    constructor(properties: WebGLProperties);

    dispose(): void;
    get(scene: Scene$1, renderCallDepth: number): WebGLRenderList;
}

/**
 * This class originall extended WebGLMultipleRenderTarget
 * However, there are some issues with this method as documented below
 */
declare class WebGLMultipleRenderTargets extends EventDispatcher {
    texture: Texture[];

    readonly isWebGLMultipleRenderTargets = true;

    /**
     * @param width The width of the render target.
     * @param height The height of the render target.
     * @param count The number of render targets.
     * @param options object that holds texture parameters for an auto-generated target texture and depthBuffer/stencilBuffer booleans.
     * For an explanation of the texture parameters see {@link Texture}.
     */
    constructor(width: number, height: number, count: number, options?: WebGLRenderTargetOptions);

    setSize(width: number, height: number, depth?: number): this;
    copy(source: WebGLMultipleRenderTargets): this;
    clone(): this;
    dispose(): void;
    // This is an available method, however it will break the code see https://github.com/mrdoob/three.js/issues/21930
    setTexture(texture: Texture): void;
}

/**
 * Camera with perspective projection.
 *
 * @source https://github.com/mrdoob/three.js/blob/master/src/cameras/PerspectiveCamera.js
 */
declare class PerspectiveCamera extends Camera {
    /**
     * @param [fov=50] Camera frustum vertical field of view. Default value is 50.
     * @param [aspect=1] Camera frustum aspect ratio. Default value is 1.
     * @param [near=0.1] Camera frustum near plane. Default value is 0.1.
     * @param [far=2000] Camera frustum far plane. Default value is 2000.
     */
    constructor(fov?: number, aspect?: number, near?: number, far?: number);

    type: 'PerspectiveCamera';

    readonly isPerspectiveCamera: true;

    /**
     * @default 1
     */
    zoom: number;

    /**
     * Camera frustum vertical field of view, from bottom to top of view, in degrees.
     * @default 50
     */
    fov: number;

    /**
     * Camera frustum aspect ratio, window width divided by window height.
     * @default 1
     */
    aspect: number;

    /**
     * Camera frustum near plane.
     * @default 0.1
     */
    near: number;

    /**
     * Camera frustum far plane.
     * @default 2000
     */
    far: number;

    /**
     * @default 10
     */
    focus: number;

    /**
     * @default null
     */
    view: null | {
        enabled: boolean;
        fullWidth: number;
        fullHeight: number;
        offsetX: number;
        offsetY: number;
        width: number;
        height: number;
    };

    /**
     * @default 35
     */
    filmGauge: number;

    /**
     * @default 0
     */
    filmOffset: number;

    setFocalLength(focalLength: number): void;
    getFocalLength(): number;
    getEffectiveFOV(): number;
    getFilmWidth(): number;
    getFilmHeight(): number;

    /**
     * Sets an offset in a larger frustum. This is useful for multi-window or multi-monitor/multi-machine setups.
     * For example, if you have 3x2 monitors and each monitor is 1920x1080 and the monitors are in grid like this:
     *
     * +---+---+---+
     * | A | B | C |
     * +---+---+---+
     * | D | E | F |
     * +---+---+---+
     *
     * then for each monitor you would call it like this:
     *
     * const w = 1920;
     * const h = 1080;
     * const fullWidth = w * 3;
     * const fullHeight = h * 2;
     *
     * // A
     * camera.setViewOffset( fullWidth, fullHeight, w * 0, h * 0, w, h );
     * // B
     * camera.setViewOffset( fullWidth, fullHeight, w * 1, h * 0, w, h );
     * // C
     * camera.setViewOffset( fullWidth, fullHeight, w * 2, h * 0, w, h );
     * // D
     * camera.setViewOffset( fullWidth, fullHeight, w * 0, h * 1, w, h );
     * // E
     * camera.setViewOffset( fullWidth, fullHeight, w * 1, h * 1, w, h );
     * // F
     * camera.setViewOffset( fullWidth, fullHeight, w * 2, h * 1, w, h ); Note there is no reason monitors have to be the same size or in a grid.
     *
     * @param fullWidth full width of multiview setup
     * @param fullHeight full height of multiview setup
     * @param x horizontal offset of subcamera
     * @param y vertical offset of subcamera
     * @param width width of subcamera
     * @param height height of subcamera
     */
    setViewOffset(fullWidth: number, fullHeight: number, x: number, y: number, width: number, height: number): void;
    clearViewOffset(): void;

    /**
     * Updates the camera projection matrix. Must be called after change of parameters.
     */
    updateProjectionMatrix(): void;
    toJSON(meta?: any): any;

    /**
     * @deprecated Use {@link PerspectiveCamera#setFocalLength .setFocalLength()} and {@link PerspectiveCamera#filmGauge .filmGauge} instead.
     */
    setLens(focalLength: number, frameHeight?: number): void;
}

declare class ArrayCamera extends PerspectiveCamera {
    constructor(cameras?: PerspectiveCamera[]);

    /**
     * @default []
     */
    cameras: PerspectiveCamera[];
    readonly isArrayCamera: true;
}

type XRControllerEventType = XRSessionEventType | XRInputSourceEventType | 'disconnected' | 'connected';

type XRHandJoints = Record<XRHandJoint, number>;

interface XRHandInputState {
    pinching: boolean;
}

declare class XRHandSpace extends Group {
    readonly joints: Partial<XRHandJoints>;
    readonly inputState: XRHandInputState;
}

declare class XRTargetRaySpace extends Group {
    hasLinearVelocity: boolean;
    readonly linearVelocity: Vector3;
    hasAngularVelocity: boolean;
    readonly angularVelocity: Vector3;
}

declare class XRGripSpace extends Group {
    hasLinearVelocity: boolean;
    readonly linearVelocity: Vector3;
    hasAngularVelocity: boolean;
    readonly angularVelocity: Vector3;
}

declare class WebXRController {
    constructor();

    getHandSpace(): XRHandSpace;
    getTargetRaySpace(): XRTargetRaySpace;
    getGripSpace(): XRGripSpace;
    dispatchEvent(event: { type: XRControllerEventType; data?: XRInputSource }): this;
    disconnect(inputSource: XRInputSource): this;
    update(inputSource: XRInputSource, frame: XRFrame, referenceSpace: XRReferenceSpace): this;
}

type WebXRCamera = PerspectiveCamera & { viewport: Vector4 };
type WebXRArrayCamera = Omit<ArrayCamera, 'cameras'> & { cameras: [WebXRCamera, WebXRCamera] };

declare class WebXRManager extends EventDispatcher {
    constructor(renderer: any, gl: WebGLRenderingContext);

    /**
     * @default false
     */
    enabled: boolean;

    /**
     * @default false
     */
    isPresenting: boolean;

    /**
     * @default true
     */
    cameraAutoUpdate: boolean;

    getController(index: number): XRTargetRaySpace;
    getControllerGrip(index: number): XRGripSpace;
    getHand(index: number): XRHandSpace;
    setFramebufferScaleFactor(value: number): void;
    setReferenceSpaceType(value: XRReferenceSpaceType): void;
    getReferenceSpace(): XRReferenceSpace | null;
    setReferenceSpace(value: XRReferenceSpace): void;
    getBaseLayer(): XRWebGLLayer | XRProjectionLayer;
    getBinding(): XRWebGLBinding;
    getFrame(): XRFrame;
    getSession(): XRSession | null;
    setSession(value: XRSession): Promise<void>;
    getCamera(): WebXRArrayCamera;
    updateCamera(camera: PerspectiveCamera): void;
    setAnimationLoop(callback: XRFrameRequestCallback | null): void;
    getFoveation(): number | undefined;
    setFoveation(foveation: number): void;
    dispose(): void;
}

declare class Data3DTexture extends Texture {
    constructor(data: BufferSource, width: number, height: number, depth: number);

    /**
     * @default THREE.NearestFilter
     */
    magFilter: TextureFilter;

    /**
     * @default THREE.NearestFilter
     */
    minFilter: TextureFilter;

    /**
     * @default THREE.ClampToEdgeWrapping
     */
    wrapR: boolean;

    /**
     * @default false
     */
    flipY: boolean;

    /**
     * @default false
     */
    generateMipmaps: boolean;

    readonly isData3DTexture: true;
}

declare class DataArrayTexture extends Texture {
    constructor(data?: BufferSource, width?: number, height?: number, depth?: number);

    /**
     * @default THREE.NearestFilter
     */
    magFilter: TextureFilter;

    /**
     * @default THREE.NearestFilter
     */
    minFilter: TextureFilter;

    /**
     * @default THREE.ClampToEdgeWrapping
     */
    wrapR: boolean;

    /**
     * @default false
     */
    flipY: boolean;

    /**
     * @default false
     */
    generateMipmaps: boolean;

    readonly isDataArrayTexture: true;
}

interface Renderer {
    domElement: HTMLCanvasElement;

    render(scene: Object3D, camera: Camera): void;
    setSize(width: number, height: number, updateStyle?: boolean): void;
}

/** This is only available in worker JS contexts, not the DOM. */
// tslint:disable-next-line:no-empty-interface
interface OffscreenCanvas extends EventTarget {}

interface WebGLRendererParameters {
    /**
     * A Canvas where the renderer draws its output.
     */
    canvas?: HTMLCanvasElement | OffscreenCanvas | undefined;

    /**
     * A WebGL Rendering Context.
     * (https://developer.mozilla.org/en-US/docs/Web/API/WebGLRenderingContext)
     * Default is null
     */
    context?: WebGLRenderingContext | undefined;

    /**
     * shader precision. Can be "highp", "mediump" or "lowp".
     */
    precision?: string | undefined;

    /**
     * default is false.
     */
    alpha?: boolean | undefined;

    /**
     * default is true.
     */
    premultipliedAlpha?: boolean | undefined;

    /**
     * default is false.
     */
    antialias?: boolean | undefined;

    /**
     * default is true.
     */
    stencil?: boolean | undefined;

    /**
     * default is false.
     */
    preserveDrawingBuffer?: boolean | undefined;

    /**
     * Can be "high-performance", "low-power" or "default"
     */
    powerPreference?: string | undefined;

    /**
     * default is true.
     */
    depth?: boolean | undefined;

    /**
     * default is false.
     */
    logarithmicDepthBuffer?: boolean | undefined;

    /**
     * default is false.
     */
    failIfMajorPerformanceCaveat?: boolean | undefined;
}

interface WebGLDebug {
    /**
     * Enables error checking and reporting when shader programs are being compiled.
     */
    checkShaderErrors: boolean;
}

/**
 * The WebGL renderer displays your beautifully crafted scenes using WebGL, if your device supports it.
 * This renderer has way better performance than CanvasRenderer.
 *
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/renderers/WebGLRenderer.js|src/renderers/WebGLRenderer.js}
 */
declare class WebGLRenderer implements Renderer {
    /**
     * parameters is an optional object with properties defining the renderer's behaviour.
     * The constructor also accepts no parameters at all.
     * In all cases, it will assume sane defaults when parameters are missing.
     */
    constructor(parameters?: WebGLRendererParameters);

    /**
     * A Canvas where the renderer draws its output.
     * This is automatically created by the renderer in the constructor (if not provided already); you just need to add it to your page.
     * @default document.createElementNS( 'http://www.w3.org/1999/xhtml', 'canvas' )
     */
    domElement: HTMLCanvasElement;

    /**
     * Defines whether the renderer should automatically clear its output before rendering.
     * @default true
     */
    autoClear: boolean;

    /**
     * If autoClear is true, defines whether the renderer should clear the color buffer. Default is true.
     * @default true
     */
    autoClearColor: boolean;

    /**
     * If autoClear is true, defines whether the renderer should clear the depth buffer. Default is true.
     * @default true
     */
    autoClearDepth: boolean;

    /**
     * If autoClear is true, defines whether the renderer should clear the stencil buffer. Default is true.
     * @default true
     */
    autoClearStencil: boolean;

    /**
     * Debug configurations.
     * @default { checkShaderErrors: true }
     */
    debug: WebGLDebug;

    /**
     * Defines whether the renderer should sort objects. Default is true.
     * @default true
     */
    sortObjects: boolean;

    /**
     * @default []
     */
    clippingPlanes: any[];

    /**
     * @default false
     */
    localClippingEnabled: boolean;

    extensions: WebGLExtensions;

    /**
     * Default is LinearEncoding.
     * @default THREE.LinearEncoding
     */
    outputEncoding: TextureEncoding;

    /**
     * @default false
     */
    physicallyCorrectLights: boolean;

    /**
     * @default THREE.NoToneMapping
     */
    toneMapping: ToneMapping;

    /**
     * @default 1
     */
    toneMappingExposure: number;

    info: WebGLInfo;

    shadowMap: WebGLShadowMap;

    pixelRatio: number;

    capabilities: WebGLCapabilities;
    properties: WebGLProperties;
    renderLists: WebGLRenderLists;
    state: WebGLState;

    xr: WebXRManager;

    /**
     * Return the WebGL context.
     */
    getContext(): WebGLRenderingContext | WebGL2RenderingContext;
    getContextAttributes(): any;
    forceContextLoss(): void;
    forceContextRestore(): void;

    /**
     * @deprecated Use {@link WebGLCapabilities#getMaxAnisotropy .capabilities.getMaxAnisotropy()} instead.
     */
    getMaxAnisotropy(): number;

    /**
     * @deprecated Use {@link WebGLCapabilities#precision .capabilities.precision} instead.
     */
    getPrecision(): string;

    getPixelRatio(): number;
    setPixelRatio(value: number): void;

    getDrawingBufferSize(target: Vector2): Vector2;
    setDrawingBufferSize(width: number, height: number, pixelRatio: number): void;

    getSize(target: Vector2): Vector2;

    /**
     * Resizes the output canvas to (width, height), and also sets the viewport to fit that size, starting in (0, 0).
     */
    setSize(width: number, height: number, updateStyle?: boolean): void;

    getCurrentViewport(target: Vector4): Vector4;

    /**
     * Copies the viewport into target.
     */
    getViewport(target: Vector4): Vector4;

    /**
     * Sets the viewport to render from (x, y) to (x + width, y + height).
     * (x, y) is the lower-left corner of the region.
     */
    setViewport(x: Vector4 | number, y?: number, width?: number, height?: number): void;

    /**
     * Copies the scissor area into target.
     */
    getScissor(target: Vector4): Vector4;

    /**
     * Sets the scissor area from (x, y) to (x + width, y + height).
     */
    setScissor(x: Vector4 | number, y?: number, width?: number, height?: number): void;

    /**
     * Returns true if scissor test is enabled; returns false otherwise.
     */
    getScissorTest(): boolean;

    /**
     * Enable the scissor test. When this is enabled, only the pixels within the defined scissor area will be affected by further renderer actions.
     */
    setScissorTest(enable: boolean): void;

    /**
     * Sets the custom opaque sort function for the WebGLRenderLists. Pass null to use the default painterSortStable function.
     */
    setOpaqueSort(method: (a: any, b: any) => number): void;

    /**
     * Sets the custom transparent sort function for the WebGLRenderLists. Pass null to use the default reversePainterSortStable function.
     */
    setTransparentSort(method: (a: any, b: any) => number): void;

    /**
     * Returns a THREE.Color instance with the current clear color.
     */
    getClearColor(target: Color): Color;

    /**
     * Sets the clear color, using color for the color and alpha for the opacity.
     */
    setClearColor(color: ColorRepresentation, alpha?: number): void;

    /**
     * Returns a float with the current clear alpha. Ranges from 0 to 1.
     */
    getClearAlpha(): number;

    setClearAlpha(alpha: number): void;

    /**
     * Tells the renderer to clear its color, depth or stencil drawing buffer(s).
     * Arguments default to true
     */
    clear(color?: boolean, depth?: boolean, stencil?: boolean): void;

    clearColor(): void;
    clearDepth(): void;
    clearStencil(): void;
    clearTarget(renderTarget: WebGLRenderTarget, color: boolean, depth: boolean, stencil: boolean): void;

    /**
     * @deprecated Use {@link WebGLState#reset .state.reset()} instead.
     */
    resetGLState(): void;
    dispose(): void;

    renderBufferDirect(
        camera: Camera,
        scene: Scene$1,
        geometry: BufferGeometry,
        material: Material,
        object: Object3D,
        geometryGroup: any,
    ): void;

    /**
     * A build in function that can be used instead of requestAnimationFrame. For WebXR projects this function must be used.
     * @param callback The function will be called every available frame. If `null` is passed it will stop any already ongoing animation.
     */
    setAnimationLoop(callback: XRFrameRequestCallback | null): void;

    /**
     * @deprecated Use {@link WebGLRenderer#setAnimationLoop .setAnimationLoop()} instead.
     */
    animate(callback: () => void): void;

    /**
     * Compiles all materials in the scene with the camera. This is useful to precompile shaders before the first rendering.
     */
    compile(scene: Object3D, camera: Camera): void;

    /**
     * Render a scene or an object using a camera.
     * The render is done to a previously specified {@link WebGLRenderTarget#renderTarget .renderTarget} set by calling
     * {@link WebGLRenderer#setRenderTarget .setRenderTarget} or to the canvas as usual.
     *
     * By default render buffers are cleared before rendering but you can prevent this by setting the property
     * {@link WebGLRenderer#autoClear autoClear} to false. If you want to prevent only certain buffers being cleared
     * you can set either the {@link WebGLRenderer#autoClearColor autoClearColor},
     * {@link WebGLRenderer#autoClearStencil autoClearStencil} or {@link WebGLRenderer#autoClearDepth autoClearDepth}
     * properties to false. To forcibly clear one ore more buffers call {@link WebGLRenderer#clear .clear}.
     */
    render(scene: Object3D, camera: Camera): void;

    /**
     * Returns the current active cube face.
     */
    getActiveCubeFace(): number;

    /**
     * Returns the current active mipmap level.
     */
    getActiveMipmapLevel(): number;

    /**
     * Returns the current render target. If no render target is set, null is returned.
     */
    getRenderTarget(): WebGLRenderTarget | null;

    /**
     * @deprecated Use {@link WebGLRenderer#getRenderTarget .getRenderTarget()} instead.
     */
    getCurrentRenderTarget(): WebGLRenderTarget | null;

    /**
     * Sets the active render target.
     *
     * @param renderTarget The {@link WebGLRenderTarget renderTarget} that needs to be activated. When `null` is given, the canvas is set as the active render target instead.
     * @param activeCubeFace Specifies the active cube side (PX 0, NX 1, PY 2, NY 3, PZ 4, NZ 5) of {@link WebGLCubeRenderTarget}.
     * @param activeMipmapLevel Specifies the active mipmap level.
     */
    setRenderTarget(
        renderTarget: WebGLRenderTarget | WebGLMultipleRenderTargets | null,
        activeCubeFace?: number,
        activeMipmapLevel?: number,
    ): void;

    readRenderTargetPixels(
        renderTarget: WebGLRenderTarget | WebGLMultipleRenderTargets,
        x: number,
        y: number,
        width: number,
        height: number,
        buffer: any,
        activeCubeFaceIndex?: number,
    ): void;

    /**
     * Copies a region of the currently bound framebuffer into the selected mipmap level of the selected texture.
     * This region is defined by the size of the destination texture's mip level, offset by the input position.
     *
     * @param position Specifies the pixel offset from which to copy out of the framebuffer.
     * @param texture Specifies the destination texture.
     * @param level Specifies the destination mipmap level of the texture.
     */
    copyFramebufferToTexture(position: Vector2, texture: Texture, level?: number): void;

    /**
     * Copies srcTexture to the specified level of dstTexture, offset by the input position.
     *
     * @param position Specifies the pixel offset into the dstTexture where the copy will occur.
     * @param srcTexture Specifies the source texture.
     * @param dstTexture Specifies the destination texture.
     * @param level Specifies the destination mipmap level of the texture.
     */
    copyTextureToTexture(position: Vector2, srcTexture: Texture, dstTexture: Texture, level?: number): void;

    /**
     * Copies the pixels of a texture in the bounds sourceBox in the desination texture starting from the given position.
     * @param sourceBox Specifies the bounds
     * @param position Specifies the pixel offset into the dstTexture where the copy will occur.
     * @param srcTexture Specifies the source texture.
     * @param dstTexture Specifies the destination texture.
     * @param level Specifies the destination mipmap level of the texture.
     */
    copyTextureToTexture3D(
        sourceBox: Box3,
        position: Vector3,
        srcTexture: Texture,
        dstTexture: Data3DTexture | DataArrayTexture,
        level?: number,
    ): void;

    /**
     * Initializes the given texture. Can be used to preload a texture rather than waiting until first render (which can cause noticeable lags due to decode and GPU upload overhead).
     *
     * @param texture The texture to Initialize.
     */
    initTexture(texture: Texture): void;

    /**
     * Can be used to reset the internal WebGL state.
     */
    resetState(): void;

    /**
     * @deprecated Use {@link WebGLRenderer#xr .xr} instead.
     */
    vr: boolean;

    /**
     * @deprecated Use {@link WebGLShadowMap#enabled .shadowMap.enabled} instead.
     */
    shadowMapEnabled: boolean;

    /**
     * @deprecated Use {@link WebGLShadowMap#type .shadowMap.type} instead.
     */
    shadowMapType: ShadowMapType;

    /**
     * @deprecated Use {@link WebGLShadowMap#cullFace .shadowMap.cullFace} instead.
     */
    shadowMapCullFace: CullFace;

    /**
     * @deprecated Use {@link WebGLExtensions#get .extensions.get( 'OES_texture_float' )} instead.
     */
    supportsFloatTextures(): any;

    /**
     * @deprecated Use {@link WebGLExtensions#get .extensions.get( 'OES_texture_half_float' )} instead.
     */
    supportsHalfFloatTextures(): any;

    /**
     * @deprecated Use {@link WebGLExtensions#get .extensions.get( 'OES_standard_derivatives' )} instead.
     */
    supportsStandardDerivatives(): any;

    /**
     * @deprecated Use {@link WebGLExtensions#get .extensions.get( 'WEBGL_compressed_texture_s3tc' )} instead.
     */
    supportsCompressedTextureS3TC(): any;

    /**
     * @deprecated Use {@link WebGLExtensions#get .extensions.get( 'WEBGL_compressed_texture_pvrtc' )} instead.
     */
    supportsCompressedTexturePVRTC(): any;

    /**
     * @deprecated Use {@link WebGLExtensions#get .extensions.get( 'EXT_blend_minmax' )} instead.
     */
    supportsBlendMinMax(): any;

    /**
     * @deprecated Use {@link WebGLCapabilities#vertexTextures .capabilities.vertexTextures} instead.
     */
    supportsVertexTextures(): any;

    /**
     * @deprecated Use {@link WebGLExtensions#get .extensions.get( 'ANGLE_instanced_arrays' )} instead.
     */
    supportsInstancedArrays(): any;

    /**
     * @deprecated Use {@link WebGLRenderer#setScissorTest .setScissorTest()} instead.
     */
    enableScissorTest(boolean: any): any;
}

declare class Ray {
    constructor(origin?: Vector3, direction?: Vector3);

    /**
     * @default new THREE.Vector3()
     */
    origin: Vector3;

    /**
     * @default new THREE.Vector3( 0, 0, - 1 )
     */
    direction: Vector3;

    set(origin: Vector3, direction: Vector3): Ray;
    clone(): this;
    copy(ray: Ray): this;
    at(t: number, target: Vector3): Vector3;
    lookAt(v: Vector3): Ray;
    recast(t: number): Ray;
    closestPointToPoint(point: Vector3, target: Vector3): Vector3;
    distanceToPoint(point: Vector3): number;
    distanceSqToPoint(point: Vector3): number;
    distanceSqToSegment(
        v0: Vector3,
        v1: Vector3,
        optionalPointOnRay?: Vector3,
        optionalPointOnSegment?: Vector3,
    ): number;
    intersectSphere(sphere: Sphere, target: Vector3): Vector3 | null;
    intersectsSphere(sphere: Sphere): boolean;
    distanceToPlane(plane: Plane): number;
    intersectPlane(plane: Plane, target: Vector3): Vector3 | null;
    intersectsPlane(plane: Plane): boolean;
    intersectBox(box: Box3, target: Vector3): Vector3 | null;
    intersectsBox(box: Box3): boolean;
    intersectTriangle(a: Vector3, b: Vector3, c: Vector3, backfaceCulling: boolean, target: Vector3): Vector3 | null;
    applyMatrix4(matrix4: Matrix4): Ray;
    equals(ray: Ray): boolean;

    /**
     * @deprecated Use {@link Ray#intersectsBox .intersectsBox()} instead.
     */
    isIntersectionBox(b: any): any;

    /**
     * @deprecated Use {@link Ray#intersectsPlane .intersectsPlane()} instead.
     */
    isIntersectionPlane(p: any): any;

    /**
     * @deprecated Use {@link Ray#intersectsSphere .intersectsSphere()} instead.
     */
    isIntersectionSphere(s: any): any;
}

interface Face {
    a: number;
    b: number;
    c: number;
    normal: Vector3;
    materialIndex: number;
}

interface Intersection<TIntersected extends Object3D = Object3D> {
    distance: number;
    distanceToRay?: number | undefined;
    point: Vector3;
    index?: number | undefined;
    face?: Face | null | undefined;
    faceIndex?: number | undefined;
    object: TIntersected;
    uv?: Vector2 | undefined;
    uv2?: Vector2 | undefined;
    instanceId?: number | undefined;
}

interface RaycasterParameters {
    Mesh?: any;
    Line?: { threshold: number } | undefined;
    LOD?: any;
    Points?: { threshold: number } | undefined;
    Sprite?: any;
}

declare class Raycaster {
    /**
     * This creates a new raycaster object.
     * @param origin The origin vector where the ray casts from.
     * @param direction The direction vector that gives direction to the ray. Should be normalized.
     * @param near All results returned are further away than near. Near can't be negative. Default value is 0.
     * @param far All results returned are closer then far. Far can't be lower then near . Default value is Infinity.
     */
    constructor(origin?: Vector3, direction?: Vector3, near?: number, far?: number);

    /** The Ray used for the raycasting. */
    ray: Ray;

    /**
     * The near factor of the raycaster. This value indicates which objects can be discarded based on the
     * distance. This value shouldn't be negative and should be smaller than the far property.
     * @default 0
     */
    near: number;

    /**
     * The far factor of the raycaster. This value indicates which objects can be discarded based on the
     * distance. This value shouldn't be negative and should be larger than the near property.
     * @default Infinity
     */
    far: number;

    /**
     * The camera to use when raycasting against view-dependent objects such as billboarded objects like Sprites. This field
     * can be set manually or is set when calling "setFromCamera".
     */
    camera: Camera;

    /**
     * Used by Raycaster to selectively ignore 3D objects when performing intersection tests.
     * @default new THREE.Layers()
     */
    layers: Layers;

    /**
     * @default { Mesh: {}, Line: { threshold: 1 }, LOD: {}, Points: { threshold: 1 }, Sprite: {} }
     */
    params: RaycasterParameters;

    /**
     * Updates the ray with a new origin and direction.
     * @param origin The origin vector where the ray casts from.
     * @param direction The normalized direction vector that gives direction to the ray.
     */
    set(origin: Vector3, direction: Vector3): void;

    /**
     * Updates the ray with a new origin and direction.
     * @param coords 2D coordinates of the mouse, in normalized device coordinates (NDC)---X and Y components should be between -1 and 1.
     * @param camera camera from which the ray should originate
     */
    setFromCamera(coords: { x: number; y: number }, camera: Camera): void;

    /**
     * Checks all intersection between the ray and the object with or without the descendants. Intersections are returned sorted by distance, closest first.
     * @param object The object to check for intersection with the ray.
     * @param recursive If true, it also checks all descendants. Otherwise it only checks intersecton with the object. Default is true.
     * @param optionalTarget (optional) target to set the result. Otherwise a new Array is instantiated. If set, you must clear this array prior to each call (i.e., array.length = 0;).
     */
    intersectObject<TIntersected extends Object3D>(
        object: Object3D,
        recursive?: boolean,
        optionalTarget?: Array<Intersection<TIntersected>>,
    ): Array<Intersection<TIntersected>>;

    /**
     * Checks all intersection between the ray and the objects with or without the descendants.
     * Intersections are returned sorted by distance, closest first.
     * Intersections are of the same form as those returned by .intersectObject.
     * @param objects The objects to check for intersection with the ray.
     * @param recursive If true, it also checks all descendants of the objects. Otherwise it only checks intersecton with the objects. Default is true.
     * @param optionalTarget (optional) target to set the result. Otherwise a new Array is instantiated. If set, you must clear this array prior to each call (i.e., array.length = 0;).
     */
    intersectObjects<TIntersected extends Object3D>(
        objects: Object3D[],
        recursive?: boolean,
        optionalTarget?: Array<Intersection<TIntersected>>,
    ): Array<Intersection<TIntersected>>;
}

declare abstract class Interpolant {
    constructor(parameterPositions: any, sampleValues: any, sampleSize: number, resultBuffer?: any);

    parameterPositions: any;
    sampleValues: any;
    valueSize: number;
    resultBuffer: any;

    evaluate(time: number): any;
}

declare class DiscreteInterpolant extends Interpolant {
    constructor(parameterPositions: any, samplesValues: any, sampleSize: number, resultBuffer?: any);

    interpolate_(i1: number, t0: number, t: number, t1: number): any;
}

declare class LinearInterpolant extends Interpolant {
    constructor(parameterPositions: any, samplesValues: any, sampleSize: number, resultBuffer?: any);

    interpolate_(i1: number, t0: number, t: number, t1: number): any;
}

declare class CubicInterpolant extends Interpolant {
    constructor(parameterPositions: any, samplesValues: any, sampleSize: number, resultBuffer?: any);

    interpolate_(i1: number, t0: number, t: number, t1: number): any;
}

declare class KeyframeTrack {
    /**
     * @param name
     * @param times
     * @param values
     * @param [interpolation=THREE.InterpolateLinear]
     */
    constructor(name: string, times: ArrayLike<any>, values: ArrayLike<any>, interpolation?: InterpolationModes);

    name: string;
    times: Float32Array;
    values: Float32Array;

    ValueTypeName: string;
    TimeBufferType: Float32Array;
    ValueBufferType: Float32Array;

    /**
     * @default THREE.InterpolateLinear
     */
    DefaultInterpolation: InterpolationModes;

    InterpolantFactoryMethodDiscrete(result: any): DiscreteInterpolant;
    InterpolantFactoryMethodLinear(result: any): LinearInterpolant;
    InterpolantFactoryMethodSmooth(result: any): CubicInterpolant;

    setInterpolation(interpolation: InterpolationModes): KeyframeTrack;
    getInterpolation(): InterpolationModes;

    getValueSize(): number;

    shift(timeOffset: number): KeyframeTrack;
    scale(timeScale: number): KeyframeTrack;
    trim(startTime: number, endTime: number): KeyframeTrack;
    validate(): boolean;
    optimize(): KeyframeTrack;
    clone(): this;

    static toJSON(track: KeyframeTrack): any;
}

// Objects //////////////////////////////////////////////////////////////////////////////////

declare class Bone extends Object3D {
    constructor();
    readonly isBone: true;
    type: 'Bone';
}

interface MorphTarget {
    name: string;
    vertices: Vector3[];
}

declare class AnimationClip {
    constructor(name?: string, duration?: number, tracks?: KeyframeTrack[], blendMode?: AnimationBlendMode);

    name: string;
    tracks: KeyframeTrack[];

    /**
     * @default THREE.NormalAnimationBlendMode
     */
    blendMode: AnimationBlendMode;

    /**
     * @default -1
     */
    duration: number;
    uuid: string;
    results: any[];

    resetDuration(): AnimationClip;
    trim(): AnimationClip;
    validate(): boolean;
    optimize(): AnimationClip;
    clone(): this;
    toJSON(clip: AnimationClip): any;

    static CreateFromMorphTargetSequence(
        name: string,
        morphTargetSequence: MorphTarget[],
        fps: number,
        noLoop: boolean,
    ): AnimationClip;
    static findByName(clipArray: AnimationClip[], name: string): AnimationClip;
    static CreateClipsFromMorphTargetSequences(
        morphTargets: MorphTarget[],
        fps: number,
        noLoop: boolean,
    ): AnimationClip[];
    static parse(json: any): AnimationClip;
    static parseAnimation(animation: any, bones: Bone[]): AnimationClip;
    static toJSON(clip: AnimationClip): any;
}

/**
 * Base class for scene graph objects
 */
declare class Object3D<E extends BaseEvent = Event$1> extends EventDispatcher<E> {
    constructor();

    /**
     * Unique number of this object instance.
     */
    id: number;

    uuid: string;

    /**
     * Optional name of the object (doesn't need to be unique).
     * @default ''
     */
    name: string;

    /**
     * @default 'Object3D'
     */
    type: string;

    /**
     * Object's parent in the scene graph.
     * @default null
     */
    parent: Object3D | null;

    /**
     * Array with object's children.
     * @default []
     */
    children: Object3D[];

    /**
     * Up direction.
     * @default THREE.Object3D.DefaultUp.clone()
     */
    up: Vector3;

    /**
     * Object's local position.
     * @default new THREE.Vector3()
     */
    readonly position: Vector3;

    /**
     * Object's local rotation (Euler angles), in radians.
     * @default new THREE.Euler()
     */
    readonly rotation: Euler;

    /**
     * Object's local rotation as a Quaternion.
     * @default new THREE.Quaternion()
     */
    readonly quaternion: Quaternion;

    /**
     * Object's local scale.
     * @default new THREE.Vector3()
     */
    readonly scale: Vector3;

    /**
     * @default new THREE.Matrix4()
     */
    readonly modelViewMatrix: Matrix4;

    /**
     * @default new THREE.Matrix3()
     */
    readonly normalMatrix: Matrix3;

    /**
     * Local transform.
     * @default new THREE.Matrix4()
     */
    matrix: Matrix4;

    /**
     * The global transform of the object. If the Object3d has no parent, then it's identical to the local transform.
     * @default new THREE.Matrix4()
     */
    matrixWorld: Matrix4;

    /**
     * When this is set, it calculates the matrix of position, (rotation or quaternion) and scale every frame and also
     * recalculates the matrixWorld property.
     * @default THREE.Object3D.DefaultMatrixAutoUpdate
     */
    matrixAutoUpdate: boolean;

    /**
     * When this is set, it calculates the matrixWorld in that frame and resets this property to false.
     * @default false
     */
    matrixWorldNeedsUpdate: boolean;

    /**
     * @default new THREE.Layers()
     */
    layers: Layers;

    /**
     * Object gets rendered if true.
     * @default true
     */
    visible: boolean;

    /**
     * Gets rendered into shadow map.
     * @default false
     */
    castShadow: boolean;

    /**
     * Material gets baked in shadow receiving.
     * @default false
     */
    receiveShadow: boolean;

    /**
     * When this is set, it checks every frame if the object is in the frustum of the camera before rendering the object.
     * If set to false the object gets rendered every frame even if it is not in the frustum of the camera.
     * @default true
     */
    frustumCulled: boolean;

    /**
     * Overrides the default rendering order of scene graph objects, from lowest to highest renderOrder.
     * Opaque and transparent objects remain sorted independently though.
     * When this property is set for an instance of Group, all descendants objects will be sorted and rendered together.
     * @default 0
     */
    renderOrder: number;

    /**
     * Array with animation clips.
     * @default []
     */
    animations: AnimationClip[];

    /**
     * An object that can be used to store custom data about the Object3d. It should not hold references to functions as these will not be cloned.
     * @default {}
     */
    userData: { [key: string]: any };

    /**
     * Custom depth material to be used when rendering to the depth map. Can only be used in context of meshes.
     * When shadow-casting with a DirectionalLight or SpotLight, if you are (a) modifying vertex positions in
     * the vertex shader, (b) using a displacement map, (c) using an alpha map with alphaTest, or (d) using a
     * transparent texture with alphaTest, you must specify a customDepthMaterial for proper shadows.
     */
    customDepthMaterial: Material;

    /**
     * Same as customDepthMaterial, but used with PointLight.
     */
    customDistanceMaterial: Material;

    /**
     * Used to check whether this or derived classes are Object3Ds. Default is true.
     * You should not change this, as it is used internally for optimisation.
     */
    readonly isObject3D: true;

    /**
     * Calls before rendering object
     */
    onBeforeRender: (
        renderer: WebGLRenderer,
        scene: Scene$1,
        camera: Camera,
        geometry: BufferGeometry,
        material: Material,
        group: Group,
    ) => void;

    /**
     * Calls after rendering object
     */
    onAfterRender: (
        renderer: WebGLRenderer,
        scene: Scene$1,
        camera: Camera,
        geometry: BufferGeometry,
        material: Material,
        group: Group,
    ) => void;

    static DefaultUp: Vector3;
    static DefaultMatrixAutoUpdate: boolean;

    /**
     * Applies the matrix transform to the object and updates the object's position, rotation and scale.
     */
    applyMatrix4(matrix: Matrix4): void;

    /**
     * Applies the rotation represented by the quaternion to the object.
     */
    applyQuaternion(quaternion: Quaternion): this;

    /**
     * axis -- A normalized vector in object space.
     * angle -- angle in radians
     * @param axis A normalized vector in object space.
     * @param angle angle in radians
     */
    setRotationFromAxisAngle(axis: Vector3, angle: number): void;

    /**
     * Calls setRotationFromEuler(euler) on the .quaternion.
     * @param euler Euler angle specifying rotation amount.
     */
    setRotationFromEuler(euler: Euler): void;

    /**
     * Calls setFromRotationMatrix(m) on the .quaternion.
     *
     * Note that this assumes that the upper 3x3 of m is a pure rotation matrix (i.e, unscaled).
     * @param m rotate the quaternion by the rotation component of the matrix.
     */
    setRotationFromMatrix(m: Matrix4): void;

    /**
     * Copy the given quaternion into .quaternion.
     * @param q normalized Quaternion
     */
    setRotationFromQuaternion(q: Quaternion): void;

    /**
     * Rotate an object along an axis in object space. The axis is assumed to be normalized.
     * @param axis	A normalized vector in object space.
     * @param angle	The angle in radians.
     */
    rotateOnAxis(axis: Vector3, angle: number): this;

    /**
     * Rotate an object along an axis in world space. The axis is assumed to be normalized. Method Assumes no rotated parent.
     * @param axis	A normalized vector in object space.
     * @param angle	The angle in radians.
     */
    rotateOnWorldAxis(axis: Vector3, angle: number): this;

    /**
     * Rotates the object around x axis in local space.
     * @param angle the angle to rotate in radians.
     */
    rotateX(angle: number): this;

    /**
     * Rotates the object around y axis in local space.
     * @param angle the angle to rotate in radians.
     */
    rotateY(angle: number): this;

    /**
     * Rotates the object around z axis in local space.
     * @param angle the angle to rotate in radians.
     */
    rotateZ(angle: number): this;

    /**
     * Translate an object by distance along an axis in object space. The axis is assumed to be normalized.
     * @param axis	A normalized vector in object space.
     * @param distance	The distance to translate.
     */
    translateOnAxis(axis: Vector3, distance: number): this;

    /**
     * Translates object along x axis by distance.
     * @param distance Distance.
     */
    translateX(distance: number): this;

    /**
     * Translates object along y axis by distance.
     * @param distance Distance.
     */
    translateY(distance: number): this;

    /**
     * Translates object along z axis by distance.
     * @param distance Distance.
     */
    translateZ(distance: number): this;

    /**
     * Updates the vector from local space to world space.
     * @param vector A local vector.
     */
    localToWorld(vector: Vector3): Vector3;

    /**
     * Updates the vector from world space to local space.
     * @param vector A world vector.
     */
    worldToLocal(vector: Vector3): Vector3;

    /**
     * Optionally, the x, y and z components of the world space position.
     * Rotates the object to face a point in world space.
     * This method does not support objects having non-uniformly-scaled parent(s).
     * @param vector A world vector to look at.
     */
    lookAt(vector: Vector3 | number, y?: number, z?: number): void;

    /**
     * Adds object as child of this object.
     */
    add(...object: Object3D[]): this;

    /**
     * Removes object as child of this object.
     */
    remove(...object: Object3D[]): this;

    /**
     * Removes this object from its current parent.
     */
    removeFromParent(): this;

    /**
     * Removes all child objects.
     */
    clear(): this;

    /**
     * Adds object as a child of this, while maintaining the object's world transform.
     */
    attach(object: Object3D): this;

    /**
     * Searches through the object's children and returns the first with a matching id.
     * @param id	Unique number of the object instance
     */
    getObjectById(id: number): Object3D | undefined;

    /**
     * Searches through the object's children and returns the first with a matching name.
     * @param name	String to match to the children's Object3d.name property.
     */
    getObjectByName(name: string): Object3D | undefined;

    getObjectByProperty(name: string, value: string): Object3D | undefined;

    getWorldPosition(target: Vector3): Vector3;
    getWorldQuaternion(target: Quaternion): Quaternion;
    getWorldScale(target: Vector3): Vector3;
    getWorldDirection(target: Vector3): Vector3;

    raycast(raycaster: Raycaster, intersects: Intersection[]): void;

    traverse(callback: (object: Object3D) => any): void;

    traverseVisible(callback: (object: Object3D) => any): void;

    traverseAncestors(callback: (object: Object3D) => any): void;

    /**
     * Updates local transform.
     */
    updateMatrix(): void;

    /**
     * Updates global transform of the object and its children.
     */
    updateMatrixWorld(force?: boolean): void;

    /**
     * Updates the global transform of the object.
     * @param updateParents recursively updates global transform of ancestors.
     * @param updateChildren recursively updates global transform of descendants.
     */
    updateWorldMatrix(updateParents: boolean, updateChildren: boolean): void;

    toJSON(meta?: { geometries: any; materials: any; textures: any; images: any }): any;

    clone(recursive?: boolean): this;

    /**
     *
     * @param object
     * @param recursive
     */
    copy(source: this, recursive?: boolean): this;
}

// Cameras ////////////////////////////////////////////////////////////////////////////////////////

/**
 * Abstract base class for cameras. This class should always be inherited when you build a new camera.
 */
declare class Camera extends Object3D {
    /**
     * This constructor sets following properties to the correct type: matrixWorldInverse, projectionMatrix and projectionMatrixInverse.
     */
    constructor();

    /**
     * This is the inverse of matrixWorld. MatrixWorld contains the Matrix which has the world transform of the Camera.
     * @default new THREE.Matrix4()
     */
    matrixWorldInverse: Matrix4;

    /**
     * This is the matrix which contains the projection.
     * @default new THREE.Matrix4()
     */
    projectionMatrix: Matrix4;

    /**
     * This is the inverse of projectionMatrix.
     * @default new THREE.Matrix4()
     */
    projectionMatrixInverse: Matrix4;

    readonly isCamera: true;

    getWorldDirection(target: Vector3): Vector3;

    updateMatrixWorld(force?: boolean): void;
}

declare class Spherical {
    constructor(radius?: number, phi?: number, theta?: number);

    /**
     * @default 1
     */
    radius: number;

    /**
     * @default 0
     */
    phi: number;

    /**
     * @default 0
     */
    theta: number;

    set(radius: number, phi: number, theta: number): this;
    clone(): this;
    copy(other: Spherical): this;
    makeSafe(): this;
    setFromVector3(v: Vector3): this;
    setFromCartesianCoords(x: number, y: number, z: number): this;
}

declare class Cylindrical {
    constructor(radius?: number, theta?: number, y?: number);

    /**
     * @default 1
     */
    radius: number;

    /**
     * @default 0
     */
    theta: number;

    /**
     * @default 0
     */
    y: number;

    clone(): this;
    copy(other: Cylindrical): this;
    set(radius: number, theta: number, y: number): this;
    setFromVector3(vec3: Vector3): this;
    setFromCartesianCoords(x: number, y: number, z: number): this;
}

type Vector3Tuple = [number, number, number];

/**
 * 3D vector. ( class Vector3 implements Vector<Vector3> )
 *
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/math/Vector3.js}
 *
 * @example
 * const a = new THREE.Vector3( 1, 0, 0 );
 * const b = new THREE.Vector3( 0, 1, 0 );
 * const c = new THREE.Vector3();
 * c.crossVectors( a, b );
 */
declare class Vector3 implements Vector {
    constructor(x?: number, y?: number, z?: number);

    /**
     * @default 0
     */
    x: number;

    /**
     * @default 0
     */
    y: number;

    /**
     * @default 0
     */
    z: number;
    readonly isVector3: true;

    /**
     * Sets value of this vector.
     */
    set(x: number, y: number, z: number): this;

    /**
     * Sets all values of this vector.
     */
    setScalar(scalar: number): this;

    /**
     * Sets x value of this vector.
     */
    setX(x: number): Vector3;

    /**
     * Sets y value of this vector.
     */
    setY(y: number): Vector3;

    /**
     * Sets z value of this vector.
     */
    setZ(z: number): Vector3;

    setComponent(index: number, value: number): this;

    getComponent(index: number): number;

    /**
     * Clones this vector.
     */
    clone(): this;

    /**
     * Copies value of v to this vector.
     */
    copy(v: Vector3): this;

    /**
     * Adds v to this vector.
     */
    add(v: Vector3): this;

    addScalar(s: number): this;

    addScaledVector(v: Vector3, s: number): this;

    /**
     * Sets this vector to a + b.
     */
    addVectors(a: Vector3, b: Vector3): this;

    /**
     * Subtracts v from this vector.
     */
    sub(a: Vector3): this;

    subScalar(s: number): this;

    /**
     * Sets this vector to a - b.
     */
    subVectors(a: Vector3, b: Vector3): this;

    multiply(v: Vector3): this;

    /**
     * Multiplies this vector by scalar s.
     */
    multiplyScalar(s: number): this;

    multiplyVectors(a: Vector3, b: Vector3): this;

    applyEuler(euler: Euler): this;

    applyAxisAngle(axis: Vector3, angle: number): this;

    applyMatrix3(m: Matrix3): this;

    applyNormalMatrix(m: Matrix3): this;

    applyMatrix4(m: Matrix4): this;

    applyQuaternion(q: Quaternion): this;

    project(camera: Camera): this;

    unproject(camera: Camera): this;

    transformDirection(m: Matrix4): this;

    divide(v: Vector3): this;

    /**
     * Divides this vector by scalar s.
     * Set vector to ( 0, 0, 0 ) if s == 0.
     */
    divideScalar(s: number): this;

    min(v: Vector3): this;

    max(v: Vector3): this;

    clamp(min: Vector3, max: Vector3): this;

    clampScalar(min: number, max: number): this;

    clampLength(min: number, max: number): this;

    floor(): this;

    ceil(): this;

    round(): this;

    roundToZero(): this;

    /**
     * Inverts this vector.
     */
    negate(): this;

    /**
     * Computes dot product of this vector and v.
     */
    dot(v: Vector3): number;

    /**
     * Computes squared length of this vector.
     */
    lengthSq(): number;

    /**
     * Computes length of this vector.
     */
    length(): number;

    /**
     * Computes Manhattan length of this vector.
     * http://en.wikipedia.org/wiki/Taxicab_geometry
     *
     * @deprecated Use {@link Vector3#manhattanLength .manhattanLength()} instead.
     */
    lengthManhattan(): number;

    /**
     * Computes the Manhattan length of this vector.
     *
     * see {@link http://en.wikipedia.org/wiki/Taxicab_geometry|Wikipedia: Taxicab Geometry}
     */
    manhattanLength(): number;

    /**
     * Computes the Manhattan length (distance) from this vector to the given vector v
     *
     * see {@link http://en.wikipedia.org/wiki/Taxicab_geometry|Wikipedia: Taxicab Geometry}
     */
    manhattanDistanceTo(v: Vector3): number;

    /**
     * Normalizes this vector.
     */
    normalize(): this;

    /**
     * Normalizes this vector and multiplies it by l.
     */
    setLength(l: number): this;
    lerp(v: Vector3, alpha: number): this;

    lerpVectors(v1: Vector3, v2: Vector3, alpha: number): this;

    /**
     * Sets this vector to cross product of itself and v.
     */
    cross(a: Vector3): this;

    /**
     * Sets this vector to cross product of a and b.
     */
    crossVectors(a: Vector3, b: Vector3): this;
    projectOnVector(v: Vector3): this;
    projectOnPlane(planeNormal: Vector3): this;
    reflect(vector: Vector3): this;
    angleTo(v: Vector3): number;

    /**
     * Computes distance of this vector to v.
     */
    distanceTo(v: Vector3): number;

    /**
     * Computes squared distance of this vector to v.
     */
    distanceToSquared(v: Vector3): number;

    /**
     * @deprecated Use {@link Vector3#manhattanDistanceTo .manhattanDistanceTo()} instead.
     */
    distanceToManhattan(v: Vector3): number;

    setFromSpherical(s: Spherical): this;
    setFromSphericalCoords(r: number, phi: number, theta: number): this;
    setFromCylindrical(s: Cylindrical): this;
    setFromCylindricalCoords(radius: number, theta: number, y: number): this;
    setFromMatrixPosition(m: Matrix4): this;
    setFromMatrixScale(m: Matrix4): this;
    setFromMatrixColumn(matrix: Matrix4, index: number): this;
    setFromMatrix3Column(matrix: Matrix3, index: number): this;

    /**
     * Sets this vector's {@link x}, {@link y} and {@link z} components from the x, y, and z components of the specified {@link Euler Euler Angle}.
     */
    setFromEuler(e: Euler): this;

    /**
     * Checks for strict equality of this vector and v.
     */
    equals(v: Vector3): boolean;

    /**
     * Sets this vector's x, y and z value from the provided array or array-like.
     * @param array the source array or array-like.
     * @param offset (optional) offset into the array. Default is 0.
     */
    fromArray(array: number[] | ArrayLike<number>, offset?: number): this;

    /**
     * Returns an array [x, y, z], or copies x, y and z into the provided array.
     * @param array (optional) array to store the vector to. If this is not provided, a new array will be created.
     * @param offset (optional) optional offset into the array.
     * @return The created or provided array.
     */
    toArray(array?: number[], offset?: number): number[];
    toArray(array?: Vector3Tuple, offset?: 0): Vector3Tuple;

    /**
     * Copies x, y and z into the provided array-like.
     * @param array array-like to store the vector to.
     * @param offset (optional) optional offset into the array-like.
     * @return The provided array-like.
     */
    toArray(array: ArrayLike<number>, offset?: number): ArrayLike<number>;

    fromBufferAttribute(attribute: BufferAttribute | InterleavedBufferAttribute, index: number): this;

    /**
     * Sets this vector's x, y and z from Math.random
     */
    random(): this;

    randomDirection(): this;
}

type Matrix4Tuple = [
    number,
    number,
    number,
    number,
    number,
    number,
    number,
    number,
    number,
    number,
    number,
    number,
    number,
    number,
    number,
    number,
];

/**
 * A 4x4 Matrix.
 *
 * @example
 * // Simple rig for rotating around 3 axes
 * const m = new THREE.Matrix4();
 * const m1 = new THREE.Matrix4();
 * const m2 = new THREE.Matrix4();
 * const m3 = new THREE.Matrix4();
 * const alpha = 0;
 * const beta = Math.PI;
 * const gamma = Math.PI/2;
 * m1.makeRotationX( alpha );
 * m2.makeRotationY( beta );
 * m3.makeRotationZ( gamma );
 * m.multiplyMatrices( m1, m2 );
 * m.multiply( m3 );
 */
declare class Matrix4 implements Matrix {
    constructor();

    /**
     * Array with matrix values.
     * @default [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1]
     */
    elements: number[];

    /**
     * Sets all fields of this matrix.
     */
    set(
        n11: number,
        n12: number,
        n13: number,
        n14: number,
        n21: number,
        n22: number,
        n23: number,
        n24: number,
        n31: number,
        n32: number,
        n33: number,
        n34: number,
        n41: number,
        n42: number,
        n43: number,
        n44: number,
    ): Matrix4;

    /**
     * Resets this matrix to identity.
     */
    identity(): Matrix4;
    clone(): Matrix4;
    copy(m: Matrix4): this;
    copyPosition(m: Matrix4): Matrix4;
    extractBasis(xAxis: Vector3, yAxis: Vector3, zAxis: Vector3): Matrix4;
    makeBasis(xAxis: Vector3, yAxis: Vector3, zAxis: Vector3): Matrix4;

    /**
     * Copies the rotation component of the supplied matrix m into this matrix rotation component.
     */
    extractRotation(m: Matrix4): Matrix4;
    makeRotationFromEuler(euler: Euler): Matrix4;
    makeRotationFromQuaternion(q: Quaternion): Matrix4;
    /**
     * Constructs a rotation matrix, looking from eye towards center with defined up vector.
     */
    lookAt(eye: Vector3, target: Vector3, up: Vector3): Matrix4;

    /**
     * Multiplies this matrix by m.
     */
    multiply(m: Matrix4): Matrix4;

    premultiply(m: Matrix4): Matrix4;

    /**
     * Sets this matrix to a x b.
     */
    multiplyMatrices(a: Matrix4, b: Matrix4): Matrix4;

    /**
     * Sets this matrix to a x b and stores the result into the flat array r.
     * r can be either a regular Array or a TypedArray.
     *
     * @deprecated This method has been removed completely.
     */
    multiplyToArray(a: Matrix4, b: Matrix4, r: number[]): Matrix4;

    /**
     * Multiplies this matrix by s.
     */
    multiplyScalar(s: number): Matrix4;

    /**
     * Computes determinant of this matrix.
     * Based on http://www.euclideanspace.com/maths/algebra/matrix/functions/inverse/fourD/index.htm
     */
    determinant(): number;

    /**
     * Transposes this matrix.
     */
    transpose(): Matrix4;

    /**
     * Sets the position component for this matrix from vector v.
     */
    setPosition(v: Vector3 | number, y?: number, z?: number): Matrix4;

    /**
     * Inverts this matrix.
     */
    invert(): Matrix4;

    /**
     * Multiplies the columns of this matrix by vector v.
     */
    scale(v: Vector3): Matrix4;

    getMaxScaleOnAxis(): number;
    /**
     * Sets this matrix as translation transform.
     */
    makeTranslation(x: number, y: number, z: number): Matrix4;

    /**
     * Sets this matrix as rotation transform around x axis by theta radians.
     *
     * @param theta Rotation angle in radians.
     */
    makeRotationX(theta: number): Matrix4;

    /**
     * Sets this matrix as rotation transform around y axis by theta radians.
     *
     * @param theta Rotation angle in radians.
     */
    makeRotationY(theta: number): Matrix4;

    /**
     * Sets this matrix as rotation transform around z axis by theta radians.
     *
     * @param theta Rotation angle in radians.
     */
    makeRotationZ(theta: number): Matrix4;

    /**
     * Sets this matrix as rotation transform around axis by angle radians.
     * Based on http://www.gamedev.net/reference/articles/article1199.asp.
     *
     * @param axis Rotation axis.
     * @param theta Rotation angle in radians.
     */
    makeRotationAxis(axis: Vector3, angle: number): Matrix4;

    /**
     * Sets this matrix as scale transform.
     */
    makeScale(x: number, y: number, z: number): Matrix4;

    /**
     * Sets this matrix as shear transform.
     */
    makeShear(xy: number, xz: number, yx: number, yz: number, zx: number, zy: number): Matrix4;

    /**
     * Sets this matrix to the transformation composed of translation, rotation and scale.
     */
    compose(translation: Vector3, rotation: Quaternion, scale: Vector3): Matrix4;

    /**
     * Decomposes this matrix into it's position, quaternion and scale components.
     */
    decompose(translation: Vector3, rotation: Quaternion, scale: Vector3): Matrix4;

    /**
     * Creates a frustum matrix.
     */
    makePerspective(left: number, right: number, bottom: number, top: number, near: number, far: number): Matrix4;

    /**
     * Creates a perspective projection matrix.
     */
    makePerspective(fov: number, aspect: number, near: number, far: number): Matrix4;

    /**
     * Creates an orthographic projection matrix.
     */
    makeOrthographic(left: number, right: number, top: number, bottom: number, near: number, far: number): Matrix4;
    equals(matrix: Matrix4): boolean;

    /**
     * Sets the values of this matrix from the provided array or array-like.
     * @param array the source array or array-like.
     * @param offset (optional) offset into the array-like. Default is 0.
     */
    fromArray(array: number[] | ArrayLike<number>, offset?: number): Matrix4;

    /**
     * Returns an array with the values of this matrix, or copies them into the provided array.
     * @param array (optional) array to store the matrix to. If this is not provided, a new array will be created.
     * @param offset (optional) optional offset into the array.
     * @return The created or provided array.
     */
    toArray(array?: number[], offset?: number): number[];
    toArray(array?: Matrix4Tuple, offset?: 0): Matrix4Tuple;

    /**
     * Copies he values of this matrix into the provided array-like.
     * @param array array-like to store the matrix to.
     * @param offset (optional) optional offset into the array-like.
     * @return The provided array-like.
     */
    toArray(array?: ArrayLike<number>, offset?: number): ArrayLike<number>;

    /**
     * Set the upper 3x3 elements of this matrix to the values of the Matrix3 m.
     */
    setFromMatrix3(m: Matrix3): Matrix4;

    /**
     * @deprecated Use {@link Matrix4#copyPosition .copyPosition()} instead.
     */
    extractPosition(m: Matrix4): Matrix4;

    /**
     * @deprecated Use {@link Matrix4#makeRotationFromQuaternion .makeRotationFromQuaternion()} instead.
     */
    setRotationFromQuaternion(q: Quaternion): Matrix4;

    /**
     * @deprecated Use {@link Vector3#applyMatrix4 vector.applyMatrix4( matrix )} instead.
     */
    multiplyVector3(v: any): any;

    /**
     * @deprecated Use {@link Vector4#applyMatrix4 vector.applyMatrix4( matrix )} instead.
     */
    multiplyVector4(v: any): any;

    /**
     * @deprecated This method has been removed completely.
     */
    multiplyVector3Array(array: number[]): number[];

    /**
     * @deprecated Use {@link Vector3#transformDirection Vector3.transformDirection( matrix )} instead.
     */
    rotateAxis(v: any): void;

    /**
     * @deprecated Use {@link Vector3#applyMatrix4 vector.applyMatrix4( matrix )} instead.
     */
    crossVector(v: any): void;

    /**
     * @deprecated Use {@link Matrix4#toArray .toArray()} instead.
     */
    flattenToArrayOffset(array: number[], offset: number): number[];

    /**
     * @deprecated Use {@link Matrix4#invert .invert()} instead.
     */
    getInverse(matrix: Matrix): Matrix;
}

type Matrix3Tuple = [number, number, number, number, number, number, number, number, number];

/**
 * ( interface Matrix<T> )
 */
interface Matrix {
    /**
     * Array with matrix values.
     */
    elements: number[];

    /**
     * identity():T;
     */
    identity(): Matrix;

    /**
     * copy(m:T):T;
     */
    copy(m: this): this;

    /**
     * multiplyScalar(s:number):T;
     */
    multiplyScalar(s: number): Matrix;

    determinant(): number;

    /**
     * transpose():T;
     */
    transpose(): Matrix;

    /**
     * invert():T;
     */
    invert(): Matrix;

    /**
     * clone():T;
     */
    clone(): Matrix;
}

/**
 * ( class Matrix3 implements Matrix<Matrix3> )
 */
declare class Matrix3 implements Matrix {
    /**
     * Creates an identity matrix.
     */
    constructor();

    /**
     * Array with matrix values.
     * @default [1, 0, 0, 0, 1, 0, 0, 0, 1]
     */
    elements: number[];

    set(
        n11: number,
        n12: number,
        n13: number,
        n21: number,
        n22: number,
        n23: number,
        n31: number,
        n32: number,
        n33: number,
    ): Matrix3;
    identity(): Matrix3;
    clone(): this;
    copy(m: Matrix3): this;
    extractBasis(xAxis: Vector3, yAxis: Vector3, zAxis: Vector3): Matrix3;
    setFromMatrix4(m: Matrix4): Matrix3;
    multiplyScalar(s: number): Matrix3;
    determinant(): number;

    /**
     * Inverts this matrix in place.
     */
    invert(): Matrix3;

    /**
     * Transposes this matrix in place.
     */
    transpose(): Matrix3;
    getNormalMatrix(matrix4: Matrix4): Matrix3;

    /**
     * Transposes this matrix into the supplied array r, and returns itself.
     */
    transposeIntoArray(r: number[]): Matrix3;

    setUvTransform(tx: number, ty: number, sx: number, sy: number, rotation: number, cx: number, cy: number): Matrix3;

    scale(sx: number, sy: number): Matrix3;

    rotate(theta: number): Matrix3;

    translate(tx: number, ty: number): Matrix3;

    equals(matrix: Matrix3): boolean;

    /**
     * Sets the values of this matrix from the provided array or array-like.
     * @param array the source array or array-like.
     * @param offset (optional) offset into the array-like. Default is 0.
     */
    fromArray(array: number[] | ArrayLike<number>, offset?: number): Matrix3;

    /**
     * Returns an array with the values of this matrix, or copies them into the provided array.
     * @param array (optional) array to store the matrix to. If this is not provided, a new array will be created.
     * @param offset (optional) optional offset into the array.
     * @return The created or provided array.
     */
    toArray(array?: number[], offset?: number): number[];
    toArray(array?: Matrix3Tuple, offset?: 0): Matrix3Tuple;

    /**
     * Copies he values of this matrix into the provided array-like.
     * @param array array-like to store the matrix to.
     * @param offset (optional) optional offset into the array-like.
     * @return The provided array-like.
     */
    toArray(array?: ArrayLike<number>, offset?: number): ArrayLike<number>;

    /**
     * Multiplies this matrix by m.
     */
    multiply(m: Matrix3): Matrix3;

    premultiply(m: Matrix3): Matrix3;

    /**
     * Sets this matrix to a x b.
     */
    multiplyMatrices(a: Matrix3, b: Matrix3): Matrix3;

    /**
     * @deprecated Use {@link Vector3.applyMatrix3 vector.applyMatrix3( matrix )} instead.
     */
    multiplyVector3(vector: Vector3): any;

    /**
     * @deprecated This method has been removed completely.
     */
    multiplyVector3Array(a: any): any;

    /**
     * @deprecated Use {@link Matrix3#invert .invert()} instead.
     */
    getInverse(matrix: Matrix4, throwOnDegenerate?: boolean): Matrix3;
    getInverse(matrix: Matrix): Matrix;

    /**
     * @deprecated Use {@link Matrix3#toArray .toArray()} instead.
     */
    flattenToArrayOffset(array: number[], offset: number): number[];
}

/**
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/core/BufferAttribute.js|src/core/BufferAttribute.js}
 */
declare class BufferAttribute {
    constructor(array: ArrayLike<number>, itemSize: number, normalized?: boolean); // array parameter should be TypedArray.

    /**
     * @default ''
     */
    name: string;
    array: ArrayLike<number>;
    itemSize: number;

    /**
     * @default THREE.StaticDrawUsage
     */
    usage: Usage;

    /**
     * @default { offset: number; count: number }
     */
    updateRange: { offset: number; count: number };

    /**
     * @default 0
     */
    version: number;

    /**
     * @default false
     */
    normalized: boolean;

    /**
     * @default 0
     */
    count: number;

    set needsUpdate(value: boolean);

    readonly isBufferAttribute: true;

    onUploadCallback: () => void;
    onUpload(callback: () => void): this;
    setUsage(usage: Usage): this;
    clone(): this;
    copy(source: BufferAttribute): this;
    copyAt(index1: number, attribute: BufferAttribute, index2: number): this;
    copyArray(array: ArrayLike<number>): this;
    copyColorsArray(colors: Array<{ r: number; g: number; b: number }>): this;
    copyVector2sArray(vectors: Array<{ x: number; y: number }>): this;
    copyVector3sArray(vectors: Array<{ x: number; y: number; z: number }>): this;
    copyVector4sArray(vectors: Array<{ x: number; y: number; z: number; w: number }>): this;
    applyMatrix3(m: Matrix3): this;
    applyMatrix4(m: Matrix4): this;
    applyNormalMatrix(m: Matrix3): this;
    transformDirection(m: Matrix4): this;
    set(value: ArrayLike<number> | ArrayBufferView, offset?: number): this;
    getX(index: number): number;
    setX(index: number, x: number): this;
    getY(index: number): number;
    setY(index: number, y: number): this;
    getZ(index: number): number;
    setZ(index: number, z: number): this;
    getW(index: number): number;
    setW(index: number, z: number): this;
    setXY(index: number, x: number, y: number): this;
    setXYZ(index: number, x: number, y: number, z: number): this;
    setXYZW(index: number, x: number, y: number, z: number, w: number): this;
    toJSON(): {
        itemSize: number;
        type: string;
        array: number[];
        normalized: boolean;
    };
}

/**
 * @deprecated THREE.Int8Attribute has been removed. Use new THREE.Int8BufferAttribute() instead.
 */
declare class Int8Attribute extends BufferAttribute {
    constructor(array: any, itemSize: number);
}

/**
 * @deprecated THREE.Uint8Attribute has been removed. Use new THREE.Uint8BufferAttribute() instead.
 */
declare class Uint8Attribute extends BufferAttribute {
    constructor(array: any, itemSize: number);
}

/**
 * @deprecated THREE.Uint8ClampedAttribute has been removed. Use new THREE.Uint8ClampedBufferAttribute() instead.
 */
declare class Uint8ClampedAttribute extends BufferAttribute {
    constructor(array: any, itemSize: number);
}

/**
 * @deprecated THREE.Int16Attribute has been removed. Use new THREE.Int16BufferAttribute() instead.
 */
declare class Int16Attribute extends BufferAttribute {
    constructor(array: any, itemSize: number);
}

/**
 * @deprecated THREE.Uint16Attribute has been removed. Use new THREE.Uint16BufferAttribute() instead.
 */
declare class Uint16Attribute extends BufferAttribute {
    constructor(array: any, itemSize: number);
}

/**
 * @deprecated THREE.Int32Attribute has been removed. Use new THREE.Int32BufferAttribute() instead.
 */
declare class Int32Attribute extends BufferAttribute {
    constructor(array: any, itemSize: number);
}

/**
 * @deprecated THREE.Uint32Attribute has been removed. Use new THREE.Uint32BufferAttribute() instead.
 */
declare class Uint32Attribute extends BufferAttribute {
    constructor(array: any, itemSize: number);
}

/**
 * @deprecated THREE.Float32Attribute has been removed. Use new THREE.Float32BufferAttribute() instead.
 */
declare class Float32Attribute extends BufferAttribute {
    constructor(array: any, itemSize: number);
}

/**
 * @deprecated THREE.Float64Attribute has been removed. Use new THREE.Float64BufferAttribute() instead.
 */
declare class Float64Attribute extends BufferAttribute {
    constructor(array: any, itemSize: number);
}

declare class Int8BufferAttribute extends BufferAttribute {
    constructor(
        array: Iterable<number> | ArrayLike<number> | ArrayBuffer | number,
        itemSize: number,
        normalized?: boolean,
    );
}

declare class Uint8BufferAttribute extends BufferAttribute {
    constructor(
        array: Iterable<number> | ArrayLike<number> | ArrayBuffer | number,
        itemSize: number,
        normalized?: boolean,
    );
}

declare class Uint8ClampedBufferAttribute extends BufferAttribute {
    constructor(
        array: Iterable<number> | ArrayLike<number> | ArrayBuffer | number,
        itemSize: number,
        normalized?: boolean,
    );
}

declare class Int16BufferAttribute extends BufferAttribute {
    constructor(
        array: Iterable<number> | ArrayLike<number> | ArrayBuffer | number,
        itemSize: number,
        normalized?: boolean,
    );
}

declare class Uint16BufferAttribute extends BufferAttribute {
    constructor(
        array: Iterable<number> | ArrayLike<number> | ArrayBuffer | number,
        itemSize: number,
        normalized?: boolean,
    );
}

declare class Int32BufferAttribute extends BufferAttribute {
    constructor(
        array: Iterable<number> | ArrayLike<number> | ArrayBuffer | number,
        itemSize: number,
        normalized?: boolean,
    );
}

declare class Uint32BufferAttribute extends BufferAttribute {
    constructor(
        array: Iterable<number> | ArrayLike<number> | ArrayBuffer | number,
        itemSize: number,
        normalized?: boolean,
    );
}

declare class Float16BufferAttribute extends BufferAttribute {
    constructor(
        array: Iterable<number> | ArrayLike<number> | ArrayBuffer | number,
        itemSize: number,
        normalized?: boolean,
    );
}

declare class Float32BufferAttribute extends BufferAttribute {
    constructor(
        array: Iterable<number> | ArrayLike<number> | ArrayBuffer | number,
        itemSize: number,
        normalized?: boolean,
    );
}

declare class Float64BufferAttribute extends BufferAttribute {
    constructor(
        array: Iterable<number> | ArrayLike<number> | ArrayBuffer | number,
        itemSize: number,
        normalized?: boolean,
    );
}

declare function SRGBToLinear(c: number): number;

interface HSL {
    h: number;
    s: number;
    l: number;
}

interface RGB {
    r: number;
    g: number;
    b: number;
}

/**
 * Represents a color. See also {@link ColorUtils}.
 *
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/math/Color.js|src/math/Color.js}
 *
 * @example
 * const color = new THREE.Color( 0xff0000 );
 */
declare class Color {
    constructor(color?: ColorRepresentation);
    constructor(r: number, g: number, b: number);

    readonly isColor: true;

    /**
     * Red channel value between 0 and 1. Default is 1.
     * @default 1
     */
    r: number;

    /**
     * Green channel value between 0 and 1. Default is 1.
     * @default 1
     */
    g: number;

    /**
     * Blue channel value between 0 and 1. Default is 1.
     * @default 1
     */
    b: number;

    set(color: ColorRepresentation): Color;
    setScalar(scalar: number): Color;
    setHex(hex: number, colorSpace?: ColorSpace): Color;

    /**
     * Sets this color from RGB values.
     * @param r Red channel value between 0 and 1.
     * @param g Green channel value between 0 and 1.
     * @param b Blue channel value between 0 and 1.
     */
    setRGB(r: number, g: number, b: number, colorSpace?: ColorSpace): Color;

    /**
     * Sets this color from HSL values.
     * Based on MochiKit implementation by Bob Ippolito.
     *
     * @param h Hue channel value between 0 and 1.
     * @param s Saturation value channel between 0 and 1.
     * @param l Value channel value between 0 and 1.
     */
    setHSL(h: number, s: number, l: number, colorSpace?: ColorSpace): Color;

    /**
     * Sets this color from a CSS context style string.
     * @param contextStyle Color in CSS context style format.
     */
    setStyle(style: string, colorSpace?: ColorSpace): Color;

    /**
     * Sets this color from a color name.
     * Faster than {@link Color#setStyle .setStyle()} method if you don't need the other CSS-style formats.
     * @param style Color name in X11 format.
     */
    setColorName(style: string, colorSpace?: ColorSpace): Color;

    /**
     * Clones this color.
     */
    clone(): this;

    /**
     * Copies given color.
     * @param color Color to copy.
     */
    copy(color: Color): this;

    /**
     * Copies given color making conversion from sRGB to linear space.
     * @param color Color to copy.
     */
    copySRGBToLinear(color: Color): Color;

    /**
     * Copies given color making conversion from linear to sRGB space.
     * @param color Color to copy.
     */
    copyLinearToSRGB(color: Color): Color;

    /**
     * Converts this color from sRGB to linear space.
     */
    convertSRGBToLinear(): Color;

    /**
     * Converts this color from linear to sRGB space.
     */
    convertLinearToSRGB(): Color;

    /**
     * Returns the hexadecimal value of this color.
     */
    getHex(colorSpace?: ColorSpace): number;

    /**
     * Returns the string formated hexadecimal value of this color.
     */
    getHexString(colorSpace?: ColorSpace): string;

    getHSL(target: HSL, colorSpace?: ColorSpace): HSL;

    getRGB(target: RGB, colorSpace?: ColorSpace): RGB;

    /**
     * Returns the value of this color in CSS context style.
     * Example: rgb(r, g, b)
     */
    getStyle(colorSpace?: ColorSpace): string;

    offsetHSL(h: number, s: number, l: number): this;

    add(color: Color): this;
    addColors(color1: Color, color2: Color): this;
    addScalar(s: number): this;
    sub(color: Color): this;
    multiply(color: Color): this;
    multiplyScalar(s: number): this;
    lerp(color: Color, alpha: number): this;
    lerpColors(color1: Color, color2: Color, alpha: number): this;
    lerpHSL(color: Color, alpha: number): this;
    equals(color: Color): boolean;

    /**
     * Sets this color's red, green and blue value from the provided array or array-like.
     * @param array the source array or array-like.
     * @param offset (optional) offset into the array-like. Default is 0.
     */
    fromArray(array: number[] | ArrayLike<number>, offset?: number): this;

    /**
     * Returns an array [red, green, blue], or copies red, green and blue into the provided array.
     * @param array (optional) array to store the color to. If this is not provided, a new array will be created.
     * @param offset (optional) optional offset into the array.
     * @return The created or provided array.
     */
    toArray(array?: number[], offset?: number): number[];

    /**
     * Copies red, green and blue into the provided array-like.
     * @param array array-like to store the color to.
     * @param offset (optional) optional offset into the array-like.
     * @return The provided array-like.
     */
    toArray(xyz: ArrayLike<number>, offset?: number): ArrayLike<number>;

    fromBufferAttribute(attribute: BufferAttribute, index: number): this;

    [Symbol.iterator](): Generator<number, void>;

    /**
     * List of X11 color names.
     */
    static NAMES: Record<string, number>;
}

type ColorRepresentation = Color | string | number;

declare class VectorKeyframeTrack extends KeyframeTrack {
    constructor(name: string, times: any[], values: any[], interpolation?: InterpolationModes);

    /**
     * @default 'vector'
     */
    ValueTypeName: string;
}

declare class StringKeyframeTrack extends KeyframeTrack {
    constructor(name: string, times: any[], values: any[], interpolation?: InterpolationModes);

    /**
     * @default 'string'
     */
    ValueTypeName: string;
}

declare class QuaternionKeyframeTrack extends KeyframeTrack {
    constructor(name: string, times: any[], values: any[], interpolation?: InterpolationModes);

    /**
     * @default 'quaternion'
     */
    ValueTypeName: string;
}

declare class NumberKeyframeTrack extends KeyframeTrack {
    constructor(name: string, times: any[], values: any[], interpolation?: InterpolationModes);

    /**
     * @default 'number'
     */
    ValueTypeName: string;
}

declare class ColorKeyframeTrack extends KeyframeTrack {
    constructor(name: string, times: any[], values: any[], interpolation?: InterpolationModes);

    /**
     * @default 'color'
     */
    ValueTypeName: string;
}

declare class BooleanKeyframeTrack extends KeyframeTrack {
    constructor(name: string, times: any[], values: any[]);

    /**
     * @default 'bool'
     */
    ValueTypeName: string;
}

declare class PropertyMixer {
    constructor(binding: any, typeName: string, valueSize: number);

    binding: any;
    valueSize: number;
    buffer: any;
    cumulativeWeight: number;
    cumulativeWeightAdditive: number;
    useCount: number;
    referenceCount: number;

    accumulate(accuIndex: number, weight: number): void;
    accumulateAdditive(weight: number): void;
    apply(accuIndex: number): void;
    saveOriginalState(): void;
    restoreOriginalState(): void;
}

interface ParseTrackNameResults {
    nodeName: string;
    objectName: string;
    objectIndex: string;
    propertyName: string;
    propertyIndex: string;
}

declare class PropertyBinding {
    constructor(rootNode: any, path: string, parsedPath?: any);

    path: string;
    parsedPath: any;
    node: any;
    rootNode: any;

    getValue(targetArray: any, offset: number): any;
    setValue(sourceArray: any, offset: number): void;
    bind(): void;
    unbind(): void;

    BindingType: { [bindingType: string]: number };
    Versioning: { [versioning: string]: number };

    GetterByBindingType: Array<() => void>;
    SetterByBindingTypeAndVersioning: Array<Array<() => void>>;

    static create(root: any, path: any, parsedPath?: any): PropertyBinding | PropertyBinding.Composite;
    static sanitizeNodeName(name: string): string;
    static parseTrackName(trackName: string): ParseTrackNameResults;
    static findNode(root: any, nodeName: string): any;
}

declare namespace PropertyBinding {
    class Composite {
        constructor(targetGroup: any, path: any, parsedPath?: any);

        getValue(array: any, offset: number): any;
        setValue(array: any, offset: number): void;
        bind(): void;
        unbind(): void;
    }
}

declare namespace AnimationUtils {
    function arraySlice(array: any, from: number, to: number): any;
    function convertArray(array: any, type: any, forceClone: boolean): any;
    function isTypedArray(object: any): boolean;
    function getKeyframeOrder(times: number[]): number[];
    function sortedArray(values: any[], stride: number, order: number[]): any[];
    function flattenJSON(jsonKeys: string[], times: any[], values: any[], valuePropertyName: string): void;

    /**
     * @param sourceClip
     * @param name
     * @param startFrame
     * @param endFrame
     * @param [fps=30]
     */
    function subclip(
        sourceClip: AnimationClip,
        name: string,
        startFrame: number,
        endFrame: number,
        fps?: number,
    ): AnimationClip;

    /**
     * @param targetClip
     * @param [referenceFrame=0]
     * @param [referenceClip=targetClip]
     * @param [fps=30]
     */
    function makeClipAdditive(
        targetClip: AnimationClip,
        referenceFrame?: number,
        referenceClip?: AnimationClip,
        fps?: number,
    ): AnimationClip;
}

declare class AnimationObjectGroup {
    constructor(...args: any[]);

    uuid: string;
    stats: {
        bindingsPerObject: number;
        objects: {
            total: number;
            inUse: number;
        };
    };
    readonly isAnimationObjectGroup: true;

    add(...args: any[]): void;
    remove(...args: any[]): void;
    uncache(...args: any[]): void;
}

// Animation ////////////////////////////////////////////////////////////////////////////////////////

declare class AnimationAction {
    constructor(mixer: AnimationMixer, clip: AnimationClip, localRoot?: Object3D, blendMode?: AnimationBlendMode);

    blendMode: AnimationBlendMode;

    /**
     * @default THREE.LoopRepeat
     */
    loop: AnimationActionLoopStyles;

    /**
     * @default 0
     */
    time: number;

    /**
     * @default 1
     */
    timeScale: number;

    /**
     * @default 1
     */
    weight: number;

    /**
     * @default Infinity
     */
    repetitions: number;

    /**
     * @default false
     */
    paused: boolean;

    /**
     * @default true
     */
    enabled: boolean;

    /**
     * @default false
     */
    clampWhenFinished: boolean;

    /**
     * @default true
     */
    zeroSlopeAtStart: boolean;

    /**
     * @default true
     */
    zeroSlopeAtEnd: boolean;

    play(): AnimationAction;
    stop(): AnimationAction;
    reset(): AnimationAction;
    isRunning(): boolean;
    isScheduled(): boolean;
    startAt(time: number): AnimationAction;
    setLoop(mode: AnimationActionLoopStyles, repetitions: number): AnimationAction;
    setEffectiveWeight(weight: number): AnimationAction;
    getEffectiveWeight(): number;
    fadeIn(duration: number): AnimationAction;
    fadeOut(duration: number): AnimationAction;
    crossFadeFrom(fadeOutAction: AnimationAction, duration: number, warp: boolean): AnimationAction;
    crossFadeTo(fadeInAction: AnimationAction, duration: number, warp: boolean): AnimationAction;
    stopFading(): AnimationAction;
    setEffectiveTimeScale(timeScale: number): AnimationAction;
    getEffectiveTimeScale(): number;
    setDuration(duration: number): AnimationAction;
    syncWith(action: AnimationAction): AnimationAction;
    halt(duration: number): AnimationAction;
    warp(statTimeScale: number, endTimeScale: number, duration: number): AnimationAction;
    stopWarping(): AnimationAction;
    getMixer(): AnimationMixer;
    getClip(): AnimationClip;
    getRoot(): Object3D;
}

declare class AnimationMixer extends EventDispatcher {
    constructor(root: Object3D | AnimationObjectGroup);

    /**
     * @default 0
     */
    time: number;

    /**
     * @default 1.0
     */
    timeScale: number;

    clipAction(
        clip: AnimationClip,
        root?: Object3D | AnimationObjectGroup,
        blendMode?: AnimationBlendMode,
    ): AnimationAction;
    existingAction(clip: AnimationClip, root?: Object3D | AnimationObjectGroup): AnimationAction | null;
    stopAllAction(): AnimationMixer;
    update(deltaTime: number): AnimationMixer;
    setTime(timeInSeconds: number): AnimationMixer;
    getRoot(): Object3D | AnimationObjectGroup;
    uncacheClip(clip: AnimationClip): void;
    uncacheRoot(root: Object3D | AnimationObjectGroup): void;
    uncacheAction(clip: AnimationClip, root?: Object3D | AnimationObjectGroup): void;
}

/**
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/audio/AudioContext.js|src/audio/AudioContext.js}
 */
declare namespace AudioContext {
    function getContext(): AudioContext;

    function setContext(context: AudioContext): void;
}

declare class AudioListener extends Object3D {
    constructor();

    type: 'AudioListener';
    context: AudioContext;
    gain: GainNode;

    /**
     * @default null
     */
    filter: any;

    /**
     * @default 0
     */
    timeDelta: number;

    getInput(): GainNode;
    removeFilter(): this;
    setFilter(value: any): this;
    getFilter(): any;
    setMasterVolume(value: number): this;
    getMasterVolume(): number;
    updateMatrixWorld(force?: boolean): void;
}

// Extras / Audio /////////////////////////////////////////////////////////////////////

declare class Audio<NodeType extends AudioNode = GainNode> extends Object3D {
    constructor(listener: AudioListener);
    type: 'Audio';

    listener: AudioListener;
    context: AudioContext;
    gain: GainNode;

    /**
     * @default false
     */
    autoplay: boolean;
    buffer: null | AudioBuffer;

    /**
     * @default 0
     */
    detune: number;

    /**
     * @default false
     */
    loop: boolean;

    /**
     * @default 0
     */
    loopStart: number;

    /**
     * @default 0
     */
    loopEnd: number;

    /**
     * @default 0
     */
    offset: number;

    /**
     * @default undefined
     */
    duration: number | undefined;

    /**
     * @default 1
     */
    playbackRate: number;

    /**
     * @default false
     */
    isPlaying: boolean;

    /**
     * @default true
     */
    hasPlaybackControl: boolean;

    /**
     * @default 'empty'
     */
    sourceType: string;
    source: null | AudioBufferSourceNode;

    /**
     * @default []
     */
    filters: AudioNode[];

    getOutput(): NodeType;
    setNodeSource(audioNode: AudioBufferSourceNode): this;
    setMediaElementSource(mediaElement: HTMLMediaElement): this;
    setMediaStreamSource(mediaStream: MediaStream): this;
    setBuffer(audioBuffer: AudioBuffer): this;
    play(delay?: number): this;
    onEnded(): void;
    pause(): this;
    stop(): this;
    connect(): this;
    disconnect(): this;
    setDetune(value: number): this;
    getDetune(): number;
    getFilters(): AudioNode[];
    setFilters(value: AudioNode[]): this;
    getFilter(): AudioNode;
    setFilter(filter: AudioNode): this;
    setPlaybackRate(value: number): this;
    getPlaybackRate(): number;
    getLoop(): boolean;
    setLoop(value: boolean): this;
    setLoopStart(value: number): this;
    setLoopEnd(value: number): this;
    getVolume(): number;
    setVolume(value: number): this;
    /**
     * @deprecated Use {@link AudioLoader} instead.
     */
    load(file: string): Audio;
}

declare class PositionalAudio extends Audio<PannerNode> {
    constructor(listener: AudioListener);

    panner: PannerNode;

    getOutput(): PannerNode;
    setRefDistance(value: number): this;
    getRefDistance(): number;
    setRolloffFactor(value: number): this;
    getRolloffFactor(): number;
    setDistanceModel(value: string): this;
    getDistanceModel(): string;
    setMaxDistance(value: number): this;
    getMaxDistance(): number;
    setDirectionalCone(coneInnerAngle: number, coneOuterAngle: number, coneOuterGain: number): this;
    updateMatrixWorld(force?: boolean): void;
}

declare class AudioAnalyser {
    /**
     * @param audio
     * @param [fftSize=2048]
     */
    constructor(audio: Audio<AudioNode>, fftSize?: number);

    analyser: AnalyserNode;
    data: Uint8Array;

    getFrequencyData(): Uint8Array;
    getAverageFrequency(): number;

    /**
     * @deprecated Use {@link AudioAnalyser#getFrequencyData .getFrequencyData()} instead.
     */
    getData(file: any): any;
}

declare class StereoCamera extends Camera {
    constructor();

    type: 'StereoCamera';

    /**
     * @default 1
     */
    aspect: number;

    /**
     * @default 0.064
     */
    eyeSep: number;

    cameraL: PerspectiveCamera;
    cameraR: PerspectiveCamera;

    update(camera: PerspectiveCamera): void;
}

/**
 * Camera with orthographic projection
 *
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/cameras/OrthographicCamera.js|src/cameras/OrthographicCamera.js}
 *
 * @example
 * const camera = new THREE.OrthographicCamera( width / - 2, width / 2, height / 2, height / - 2, 1, 1000 );
 * scene.add( camera );
 */
declare class OrthographicCamera extends Camera {
    /**
     * @param left Camera frustum left plane.
     * @param right Camera frustum right plane.
     * @param top Camera frustum top plane.
     * @param bottom Camera frustum bottom plane.
     * @param [near=0.1] Camera frustum near plane.
     * @param [far=2000] Camera frustum far plane.
     */
    constructor(left?: number, right?: number, top?: number, bottom?: number, near?: number, far?: number);

    type: 'OrthographicCamera';

    readonly isOrthographicCamera: true;

    /**
     * @default 1
     */
    zoom: number;

    /**
     * @default null
     */
    view: null | {
        enabled: boolean;
        fullWidth: number;
        fullHeight: number;
        offsetX: number;
        offsetY: number;
        width: number;
        height: number;
    };

    /**
     * Camera frustum left plane.
     * @default -1
     */
    left: number;

    /**
     * Camera frustum right plane.
     * @default 1
     */
    right: number;

    /**
     * Camera frustum top plane.
     * @default 1
     */
    top: number;

    /**
     * Camera frustum bottom plane.
     * @default -1
     */
    bottom: number;

    /**
     * Camera frustum near plane.
     * @default 0.1
     */
    near: number;

    /**
     * Camera frustum far plane.
     * @default 2000
     */
    far: number;

    /**
     * Updates the camera projection matrix. Must be called after change of parameters.
     */
    updateProjectionMatrix(): void;
    setViewOffset(
        fullWidth: number,
        fullHeight: number,
        offsetX: number,
        offsetY: number,
        width: number,
        height: number,
    ): void;
    clearViewOffset(): void;
    toJSON(meta?: any): any;
}

declare class CubeTexture extends Texture {
    /**
     * @param [images=[]]
     * @param [mapping=THREE.CubeReflectionMapping]
     * @param [wrapS=THREE.ClampToEdgeWrapping]
     * @param [wrapT=THREE.ClampToEdgeWrapping]
     * @param [magFilter=THREE.LinearFilter]
     * @param [minFilter=THREE.LinearMipmapLinearFilter]
     * @param [format=THREE.RGBAFormat]
     * @param [type=THREE.UnsignedByteType]
     * @param [anisotropy=1]
     * @param [encoding=THREE.LinearEncoding]
     */
    constructor(
        images?: any[], // HTMLImageElement or HTMLCanvasElement
        mapping?: Mapping,
        wrapS?: Wrapping,
        wrapT?: Wrapping,
        magFilter?: TextureFilter,
        minFilter?: TextureFilter,
        format?: PixelFormat,
        type?: TextureDataType,
        anisotropy?: number,
        encoding?: TextureEncoding,
    );

    images: any; // returns and sets the value of Texture.image in the codde ?

    /**
     * @default false
     */
    flipY: boolean;

    readonly isCubeTexture: true;
}

declare class WebGLCubeRenderTarget extends WebGLRenderTarget {
    constructor(size: number, options?: WebGLRenderTargetOptions);

    texture: CubeTexture;

    fromEquirectangularTexture(renderer: WebGLRenderer, texture: Texture): this;

    clear(renderer: WebGLRenderer, color: boolean, depth: boolean, stencil: boolean): void;
}

declare class CubeCamera extends Object3D {
    constructor(near: number, far: number, renderTarget: WebGLCubeRenderTarget);

    type: 'CubeCamera';

    renderTarget: WebGLCubeRenderTarget;

    update(renderer: WebGLRenderer, scene: Scene$1): void;
}

declare class Uniform {
    constructor(value: any);
    /**
     * @deprecated
     */
    constructor(type: string, value: any);
    /**
     * @deprecated
     */
    type: string;
    value: any;
    /**
     * @deprecated Use {@link Object3D#onBeforeRender object.onBeforeRender()} instead.
     */
    dynamic: boolean;

    /**
     * @deprecated Use {@link Object3D#onBeforeRender object.onBeforeRender()} instead.
     */
    onUpdate(callback: () => void): Uniform;
}

/**
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/core/InstancedBufferGeometry.js|src/core/InstancedBufferGeometry.js}
 */
declare class InstancedBufferGeometry extends BufferGeometry {
    constructor();

    /**
     * @default 'InstancedBufferGeometry
     */
    type: string;

    isInstancedBufferGeometry: boolean;

    /**
     * @default Infinity
     */
    instanceCount: number;
}

/**
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/core/InstancedInterleavedBuffer.js|src/core/InstancedInterleavedBuffer.js}
 */
declare class InstancedInterleavedBuffer extends InterleavedBuffer {
    constructor(array: ArrayLike<number>, stride: number, meshPerAttribute?: number);

    /**
     * @default 1
     */
    meshPerAttribute: number;
}

/**
 * see {@link https://github.com/mrdoob/three.js/blob/master/examples/jsm/utils/BufferGeometryUtils.js|examples/jsm/utils/BufferGeometryUtils.js}
 */
declare namespace BufferGeometryUtils {
    function mergeBufferGeometries(geometries: BufferGeometry[]): BufferGeometry;
    function computeTangents(geometry: BufferGeometry): null;
    function mergeBufferAttributes(attributes: BufferAttribute[]): BufferAttribute;
}

/**
 * @deprecated
 */
declare namespace GeometryUtils {
    /**
     * @deprecated Use {@link Geometry#merge geometry.merge( geometry2, matrix, materialIndexOffset )} instead.
     */
    function merge(geometry1: any, geometry2: any, materialIndexOffset?: any): any;
    /**
     * @deprecated Use {@link Geometry#center geometry.center()} instead.
     */
    function center(geometry: any): any;
}

/**
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/core/InstancedBufferAttribute.js|src/core/InstancedBufferAttribute.js}
 */
declare class InstancedBufferAttribute extends BufferAttribute {
    constructor(array: ArrayLike<number>, itemSize: number, normalized?: boolean, meshPerAttribute?: number);

    /**
     * @default 1
     */
    meshPerAttribute: number;
}

/**
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/core/GLBufferAttribute.js|src/core/GLBufferAttribute.js}
 */

declare class GLBufferAttribute {
    constructor(buffer: WebGLBuffer, type: number, itemSize: number, elementSize: 1 | 2 | 4, count: number);

    buffer: WebGLBuffer;
    type: number;
    itemSize: number;
    elementSize: 1 | 2 | 4;
    count: number;
    version: number;

    readonly isGLBufferAttribute: true;

    set needsUpdate(value: boolean);

    setBuffer(buffer: WebGLBuffer): this;
    setType(type: number, elementSize: 1 | 2 | 4): this;
    setItemSize(itemSize: number): this;
    setCount(count: number): this;
}

/**
 * Object for keeping track of time.
 *
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/core/Clock.js|src/core/Clock.js}
 */
declare class Clock {
    /**
     * @param [autoStart=true] Automatically start the clock.
     */
    constructor(autoStart?: boolean);

    /**
     * If set, starts the clock automatically when the first update is called.
     * @default true
     */
    autoStart: boolean;

    /**
     * When the clock is running, It holds the starttime of the clock.
     * This counted from the number of milliseconds elapsed since 1 January 1970 00:00:00 UTC.
     * @default 0
     */
    startTime: number;

    /**
     * When the clock is running, It holds the previous time from a update.
     * This counted from the number of milliseconds elapsed since 1 January 1970 00:00:00 UTC.
     * @default 0
     */
    oldTime: number;

    /**
     * When the clock is running, It holds the time elapsed between the start of the clock to the previous update.
     * This parameter is in seconds of three decimal places.
     * @default 0
     */
    elapsedTime: number;

    /**
     * This property keeps track whether the clock is running or not.
     * @default false
     */
    running: boolean;

    /**
     * Starts clock.
     */
    start(): void;

    /**
     * Stops clock.
     */
    stop(): void;

    /**
     * Get the seconds passed since the clock started.
     */
    getElapsedTime(): number;

    /**
     * Get the seconds passed since the last call to this method.
     */
    getDelta(): number;
}

// Extras / Core /////////////////////////////////////////////////////////////////////

/**
 * An extensible curve object which contains methods for interpolation
 * class Curve<T extends Vector>
 */
declare class Curve<T extends Vector> {
    /**
     * @default 'Curve'
     */
    type: string;

    /**
     * This value determines the amount of divisions when calculating the cumulative segment lengths of a curve via .getLengths.
     * To ensure precision when using methods like .getSpacedPoints, it is recommended to increase .arcLengthDivisions if the curve is very large.
     * @default 200
     */
    arcLengthDivisions: number;

    /**
     * Returns a vector for point t of the curve where t is between 0 and 1
     * getPoint(t: number, optionalTarget?: T): T;
     */
    getPoint(t: number, optionalTarget?: T): T;

    /**
     * Returns a vector for point at relative position in curve according to arc length
     * getPointAt(u: number, optionalTarget?: T): T;
     */
    getPointAt(u: number, optionalTarget?: T): T;

    /**
     * Get sequence of points using getPoint( t )
     * getPoints(divisions?: number): T[];
     */
    getPoints(divisions?: number): T[];

    /**
     * Get sequence of equi-spaced points using getPointAt( u )
     * getSpacedPoints(divisions?: number): T[];
     */
    getSpacedPoints(divisions?: number): T[];

    /**
     * Get total curve arc length
     */
    getLength(): number;

    /**
     * Get list of cumulative segment lengths
     */
    getLengths(divisions?: number): number[];

    /**
     * Update the cumlative segment distance cache
     */
    updateArcLengths(): void;

    /**
     * Given u ( 0 .. 1 ), get a t to find p. This gives you points which are equi distance
     */
    getUtoTmapping(u: number, distance: number): number;

    /**
     * Returns a unit vector tangent at t. If the subclassed curve do not implement its tangent derivation, 2 points a
     * small delta apart will be used to find its gradient which seems to give a reasonable approximation
     * getTangent(t: number, optionalTarget?: T): T;
     */
    getTangent(t: number, optionalTarget?: T): T;

    /**
     * Returns tangent at equidistance point u on the curve
     * getTangentAt(u: number, optionalTarget?: T): T;
     */
    getTangentAt(u: number, optionalTarget?: T): T;

    /**
     * Generate Frenet frames of the curve
     */
    computeFrenetFrames(
        segments: number,
        closed?: boolean,
    ): {
        tangents: Vector3[];
        normals: Vector3[];
        binormals: Vector3[];
    };

    clone(): this;
    copy(source: Curve<T>): this;
    toJSON(): object;
    fromJSON(json: object): this;

    /**
     * @deprecated since r84.
     */
    static create(constructorFunc: () => void, getPointFunc: () => void): () => void;
}

declare class EllipseCurve extends Curve<Vector2> {
    constructor(
        aX: number,
        aY: number,
        xRadius: number,
        yRadius: number,
        aStartAngle: number,
        aEndAngle: number,
        aClockwise: boolean,
        aRotation: number,
    );

    /**
     * @default 'EllipseCurve'
     */
    type: string;

    /**
     * @default 0
     */
    aX: number;

    /**
     * @default 0
     */
    aY: number;

    /**
     * @default 1
     */
    xRadius: number;

    /**
     * @default 1
     */
    yRadius: number;

    /**
     * @default 0
     */
    aStartAngle: number;

    /**
     * @default 2 * Math.PI
     */
    aEndAngle: number;

    /**
     * @default false
     */
    aClockwise: boolean;

    /**
     * @default 0
     */
    aRotation: number;
}

declare class ArcCurve extends EllipseCurve {
    constructor(aX: number, aY: number, aRadius: number, aStartAngle: number, aEndAngle: number, aClockwise: boolean);

    /**
     * @default 'ArcCurve'
     */
    type: string;
}

// Extras / Curves /////////////////////////////////////////////////////////////////////
declare namespace CurveUtils {
    function tangentQuadraticBezier(t: number, p0: number, p1: number, p2: number): number;
    function tangentCubicBezier(t: number, p0: number, p1: number, p2: number, p3: number): number;
    function tangentSpline(t: number, p0: number, p1: number, p2: number, p3: number): number;
    function interpolate(p0: number, p1: number, p2: number, p3: number, t: number): number;
}

declare class CatmullRomCurve3 extends Curve<Vector3> {
    /**
     * @param [points=[]]
     * @param [closed=false]
     * @param [curveType='centripetal']
     * @param [tension=0.5]
     */
    constructor(points?: Vector3[], closed?: boolean, curveType?: string, tension?: number);

    /**
     * @default 'CatmullRomCurve3'
     */
    type: string;

    /**
     * @default []
     */
    points: Vector3[];
}

declare class CubicBezierCurve extends Curve<Vector2> {
    constructor(v0: Vector2, v1: Vector2, v2: Vector2, v3: Vector2);

    /**
     * @default 'CubicBezierCurve'
     */
    type: string;

    /**
     * @default new THREE.Vector2()
     */
    v0: Vector2;

    /**
     * @default new THREE.Vector2()
     */
    v1: Vector2;

    /**
     * @default new THREE.Vector2()
     */
    v2: Vector2;

    /**
     * @default new THREE.Vector2()
     */
    v3: Vector2;
}

declare class CubicBezierCurve3 extends Curve<Vector3> {
    constructor(v0: Vector3, v1: Vector3, v2: Vector3, v3: Vector3);

    /**
     * @default 'CubicBezierCurve3'
     */
    type: string;

    /**
     * @default new THREE.Vector3()
     */
    v0: Vector3;

    /**
     * @default new THREE.Vector3()
     */
    v1: Vector3;

    /**
     * @default new THREE.Vector3()
     */
    v2: Vector3;

    /**
     * @default new THREE.Vector3()
     */
    v3: Vector3;
}

declare class LineCurve extends Curve<Vector2> {
    constructor(v1: Vector2, v2: Vector2);

    /**
     * @default 'LineCurve'
     */
    type: string;

    /**
     * @default new THREE.Vector2()
     */
    v1: Vector2;

    /**
     * @default new THREE.Vector2()
     */
    v2: Vector2;
}

declare class LineCurve3 extends Curve<Vector3> {
    constructor(v1: Vector3, v2: Vector3);

    /**
     * @default 'LineCurve3'
     */
    type: string;

    /**
     * @default new THREE.Vector3()
     */
    v1: Vector3;

    /**
     * @default new THREE.Vector3()
     */
    v2: Vector3;
}

declare class QuadraticBezierCurve extends Curve<Vector2> {
    constructor(v0: Vector2, v1: Vector2, v2: Vector2);

    /**
     * @default 'QuadraticBezierCurve'
     */
    type: string;

    /**
     * @default new THREE.Vector2()
     */
    v0: Vector2;

    /**
     * @default new THREE.Vector2()
     */
    v1: Vector2;

    /**
     * @default new THREE.Vector2()
     */
    v2: Vector2;
}

declare class QuadraticBezierCurve3 extends Curve<Vector3> {
    constructor(v0: Vector3, v1: Vector3, v2: Vector3);

    /**
     * @default 'QuadraticBezierCurve3'
     */
    type: string;

    /**
     * @default new THREE.Vector3()
     */
    v0: Vector3;

    /**
     * @default new THREE.Vector3()
     */
    v1: Vector3;

    /**
     * @default new THREE.Vector3()
     */
    v2: Vector3;
}

declare class SplineCurve extends Curve<Vector2> {
    constructor(points?: Vector2[]);

    /**
     * @default 'SplineCurve'
     */
    type: string;

    /**
     * @default []
     */
    points: Vector2[];
}

declare class CurvePath<T extends Vector> extends Curve<T> {
    constructor();

    /**
     * @default 'CurvePath'
     */
    type: string;

    /**
     * @default []
     */
    curves: Array<Curve<T>>;

    /**
     * @default false
     */
    autoClose: boolean;

    add(curve: Curve<T>): void;
    closePath(): void;
    getPoint(t: number, optionalTarget?: T): T;
    getCurveLengths(): number[];
}

/**
 * a 2d path representation, comprising of points, lines, and cubes, similar to the html5 2d canvas api. It extends CurvePath.
 */
declare class Path extends CurvePath<Vector2> {
    constructor(points?: Vector2[]);

    /**
     * @default 'Path'
     */
    type: string;

    /**
     * @default new THREE.Vector2()
     */
    currentPoint: Vector2;

    /**
     * @deprecated Use {@link Path#setFromPoints .setFromPoints()} instead.
     */
    fromPoints(vectors: Vector2[]): this;
    setFromPoints(vectors: Vector2[]): this;
    moveTo(x: number, y: number): this;
    lineTo(x: number, y: number): this;
    quadraticCurveTo(aCPx: number, aCPy: number, aX: number, aY: number): this;
    bezierCurveTo(aCP1x: number, aCP1y: number, aCP2x: number, aCP2y: number, aX: number, aY: number): this;
    splineThru(pts: Vector2[]): this;
    arc(aX: number, aY: number, aRadius: number, aStartAngle: number, aEndAngle: number, aClockwise: boolean): this;
    absarc(aX: number, aY: number, aRadius: number, aStartAngle: number, aEndAngle: number, aClockwise: boolean): this;
    ellipse(
        aX: number,
        aY: number,
        xRadius: number,
        yRadius: number,
        aStartAngle: number,
        aEndAngle: number,
        aClockwise: boolean,
        aRotation: number,
    ): this;
    absellipse(
        aX: number,
        aY: number,
        xRadius: number,
        yRadius: number,
        aStartAngle: number,
        aEndAngle: number,
        aClockwise: boolean,
        aRotation: number,
    ): this;
}

/**
 * Defines a 2d shape plane using paths.
 */
declare class Shape extends Path {
    constructor(points?: Vector2[]);

    /**
     * @default 'Shape'
     */
    type: string;

    uuid: string;

    /**
     * @default []
     */
    holes: Path[];

    getPointsHoles(divisions: number): Vector2[][];

    extractPoints(divisions: number): {
        shape: Vector2[];
        holes: Vector2[][];
    };
}

declare class ShapePath {
    constructor();

    /**
     * @default 'ShapePath'
     */
    type: string;

    /**
     * @default new THREE.Color()
     */
    color: Color;

    /**
     * @default []
     */
    subPaths: any[];

    /**
     * @default null
     */
    currentPath: any;

    moveTo(x: number, y: number): this;
    lineTo(x: number, y: number): this;
    quadraticCurveTo(aCPx: number, aCPy: number, aX: number, aY: number): this;
    bezierCurveTo(aCP1x: number, aCP1y: number, aCP2x: number, aCP2y: number, aX: number, aY: number): this;
    splineThru(pts: Vector2[]): this;
    toShapes(isCCW: boolean, noHoles?: boolean): Shape[];
}

declare function toHalfFloat(val: number): number;
declare function fromHalfFloat(val: number): number;

declare namespace Earcut {
    function triangulate(data: number[], holeIndices: number[], dim: number): Triangle[];
}

declare namespace ImageUtils {
    function getDataURL(image: any): string;

    /**
     * @deprecated
     */
    let crossOrigin: string;

    /**
     * @deprecated Use {@link TextureLoader THREE.TextureLoader()} instead.
     */
    function loadTexture(
        url: string,
        mapping?: Mapping,
        onLoad?: (texture: Texture) => void,
        onError?: (message: string) => void,
    ): Texture;

    /**
     * @deprecated Use {@link CubeTextureLoader THREE.CubeTextureLoader()} instead.
     */
    function loadTextureCube(
        array: string[],
        mapping?: Mapping,
        onLoad?: (texture: Texture) => void,
        onError?: (message: string) => void,
    ): Texture;

    function sRGBToLinear(image: any): HTMLCanvasElement | { data: number[]; width: number; height: number };
}

interface Vec2 {
    x: number;
    y: number;
}

declare namespace ShapeUtils {
    function area(contour: Vec2[]): number;
    function triangulateShape(contour: Vec2[], holes: Vec2[][]): number[][];
    function isClockWise(pts: Vec2[]): boolean;
}

declare class PMREMGenerator {
    constructor(renderer: WebGLRenderer);
    fromScene(scene: Scene$1, sigma?: number, near?: number, far?: number): WebGLRenderTarget;
    fromEquirectangular(equirectangular: Texture, renderTarget?: WebGLRenderTarget | null): WebGLRenderTarget;
    fromCubemap(cubemap: CubeTexture, renderTarget?: WebGLRenderTarget | null): WebGLRenderTarget;
    compileCubemapShader(): void;
    compileEquirectangularShader(): void;
    dispose(): void;
}

declare class BoxGeometry extends BufferGeometry {
    /**
     * @param [width=1] — Width of the sides on the X axis.
     * @param [height=1] — Height of the sides on the Y axis.
     * @param [depth=1] — Depth of the sides on the Z axis.
     * @param [widthSegments=1] — Number of segmented faces along the width of the sides.
     * @param [heightSegments=1] — Number of segmented faces along the height of the sides.
     * @param [depthSegments=1] — Number of segmented faces along the depth of the sides.
     */
    constructor(
        width?: number,
        height?: number,
        depth?: number,
        widthSegments?: number,
        heightSegments?: number,
        depthSegments?: number,
    );

    /**
     * @default 'BoxGeometry'
     */
    type: string;

    parameters: {
        width: number;
        height: number;
        depth: number;
        widthSegments: number;
        heightSegments: number;
        depthSegments: number;
    };

    static fromJSON(data: any): BoxGeometry;
}

declare class CapsuleGeometry extends BufferGeometry {
    /**
     * @param [radius=1] — Radius of the capsule.
     * @param [length=1] — Length of the middle section.
     * @param [capSegments=4] — Number of curve segments used to build the caps.
     * @param [radialSegments=8] — Number of segmented faces around the circumference of the capsule.
     */
    constructor(radius?: number, length?: number, capSegments?: number, radialSegments?: number);

    /**
     * @default 'CapsuleGeometry'
     */
    type: string;

    parameters: {
        radius: number;
        length: number;
        capSegments: number;
        radialSegments: number;
    };

    static fromJSON(data: any): CapsuleGeometry;
}

declare class CircleGeometry extends BufferGeometry {
    /**
     * @param [radius=1]
     * @param [segments=8]
     * @param [thetaStart=0]
     * @param [thetaLength=Math.PI * 2]
     */
    constructor(radius?: number, segments?: number, thetaStart?: number, thetaLength?: number);

    /**
     * @default 'CircleGeometry'
     */
    type: string;

    parameters: {
        radius: number;
        segments: number;
        thetaStart: number;
        thetaLength: number;
    };

    static fromJSON(data: any): CircleGeometry;
}

declare class CylinderGeometry extends BufferGeometry {
    /**
     * @param [radiusTop=1] — Radius of the cylinder at the top.
     * @param [radiusBottom=1] — Radius of the cylinder at the bottom.
     * @param [height=1] — Height of the cylinder.
     * @param [radialSegments=8] — Number of segmented faces around the circumference of the cylinder.
     * @param [heightSegments=1] — Number of rows of faces along the height of the cylinder.
     * @param [openEnded=false] - A Boolean indicating whether or not to cap the ends of the cylinder.
     * @param [thetaStart=0]
     * @param [thetaLength=Math.PI * 2]
     */
    constructor(
        radiusTop?: number,
        radiusBottom?: number,
        height?: number,
        radialSegments?: number,
        heightSegments?: number,
        openEnded?: boolean,
        thetaStart?: number,
        thetaLength?: number,
    );

    /**
     * @default 'CylinderGeometry'
     */
    type: string;

    parameters: {
        radiusTop: number;
        radiusBottom: number;
        height: number;
        radialSegments: number;
        heightSegments: number;
        openEnded: boolean;
        thetaStart: number;
        thetaLength: number;
    };

    static fromJSON(data: any): CylinderGeometry;
}

declare class ConeGeometry extends CylinderGeometry {
    /**
     * @param [radius=1] — Radius of the cone base.
     * @param [height=1] — Height of the cone.
     * @param [radialSegments=8] — Number of segmented faces around the circumference of the cone.
     * @param [heightSegments=1] — Number of rows of faces along the height of the cone.
     * @param [openEnded=false] — A Boolean indicating whether the base of the cone is open or capped.
     * @param [thetaStart=0]
     * @param [thetaLength=Math.PI * 2]
     */
    constructor(
        radius?: number,
        height?: number,
        radialSegments?: number,
        heightSegments?: number,
        openEnded?: boolean,
        thetaStart?: number,
        thetaLength?: number,
    );

    /**
     * @default 'ConeGeometry'
     */
    type: string;

    static fromJSON(data: any): ConeGeometry;
}

declare class PolyhedronGeometry extends BufferGeometry {
    /**
     * @param vertices
     * @param indices
     * @param [radius=1]
     * @param [detail=0]
     */
    constructor(vertices?: number[], indices?: number[], radius?: number, detail?: number);

    /**
     * @default 'PolyhedronGeometry'
     */
    type: string;

    parameters: {
        vertices: number[];
        indices: number[];
        radius: number;
        detail: number;
    };

    static fromJSON(data: any): PolyhedronGeometry;
}

declare class DodecahedronGeometry extends PolyhedronGeometry {
    /**
     * @param [radius=1]
     * @param [detail=0]
     */
    constructor(radius?: number, detail?: number);

    /**
     * @default 'DodecahedronGeometry'
     */
    type: string;

    static fromJSON(data: any): DodecahedronGeometry;
}

declare class EdgesGeometry<TBufferGeometry extends BufferGeometry = BufferGeometry> extends BufferGeometry {
    /**
     * @param geometry
     * @param [thresholdAngle=1]
     */
    constructor(geometry?: TBufferGeometry, thresholdAngle?: number);

    /**
     * @default 'EdgesGeometry'
     */
    type: string;

    parameters: {
        geometry: TBufferGeometry;
        thresholdAngle: number;
    };
}

interface ExtrudeGeometryOptions {
    /**
     * @default 12
     */
    curveSegments?: number | undefined;
    /**
     * @default 1
     */
    steps?: number | undefined;
    /**
     * @default 100
     */
    depth?: number | undefined;
    /**
     * @default true
     */
    bevelEnabled?: boolean | undefined;
    /**
     * @default 6
     */
    bevelThickness?: number | undefined;
    bevelSize?: number | undefined;
    /**
     * @default 0
     */
    bevelOffset?: number | undefined;
    /**
     * @default 3
     */
    bevelSegments?: number | undefined;
    extrudePath?: Curve<Vector3> | undefined;
    UVGenerator?: UVGenerator | undefined;
}

interface UVGenerator {
    generateTopUV(
        geometry: ExtrudeGeometry,
        vertices: number[],
        indexA: number,
        indexB: number,
        indexC: number,
    ): Vector2[];
    generateSideWallUV(
        geometry: ExtrudeGeometry,
        vertices: number[],
        indexA: number,
        indexB: number,
        indexC: number,
        indexD: number,
    ): Vector2[];
}

declare class ExtrudeGeometry extends BufferGeometry {
    constructor(shapes?: Shape | Shape[], options?: ExtrudeGeometryOptions);

    /**
     * @default 'ExtrudeGeometry'
     */
    type: string;

    addShapeList(shapes: Shape[], options?: any): void;
    addShape(shape: Shape, options?: any): void;

    static fromJSON(data: any): ExtrudeGeometry;
}

declare class IcosahedronGeometry extends PolyhedronGeometry {
    /**
     * @param [radius=1]
     * @param [detail=0]
     */
    constructor(radius?: number, detail?: number);

    /**
     * @default 'IcosahedronGeometry'
     */
    type: string;

    static fromJSON(data: any): IcosahedronGeometry;
}

declare class LatheGeometry extends BufferGeometry {
    /**
     * @param points
     * @param [segments=12]
     * @param [phiStart=0]
     * @param [phiLength=Math.PI * 2]
     */
    constructor(points?: Vector2[], segments?: number, phiStart?: number, phiLength?: number);

    /**
     * @default 'LatheGeometry'
     */
    type: string;

    parameters: {
        points: Vector2[];
        segments: number;
        phiStart: number;
        phiLength: number;
    };

    static fromJSON(data: any): LatheGeometry;
}

declare class OctahedronGeometry extends PolyhedronGeometry {
    /**
     * @param [radius=1]
     * @param [detail=0]
     */
    constructor(radius?: number, detail?: number);

    /**
     * @default 'OctahedronGeometry'
     */
    type: string;

    static fromJSON(data: any): OctahedronGeometry;
}

declare class PlaneGeometry extends BufferGeometry {
    /**
     * @param [width=1] — Width of the sides on the X axis.
     * @param [height=1] — Height of the sides on the Y axis.
     * @param [widthSegments=1] — Number of segmented faces along the width of the sides.
     * @param [heightSegments=1] — Number of segmented faces along the height of the sides.
     */
    constructor(width?: number, height?: number, widthSegments?: number, heightSegments?: number);

    /**
     * @default 'PlaneGeometry'
     */
    type: string;

    parameters: {
        width: number;
        height: number;
        widthSegments: number;
        heightSegments: number;
    };

    static fromJSON(data: any): PlaneGeometry;
}

declare class RingGeometry extends BufferGeometry {
    /**
     * @param [innerRadius=0.5]
     * @param [outerRadius=1]
     * @param [thetaSegments=8]
     * @param [phiSegments=1]
     * @param [thetaStart=0]
     * @param [thetaLength=Math.PI * 2]
     */
    constructor(
        innerRadius?: number,
        outerRadius?: number,
        thetaSegments?: number,
        phiSegments?: number,
        thetaStart?: number,
        thetaLength?: number,
    );

    /**
     * @default 'RingGeometry'
     */
    type: string;

    parameters: {
        innerRadius: number;
        outerRadius: number;
        thetaSegments: number;
        phiSegments: number;
        thetaStart: number;
        thetaLength: number;
    };

    static fromJSON(data: any): RingGeometry;
}

declare class ShapeGeometry extends BufferGeometry {
    /**
     * @default 'ShapShapeGeometryeBufferGeometry'
     */
    type: string;

    constructor(shapes?: Shape | Shape[], curveSegments?: number);

    static fromJSON(data: any): ShapeGeometry;
}

declare class SphereGeometry extends BufferGeometry {
    /**
     * @param [radius=50] — sphere radius. Default is 50.
     * @param [widthSegments=8] — number of horizontal segments. Minimum value is 3, and the default is 8.
     * @param [heightSegments=6] — number of vertical segments. Minimum value is 2, and the default is 6.
     * @param [phiStart=0] — specify horizontal starting angle. Default is 0.
     * @param [phiLength=Math.PI * 2] — specify horizontal sweep angle size. Default is Math.PI * 2.
     * @param [thetaStart=0] — specify vertical starting angle. Default is 0.
     * @param [thetaLength=Math.PI * 2] — specify vertical sweep angle size. Default is Math.PI.
     */
    constructor(
        radius?: number,
        widthSegments?: number,
        heightSegments?: number,
        phiStart?: number,
        phiLength?: number,
        thetaStart?: number,
        thetaLength?: number,
    );

    /**
     * @default 'SphereGeometry'
     */
    type: string;

    parameters: {
        radius: number;
        widthSegments: number;
        heightSegments: number;
        phiStart: number;
        phiLength: number;
        thetaStart: number;
        thetaLength: number;
    };

    static fromJSON(data: any): SphereGeometry;
}

declare class TetrahedronGeometry extends PolyhedronGeometry {
    /**
     * @param [radius=1]
     * @param [detail=0]
     */
    constructor(radius?: number, detail?: number);

    /**
     * @default 'TetrahedronGeometry'
     */
    type: string;

    static fromJSON(data: any): TetrahedronGeometry;
}

declare class TorusGeometry extends BufferGeometry {
    /**
     * @param [radius=1]
     * @param [tube=0.4]
     * @param [radialSegments=8]
     * @param [tubularSegments=6]
     * @param [arc=Math.PI * 2]
     */
    constructor(radius?: number, tube?: number, radialSegments?: number, tubularSegments?: number, arc?: number);

    /**
     * @default 'TorusGeometry'
     */
    type: string;

    parameters: {
        radius: number;
        tube: number;
        radialSegments: number;
        tubularSegments: number;
        arc: number;
    };

    static fromJSON(data: any): TorusGeometry;
}

declare class TorusKnotGeometry extends BufferGeometry {
    /**
     * @param [radius=1]
     * @param [tube=0.4]
     * @param [radialSegments=64]
     * @param [tubularSegments=8]
     * @param [p=2]
     * @param [q=3]
     */
    constructor(
        radius?: number,
        tube?: number,
        tubularSegments?: number,
        radialSegments?: number,
        p?: number,
        q?: number,
    );

    /**
     * @default 'TorusKnotGeometry'
     */
    type: string;

    parameters: {
        radius: number;
        tube: number;
        tubularSegments: number;
        radialSegments: number;
        p: number;
        q: number;
    };

    static fromJSON(data: any): TorusKnotGeometry;
}

declare class TubeGeometry extends BufferGeometry {
    /**
     * @param path
     * @param [tubularSegments=64]
     * @param [radius=1]
     * @param [radiusSegments=8]
     * @param [closed=false]
     */
    constructor(
        path?: Curve<Vector3>,
        tubularSegments?: number,
        radius?: number,
        radiusSegments?: number,
        closed?: boolean,
    );

    /**
     * @default 'TubeGeometry'
     */
    type: string;

    parameters: {
        path: Curve<Vector3>;
        tubularSegments: number;
        radius: number;
        radialSegments: number;
        closed: boolean;
    };
    tangents: Vector3[];
    normals: Vector3[];
    binormals: Vector3[];

    static fromJSON(data: any): TubeGeometry;
}

declare class WireframeGeometry<TBufferGeometry extends BufferGeometry = BufferGeometry> extends BufferGeometry {
    constructor(geometry?: TBufferGeometry);

    /**
     * @default 'WireframeGeometry'
     */
    type: string;

    parameters: {
        geometry: TBufferGeometry;
    };
}

declare class Line<
    TGeometry extends BufferGeometry = BufferGeometry,
    TMaterial extends Material | Material[] = Material | Material[],
> extends Object3D {
    constructor(geometry?: TGeometry, material?: TMaterial);

    geometry: TGeometry;
    material: TMaterial;

    type: 'Line' | 'LineLoop' | 'LineSegments' | string;
    readonly isLine: true;

    morphTargetInfluences?: number[] | undefined;
    morphTargetDictionary?: { [key: string]: number } | undefined;

    computeLineDistances(): this;
    raycast(raycaster: Raycaster, intersects: Intersection[]): void;
    updateMorphTargets(): void;
}

/**
 * @deprecated
 */
declare const LineStrip: number;
/**
 * @deprecated
 */
declare const LinePieces: number;

declare class LineSegments<
    TGeometry extends BufferGeometry = BufferGeometry,
    TMaterial extends Material | Material[] = Material | Material[],
> extends Line<TGeometry, TMaterial> {
    constructor(geometry?: TGeometry, material?: TMaterial);

    /**
     * @default 'LineSegments'
     */
    type: 'LineSegments' | string;
    readonly isLineSegments: true;
}

declare class SpotLightHelper extends Object3D {
    constructor(light: Light, color?: ColorRepresentation);

    light: Light;
    matrix: Matrix4;

    /**
     * @default false
     */
    matrixAutoUpdate: boolean;
    color: ColorRepresentation | undefined;
    cone: LineSegments;

    dispose(): void;
    update(): void;
}

declare class SkeletonHelper extends LineSegments {
    constructor(object: Object3D);

    /**
     * @default 'SkeletonHelper'
     */
    type: string;

    bones: Bone[];
    root: Object3D;

    readonly isSkeletonHelper: true;

    matrix: Matrix4;

    /**
     * @default false
     */
    matrixAutoUpdate: boolean;

    getBoneList(object: Object3D): Bone[];
    update(): void;
}

declare class PointLightShadow extends LightShadow {
    camera: PerspectiveCamera;
}

/**
 * @example
 * const light = new THREE.PointLight( 0xff0000, 1, 100 );
 * light.position.set( 50, 50, 50 );
 * scene.add( light );
 */
declare class PointLight extends Light {
    constructor(color?: ColorRepresentation, intensity?: number, distance?: number, decay?: number);

    /**
     * @default 'PointLight'
     */
    type: string;

    /**
     * Light's intensity.
     * @default 1
     */
    intensity: number;

    /**
     * If non-zero, light will attenuate linearly from maximum intensity at light position down to zero at distance.
     * @default 0
     */
    distance: number;

    /**
     * @default 1
     */
    decay: number;

    /**
     * @default new THREE.PointLightShadow()
     */
    shadow: PointLightShadow;

    power: number;
}

declare class PointLightHelper extends Object3D {
    constructor(light: PointLight, sphereSize?: number, color?: ColorRepresentation);

    /**
     * @default 'PointLightHelper'
     */
    type: string;

    light: PointLight;
    color: ColorRepresentation | undefined;
    matrix: Matrix4;

    /**
     * @default false
     */
    matrixAutoUpdate: boolean;

    dispose(): void;
    update(): void;
}

declare class HemisphereLight extends Light {
    /**
     * @param skyColor
     * @param groundColor
     * @param [intensity=1]
     */
    constructor(skyColor?: ColorRepresentation, groundColor?: ColorRepresentation, intensity?: number);

    /**
     * @default 'HemisphereLight'
     */
    type: string;

    /**
     * @default THREE.Object3D.DefaultUp
     */
    position: Vector3;

    groundColor: Color;

    readonly isHemisphereLight: true;
}

/**
 * parameters is an object with one or more properties defining the material's appearance.
 */
interface MeshBasicMaterialParameters extends MaterialParameters {
    color?: ColorRepresentation | undefined;
    opacity?: number | undefined;
    map?: Texture | null | undefined;
    lightMap?: Texture | null;
    lightMapIntensity?: number | undefined;
    aoMap?: Texture | null | undefined;
    aoMapIntensity?: number | undefined;
    specularMap?: Texture | null | undefined;
    alphaMap?: Texture | null | undefined;
    fog?: boolean | undefined;
    envMap?: Texture | null | undefined;
    combine?: Combine | undefined;
    reflectivity?: number | undefined;
    refractionRatio?: number | undefined;
    wireframe?: boolean | undefined;
    wireframeLinewidth?: number | undefined;
    wireframeLinecap?: string | undefined;
    wireframeLinejoin?: string | undefined;
}

declare class MeshBasicMaterial extends Material {
    constructor(parameters?: MeshBasicMaterialParameters);

    /**
     * @default 'MeshBasicMaterial'
     */
    type: string;

    /**
     * @default new THREE.Color( 0xffffff )
     */
    color: Color;

    /**
     * @default null
     */
    map: Texture | null;

    /**
     * @default null
     */
    lightMap: Texture | null;

    /**
     * @default 1
     */
    lightMapIntensity: number;

    /**
     * @default null
     */
    aoMap: Texture | null;

    /**
     * @default 1
     */
    aoMapIntensity: number;

    /**
     * @default null
     */
    specularMap: Texture | null;

    /**
     * @default null
     */
    alphaMap: Texture | null;

    /**
     * @default null
     */
    envMap: Texture | null;

    /**
     * @default THREE.MultiplyOperation
     */
    combine: Combine;

    /**
     * @default 1
     */
    reflectivity: number;

    /**
     * @default 0.98
     */
    refractionRatio: number;

    /**
     * @default false
     */
    wireframe: boolean;

    /**
     * @default 1
     */
    wireframeLinewidth: number;

    /**
     * @default 'round'
     */
    wireframeLinecap: string;

    /**
     * @default 'round'
     */
    wireframeLinejoin: string;

    /**
     * Whether the material is affected by fog. Default is true.
     * @default fog
     */
    fog: boolean;

    setValues(parameters: MeshBasicMaterialParameters): void;
}

declare class HemisphereLightHelper extends Object3D {
    constructor(light: HemisphereLight, size: number, color?: ColorRepresentation);

    light: HemisphereLight;
    matrix: Matrix4;
    matrixAutoUpdate: boolean;
    material: MeshBasicMaterial;

    color: ColorRepresentation | undefined;

    dispose(): void;
    update(): void;
}

declare class GridHelper extends LineSegments {
    /**
     * @param [size=10]
     * @param [divisions=10]
     * @param [color1=0x444444]
     * @param [color2=0x888888]
     */
    constructor(size?: number, divisions?: number, color1?: ColorRepresentation, color2?: ColorRepresentation);

    /**
     * @default 'GridHelper'
     */
    type: string;

    /**
     * @deprecated Colors should be specified in the constructor.
     */
    setColors(color1?: ColorRepresentation, color2?: ColorRepresentation): void;
}

declare class PolarGridHelper extends LineSegments {
    /**
     * @param [radius=10]
     * @param [radials=16]
     * @param [circles=8]
     * @param [divisions=64]
     * @param [color1=0x444444]
     * @param [color2=0x888888]
     */
    constructor(
        radius?: number,
        radials?: number,
        circles?: number,
        divisions?: number,
        color1?: ColorRepresentation,
        color2?: ColorRepresentation,
    );

    /**
     * @default 'PolarGridHelper'
     */
    type: string;
}

declare class DirectionalLightShadow extends LightShadow {
    camera: OrthographicCamera;
    readonly isDirectionalLightShadow: true;
}

/**
 * see {@link https://github.com/mrdoob/three.js/blob/master/src/lights/DirectionalLight.js|src/lights/DirectionalLight.js}
 *
 * @example
 * // White directional light at half intensity shining from the top.
 * const directionalLight = new THREE.DirectionalLight( 0xffffff, 0.5 );
 * directionalLight.position.set( 0, 1, 0 );
 * scene.add( directionalLight );
 */
declare class DirectionalLight extends Light {
    constructor(color?: ColorRepresentation, intensity?: number);

    /**
     * @default 'DirectionalLight'
     */
    type: string;

    /**
     * @default THREE.Object3D.DefaultUp
     */
    readonly position: Vector3;

    /**
     * Target used for shadow camera orientation.
     * @default new THREE.Object3D()
     */
    target: Object3D;

    /**
     * Light's intensity.
     * @default 1
     */
    intensity: number;

    /**
     * @default new THREE.DirectionalLightShadow()
     */
    shadow: DirectionalLightShadow;
    readonly isDirectionalLight: true;
}

declare class DirectionalLightHelper extends Object3D {
    /**
     * @param light
     * @param [size=1]
     * @param color
     */
    constructor(light: DirectionalLight, size?: number, color?: ColorRepresentation);

    light: DirectionalLight;
    lightPlane: Line;
    targetLine: Line;

    /**
     * @default undefined
     */
    color: ColorRepresentation | undefined;
    matrix: Matrix4;

    /**
     * @default false
     */
    matrixAutoUpdate: boolean;

    dispose(): void;
    update(): void;
}

declare class CameraHelper extends LineSegments {
    constructor(camera: Camera);

    camera: Camera;
    pointMap: { [id: string]: number[] };

    /**
     * @default 'CameraHelper'
     */
    type: string;

    update(): void;

    dispose(): void;
}

declare class BoxHelper extends LineSegments {
    /**
     * @param object
     * @param [color=0xffff00]
     */
    constructor(object: Object3D, color?: ColorRepresentation);

    /**
     * @default 'BoxHelper'
     */
    type: string;

    update(object?: Object3D): void;

    setFromObject(object: Object3D): this;
}

declare class Box3Helper extends LineSegments {
    /**
     * @param box
     * @param [color=0xffff00]
     */
    constructor(box: Box3, color?: Color);

    /**
     * @default 'Box3Helper'
     */
    type: string;

    box: Box3;
}

declare class PlaneHelper extends LineSegments {
    /**
     * @param plane
     * @param [size=1]
     * @param [hex=0xffff00]
     */
    constructor(plane: Plane, size?: number, hex?: number);

    /**
     * @default 'PlaneHelper'
     */
    type: string;

    plane: Plane;

    /**
     * @default 1
     */
    size: number;

    updateMatrixWorld(force?: boolean): void;
}

declare class Mesh<
    TGeometry extends BufferGeometry = BufferGeometry,
    TMaterial extends Material | Material[] = Material | Material[],
> extends Object3D {
    constructor(geometry?: TGeometry, material?: TMaterial);

    geometry: TGeometry;
    material: TMaterial;
    morphTargetInfluences?: number[] | undefined;
    morphTargetDictionary?: { [key: string]: number } | undefined;
    readonly isMesh: true;
    type: string;

    updateMorphTargets(): void;
    raycast(raycaster: Raycaster, intersects: Intersection[]): void;
}

// Extras / Helpers /////////////////////////////////////////////////////////////////////

declare class ArrowHelper extends Object3D {
    /**
     * @param [dir] Direction from origin. Must be a unit vector.
     * @param [origin] Point at which the arrow starts.
     * @param [length] Length of the arrow.
     * @param [color] Hexadecimal value to define color.
     * @param [headLength] The length of the head of the arrow.
     * @param [headWidth] The width of the head of the arrow.
     */
    constructor(
        dir?: Vector3,
        origin?: Vector3,
        length?: number,
        color?: ColorRepresentation,
        headLength?: number,
        headWidth?: number,
    );

    /**
     * @default 'ArrowHelper'
     */
    type: string;

    /**
     * Contains the line part of the arrowHelper.
     */
    line: Line;

    /**
     * Contains the cone part of the arrowHelper.
     */
    cone: Mesh;

    /**
     * @param dir The desired direction. Must be a unit vector.
     */
    setDirection(dir: Vector3): void;

    /**
     * @param length The desired length.
     * @param [headLength] The length of the head of the arrow.
     * @param [headWidth] The width of the head of the arrow.
     */
    setLength(length: number, headLength?: number, headWidth?: number): void;

    /**
     * @param color The desired color.
     */
    setColor(color: ColorRepresentation): void;
}

declare class AxesHelper extends LineSegments {
    /**
     * @param [size=1]
     */
    constructor(size?: number);

    /**
     * @default 'AxesHelper'
     */
    type: string;

    setColors(xAxisColor: Color, yAxisColor: Color, zAxisColor: Color): this;

    dispose(): void;
}

declare class SpotLightShadow extends LightShadow {
    camera: PerspectiveCamera;
    readonly isSpotLightShadow: true;

    /**
     * @default 1
     */
    focus: number;
}

/**
 * A point light that can cast shadow in one direction.
 */
declare class SpotLight extends Light {
    constructor(
        color?: ColorRepresentation,
        intensity?: number,
        distance?: number,
        angle?: number,
        penumbra?: number,
        decay?: number,
    );

    /**
     * @default 'SpotLight'
     */
    type: string;

    /**
     * @default THREE.Object3D.DefaultUp
     */
    position: Vector3;

    /**
     * Spotlight focus points at target.position.
     * @default new THREE.Object3D()
     */
    target: Object3D;

    /**
     * Light's intensity.
     * @default 1
     */
    intensity: number;

    /**
     * If non-zero, light will attenuate linearly from maximum intensity at light position down to zero at distance.
     * @default 0
     */
    distance: number;

    /**
     * Maximum extent of the spotlight, in radians, from its direction.
     * @default Math.PI / 3.
     */
    angle: number;

    /**
     * @default 1
     */
    decay: number;

    /**
     * @default new THREE.SpotLightShadow()
     */
    shadow: SpotLightShadow;
    power: number;

    /**
     * @default 0
     */
    penumbra: number;

    readonly isSpotLight: true;
}

declare class RectAreaLight extends Light {
    constructor(color?: ColorRepresentation, intensity?: number, width?: number, height?: number);

    /**
     * @default 'RectAreaLight'
     */
    type: string;

    /**
     * @default 10
     */
    width: number;

    /**
     * @default 10
     */
    height: number;

    /**
     * @default 1
     */
    intensity: number;

    power: number;

    readonly isRectAreaLight: true;
}

/**
 * This light's color gets applied to all the objects in the scene globally.
 *
 * @source https://github.com/mrdoob/three.js/blob/master/src/lights/AmbientLight.js
 */
declare class AmbientLight extends Light {
    /**
     * This creates a Ambientlight with a color.
     * @param color Numeric value of the RGB component of the color or a Color instance.
     * @param [intensity=1]
     */
    constructor(color?: ColorRepresentation, intensity?: number);

    /**
     * @default 'AmbientLight'
     */
    type: string;

    readonly isAmbientLight: true;
}

declare class SphericalHarmonics3 {
    constructor();

    /**
     * @default [new THREE.Vector3(), new THREE.Vector3(), new THREE.Vector3(), new THREE.Vector3(),
     * new THREE.Vector3(), new THREE.Vector3(), new THREE.Vector3(), new THREE.Vector3(), new THREE.Vector3()]
     */
    coefficients: Vector3[];
    readonly isSphericalHarmonics3: true;

    set(coefficients: Vector3[]): SphericalHarmonics3;
    zero(): SphericalHarmonics3;
    add(sh: SphericalHarmonics3): SphericalHarmonics3;
    addScaledSH(sh: SphericalHarmonics3, s: number): SphericalHarmonics3;
    scale(s: number): SphericalHarmonics3;
    lerp(sh: SphericalHarmonics3, alpha: number): SphericalHarmonics3;
    equals(sh: SphericalHarmonics3): boolean;
    copy(sh: SphericalHarmonics3): SphericalHarmonics3;
    clone(): this;

    /**
     * Sets the values of this spherical harmonics from the provided array or array-like.
     * @param array the source array or array-like.
     * @param offset (optional) offset into the array. Default is 0.
     */
    fromArray(array: number[] | ArrayLike<number>, offset?: number): this;

    /**
     * Returns an array with the values of this spherical harmonics, or copies them into the provided array.
     * @param array (optional) array to store the spherical harmonics to. If this is not provided, a new array will be created.
     * @param offset (optional) optional offset into the array.
     * @return The created or provided array.
     */
    toArray(array?: number[], offset?: number): number[];

    /**
     * Returns an array with the values of this spherical harmonics, or copies them into the provided array-like.
     * @param array array-like to store the spherical harmonics to.
     * @param offset (optional) optional offset into the array-like.
     * @return The provided array-like.
     */
    toArray(array: ArrayLike<number>, offset?: number): ArrayLike<number>;

    getAt(normal: Vector3, target: Vector3): Vector3;
    getIrradianceAt(normal: Vector3, target: Vector3): Vector3;

    static getBasisAt(normal: Vector3, shBasis: number[]): void;
}

declare class LightProbe extends Light {
    constructor(sh?: SphericalHarmonics3, intensity?: number);

    /**
     * @default 'LightProbe'
     */
    type: string;

    readonly isLightProbe: true;

    /**
     * @default new THREE.SphericalHarmonics3()
     */
    sh: SphericalHarmonics3;

    fromJSON(json: object): LightProbe;
}

declare class AmbientLightProbe extends LightProbe {
    constructor(color?: ColorRepresentation, intensity?: number);

    readonly isAmbientLightProbe: true;
}

declare class HemisphereLightProbe extends LightProbe {
    constructor(skyColor?: ColorRepresentation, groundColor?: ColorRepresentation, intensity?: number);

    readonly isHemisphereLightProbe: true;
}

/**
 * Base class for implementing loaders.
 */
declare class Loader {
    constructor(manager?: LoadingManager);

    /**
     * @default 'anonymous'
     */
    crossOrigin: string;

    /**
     * @default: false
     */
    withCredentials: boolean;

    /**
     * @default ''
     */
    path: string;

    /**
     * @default ''
     */
    resourcePath: string;
    manager: LoadingManager;

    /**
     * @default {}
     */
    requestHeader: { [header: string]: string };

    /*
	load(): void;
	parse(): void;
	*/

    loadAsync(url: string, onProgress?: (event: ProgressEvent) => void): Promise<any>;

    setCrossOrigin(crossOrigin: string): this;
    setWithCredentials(value: boolean): this;
    setPath(path: string): this;
    setResourcePath(resourcePath: string): this;
    setRequestHeader(requestHeader: { [header: string]: string }): this;
}

declare const DefaultLoadingManager: LoadingManager;

/**
 * Handles and keeps track of loaded and pending data.
 */
declare class LoadingManager {
    constructor(
        onLoad?: () => void,
        onProgress?: (url: string, loaded: number, total: number) => void,
        onError?: (url: string) => void,
    );

    /**
     * Will be called when loading of an item starts.
     * @param url The url of the item that started loading.
     * @param loaded The number of items already loaded so far.
     * @param total The total amount of items to be loaded.
     */
    onStart?: ((url: string, loaded: number, total: number) => void) | undefined;

    /**
     * Will be called when all items finish loading.
     * The default is a function with empty body.
     */
    onLoad: () => void;

    /**
     * Will be called for each loaded item.
     * The default is a function with empty body.
     * @param url The url of the item just loaded.
     * @param loaded The number of items already loaded so far.
     * @param total The total amount of items to be loaded.
     */
    onProgress: (url: string, loaded: number, total: number) => void;

    /**
     * Will be called when item loading fails.
     * The default is a function with empty body.
     * @param url The url of the item that errored.
     */
    onError: (url: string) => void;

    /**
     * If provided, the callback will be passed each resource URL before a request is sent.
     * The callback may return the original URL, or a new URL to override loading behavior.
     * This behavior can be used to load assets from .ZIP files, drag-and-drop APIs, and Data URIs.
     * @param callback URL modifier callback. Called with url argument, and must return resolvedURL.
     */
    setURLModifier(callback?: (url: string) => string): this;

    /**
     * Given a URL, uses the URL modifier callback (if any) and returns a resolved URL.
     * If no URL modifier is set, returns the original URL.
     * @param url the url to load
     */
    resolveURL(url: string): string;

    itemStart(url: string): void;
    itemEnd(url: string): void;
    itemError(url: string): void;

    // handlers

    addHandler(regex: RegExp, loader: Loader): this;
    removeHandler(regex: RegExp): this;
    getHandler(file: string): Loader | null;
}

declare class AnimationLoader extends Loader {
    constructor(manager?: LoadingManager);

    load(
        url: string,
        onLoad: (response: AnimationClip[]) => void,
        onProgress?: (request: ProgressEvent) => void,
        onError?: (event: ErrorEvent) => void,
    ): void;
    loadAsync(url: string, onProgress?: (event: ProgressEvent) => void): Promise<AnimationClip[]>;
    parse(json: any): AnimationClip[];
}

declare class CompressedTexture extends Texture {
    /**
     * @param mipmaps
     * @param width
     * @param height
     * @param [format=THREE.RGBAFormat]
     * @param [type=THREE.UnsignedByteType]
     * @param [mapping=THREE.Texture.DEFAULT_MAPPING]
     * @param [wrapS=THREE.ClampToEdgeWrapping]
     * @param [wrapT=THREE.ClampToEdgeWrapping]
     * @param [magFilter=THREE.LinearFilter]
     * @param [minFilter=THREE.LinearMipmapLinearFilter]
     * @param [anisotropy=1]
     * @param [encoding=THREE.LinearEncoding]
     */
    constructor(
        mipmaps: ImageData[],
        width: number,
        height: number,
        format?: CompressedPixelFormat,
        type?: TextureDataType,
        mapping?: Mapping,
        wrapS?: Wrapping,
        wrapT?: Wrapping,
        magFilter?: TextureFilter,
        minFilter?: TextureFilter,
        anisotropy?: number,
        encoding?: TextureEncoding,
    );

    get image(): { width: number; height: number };
    set image(value: { width: number; height: number });

    mipmaps: ImageData[];

    /**
     * @default false
     */
    flipY: boolean;

    /**
     * @default false
     */
    generateMipmaps: boolean;

    readonly isCompressedTexture: true;
}

declare class CompressedTextureLoader extends Loader {
    constructor(manager?: LoadingManager);

    load(
        url: string,
        onLoad: (texture: CompressedTexture) => void,
        onProgress?: (event: ProgressEvent) => void,
        onError?: (event: ErrorEvent) => void,
    ): CompressedTexture;

    loadAsync(url: string, onProgress?: (event: ProgressEvent) => void): Promise<CompressedTexture>;
}

declare class DataTexture extends Texture {
    /**
     * @param data
     * @param width
     * @param height
     * @param [format=THREE.RGBAFormat]
     * @param [type=THREE.UnsignedByteType]
     * @param [mapping=THREE.Texture.DEFAULT_MAPPING]
     * @param [wrapS=THREE.ClampToEdgeWrapping]
     * @param [wrapT=THREE.ClampToEdgeWrapping]
     * @param [magFilter=THREE.NearestFilter]
     * @param [minFilter=THREE.NearestFilter]
     * @param [anisotropy=1]
     * @param [encoding=THREE.LinearEncoding]
     */
    constructor(
        data?: BufferSource | null,
        width?: number,
        height?: number,
        format?: PixelFormat,
        type?: TextureDataType,
        mapping?: Mapping,
        wrapS?: Wrapping,
        wrapT?: Wrapping,
        magFilter?: TextureFilter,
        minFilter?: TextureFilter,
        anisotropy?: number,
        encoding?: TextureEncoding,
    );

    get image(): ImageData;
    set image(value: ImageData);

    /**
     * @default false
     */
    flipY: boolean;

    /**
     * @default false
     */
    generateMipmaps: boolean;

    /**
     * @default 1
     */
    unpackAlignment: number;

    /**
     * @default THREE.DepthFormat
     */
    format: PixelFormat;

    readonly isDataTexture: true;
}

declare class DataTextureLoader extends Loader {
    constructor(manager?: LoadingManager);

    load(
        url: string,
        onLoad: (dataTexture: DataTexture, texData: object) => void,
        onProgress?: (event: ProgressEvent) => void,
        onError?: (event: ErrorEvent) => void,
    ): DataTexture;

    loadAsync(url: string, onProgress?: (event: ProgressEvent) => void): Promise<DataTexture>;
}

declare class CubeTextureLoader extends Loader {
    constructor(manager?: LoadingManager);

    load(
        urls: string[],
        onLoad?: (texture: CubeTexture) => void,
        onProgress?: (event: ProgressEvent) => void,
        onError?: (event: ErrorEvent) => void,
    ): CubeTexture;

    loadAsync(url: string, onProgress?: (event: ProgressEvent) => void): Promise<CubeTexture>;
}

/**
 * Class for loading a texture.
 * Unlike other loaders, this one emits events instead of using predefined callbacks. So if you're interested in getting notified when things happen, you need to add listeners to the object.
 */
declare class TextureLoader extends Loader {
    constructor(manager?: LoadingManager);

    load(
        url: string,
        onLoad?: (texture: Texture) => void,
        onProgress?: (event: ProgressEvent) => void,
        onError?: (event: ErrorEvent) => void,
    ): Texture;

    loadAsync(url: string, onProgress?: (event: ProgressEvent) => void): Promise<Texture>;
}

declare class ObjectLoader extends Loader {
    constructor(manager?: LoadingManager);

    load(
        url: string,
        // tslint:disable-next-line:no-unnecessary-generics
        onLoad?: <ObjectType extends Object3D>(object: ObjectType) => void,
        onProgress?: (event: ProgressEvent) => void,
        onError?: (event: Error | ErrorEvent) => void,
    ): void;
    loadAsync<ObjectType extends Object3D>(
        url: string,
        onProgress?: (event: ProgressEvent) => void,
    ): // tslint:disable-next-line:no-unnecessary-generics
    Promise<ObjectType>;
    // tslint:disable-next-line:no-unnecessary-generics
    parse<T extends Object3D>(json: any, onLoad?: (object: Object3D) => void): T;
    // tslint:disable-next-line:no-unnecessary-generics
    parseAsync<T extends Object3D>(json: any): Promise<T>;
    parseGeometries(json: any): { [key: string]: InstancedBufferGeometry | BufferGeometry }; // Array of BufferGeometry or Geometry or Geometry2.
    parseMaterials(json: any, textures: Texture[]): Material[]; // Array of Classes that inherits from Matrial.
    parseAnimations(json: any): AnimationClip[];
    parseImages(json: any, onLoad: () => void): { [key: string]: HTMLImageElement };
    parseImagesAsync(json: any): Promise<{ [key: string]: HTMLImageElement }>;
    parseTextures(json: any, images: any): Texture[];
    parseObject<T extends Object3D>(
        data: any,
        geometries: any[],
        materials: Material[],
        animations: AnimationClip[],
    ): // tslint:disable-next-line:no-unnecessary-generics
    T;
}

declare class MaterialLoader extends Loader {
    constructor(manager?: LoadingManager);

    /**
     * @default {}
     */
    textures: { [key: string]: Texture };

    load(
        url: string,
        onLoad: (material: Material) => void,
        onProgress?: (event: ProgressEvent) => void,
        onError?: (event: Error | ErrorEvent) => void,
    ): void;
    loadAsync(url: string, onProgress?: (event: ProgressEvent) => void): Promise<Material>;
    setTextures(textures: { [key: string]: Texture }): this;
    parse(json: any): Material;
}

declare class BufferGeometryLoader extends Loader {
    constructor(manager?: LoadingManager);

    load(
        url: string,
        onLoad: (bufferGeometry: InstancedBufferGeometry | BufferGeometry) => void,
        onProgress?: (request: ProgressEvent) => void,
        onError?: (event: ErrorEvent) => void,
    ): void;
    loadAsync(
        url: string,
        onProgress?: (event: ProgressEvent) => void,
    ): Promise<InstancedBufferGeometry | BufferGeometry>;
    parse(json: any): InstancedBufferGeometry | BufferGeometry;
}

/**
 * A loader for loading an image.
 * Unlike other loaders, this one emits events instead of using predefined callbacks. So if you're interested in getting notified when things happen, you need to add listeners to the object.
 */
declare class ImageLoader extends Loader {
    constructor(manager?: LoadingManager);

    load(
        url: string,
        onLoad?: (image: HTMLImageElement) => void,
        onProgress?: (event: ProgressEvent) => void,
        onError?: (event: ErrorEvent) => void,
    ): HTMLImageElement;

    loadAsync(url: string, onProgress?: (event: ProgressEvent) => void): Promise<HTMLImageElement>;
}

declare class ImageBitmapLoader extends Loader {
    constructor(manager?: LoadingManager);

    /**
     * @default { premultiplyAlpha: 'none' }
     */
    options: undefined | object;

    readonly isImageBitmapLoader: true;

    setOptions(options: object): ImageBitmapLoader;
    load(
        url: string,
        onLoad?: (response: ImageBitmap) => void,
        onProgress?: (request: ProgressEvent) => void,
        onError?: (event: ErrorEvent) => void,
    ): any;

    loadAsync(url: string, onProgress?: (event: ProgressEvent) => void): Promise<ImageBitmap>;
}

declare class FileLoader extends Loader {
    constructor(manager?: LoadingManager);

    mimeType: undefined | MimeType;
    responseType: undefined | string;

    load(
        url: string,
        onLoad?: (response: string | ArrayBuffer) => void,
        onProgress?: (request: ProgressEvent) => void,
        onError?: (event: ErrorEvent) => void,
    ): any;
    loadAsync(url: string, onProgress?: (event: ProgressEvent) => void): Promise<string | ArrayBuffer>;
    setMimeType(mimeType: MimeType): FileLoader;
    setResponseType(responseType: string): FileLoader;
}

interface LoaderUtils {
    decodeText(array: BufferSource): string;
    extractUrlBase(url: string): string;
    resolveURL(url: string, path: string): string;
}

declare const LoaderUtils: LoaderUtils;

declare namespace Cache {
    /**
     * @default false
     */
    let enabled: boolean;

    /**
     * @default {}
     */
    let files: any;

    function add(key: string, file: any): void;
    function get(key: string): any;
    function remove(key: string): void;
    function clear(): void;
}

declare class AudioLoader extends Loader {
    constructor(manager?: LoadingManager);

    load(
        url: string,
        onLoad: (audioBuffer: AudioBuffer) => void,
        onProgress?: (request: ProgressEvent) => void,
        onError?: (event: ErrorEvent) => void,
    ): void;

    loadAsync(url: string, onProgress?: (event: ProgressEvent) => void): Promise<AudioBuffer>;
}

interface ShadowMaterialParameters extends MaterialParameters {
    color?: ColorRepresentation | undefined;
    fog?: boolean | undefined;
}

declare class ShadowMaterial extends Material {
    constructor(parameters?: ShadowMaterialParameters);

    /**
     * @default 'ShadowMaterial'
     */
    type: string;

    /**
     * @default new THREE.Color( 0x000000 )
     */
    color: Color;

    /**
     * @default true
     */
    transparent: boolean;

    /**
     * Whether the material is affected by fog. Default is true.
     * @default fog
     */
    fog: boolean;
}

interface SpriteMaterialParameters extends MaterialParameters {
    color?: ColorRepresentation | undefined;
    map?: Texture | null | undefined;
    alphaMap?: Texture | null | undefined;
    rotation?: number | undefined;
    sizeAttenuation?: boolean | undefined;
    fog?: boolean | undefined;
}

declare class SpriteMaterial extends Material {
    constructor(parameters?: SpriteMaterialParameters);
    /**
     * @default 'SpriteMaterial'
     */
    type: string;

    /**
     * @default new THREE.Color( 0xffffff )
     */
    color: Color;

    /**
     * @default null
     */
    map: Texture | null;

    /**
     * @default null
     */
    alphaMap: Texture | null;

    /**
     * @default 0
     */
    rotation: number;

    /**
     * @default true
     */
    sizeAttenuation: boolean;

    /**
     * @default true
     */
    transparent: boolean;

    /**
     * Whether the material is affected by fog. Default is true.
     * @default fog
     */
    fog: boolean;

    readonly isSpriteMaterial: true;

    setValues(parameters: SpriteMaterialParameters): void;
    copy(source: SpriteMaterial): this;
}

interface ShaderMaterialParameters extends MaterialParameters {
    uniforms?: { [uniform: string]: IUniform } | undefined;
    vertexShader?: string | undefined;
    fragmentShader?: string | undefined;
    linewidth?: number | undefined;
    wireframe?: boolean | undefined;
    wireframeLinewidth?: number | undefined;
    lights?: boolean | undefined;
    clipping?: boolean | undefined;

    extensions?:
        | {
              derivatives?: boolean | undefined;
              fragDepth?: boolean | undefined;
              drawBuffers?: boolean | undefined;
              shaderTextureLOD?: boolean | undefined;
          }
        | undefined;
    glslVersion?: GLSLVersion | undefined;
}

declare class ShaderMaterial extends Material {
    constructor(parameters?: ShaderMaterialParameters);

    /**
     * @default 'ShaderMaterial'
     */
    type: string;

    /**
     * @default {}
     */
    defines: { [key: string]: any };

    /**
     * @default {}
     */
    uniforms: { [uniform: string]: IUniform };
    vertexShader: string;
    fragmentShader: string;

    /**
     * @default 1
     */
    linewidth: number;

    /**
     * @default false
     */
    wireframe: boolean;

    /**
     * @default 1
     */
    wireframeLinewidth: number;

    /**
     * @default false
     */
    fog: boolean;

    /**
     * @default false
     */
    lights: boolean;

    /**
     * @default false
     */
    clipping: boolean;

    /**
     * @deprecated Use {@link ShaderMaterial#extensions.derivatives extensions.derivatives} instead.
     */
    derivatives: any;

    /**
     * @default { derivatives: false, fragDepth: false, drawBuffers: false, shaderTextureLOD: false }
     */
    extensions: {
        derivatives: boolean;
        fragDepth: boolean;
        drawBuffers: boolean;
        shaderTextureLOD: boolean;
    };

    /**
     * @default { 'color': [ 1, 1, 1 ], 'uv': [ 0, 0 ], 'uv2': [ 0, 0 ] }
     */
    defaultAttributeValues: any;

    /**
     * @default undefined
     */
    index0AttributeName: string | undefined;

    /**
     * @default false
     */
    uniformsNeedUpdate: boolean;

    /**
     * @default null
     */
    glslVersion: GLSLVersion | null;

    isShaderMaterial: boolean;

    setValues(parameters: ShaderMaterialParameters): void;
    toJSON(meta: any): any;
}

declare class RawShaderMaterial extends ShaderMaterial {
    constructor(parameters?: ShaderMaterialParameters);
}

interface PointsMaterialParameters extends MaterialParameters {
    color?: ColorRepresentation | undefined;
    map?: Texture | null | undefined;
    alphaMap?: Texture | null | undefined;
    size?: number | undefined;
    sizeAttenuation?: boolean | undefined;
    fog?: boolean | undefined;
}

declare class PointsMaterial extends Material {
    constructor(parameters?: PointsMaterialParameters);

    /**
     * @default 'PointsMaterial'
     */
    type: string;

    /**
     * @default new THREE.Color( 0xffffff )
     */
    color: Color;

    /**
     * @default null
     */
    map: Texture | null;

    /**
     * @default null
     */
    alphaMap: Texture | null;

    /**
     * @default 1
     */
    size: number;

    /**
     * @default true
     */
    sizeAttenuation: boolean;

    /**
     * Whether the material is affected by fog. Default is true.
     * @default fog
     */
    fog: boolean;

    setValues(parameters: PointsMaterialParameters): void;
}

interface MeshStandardMaterialParameters extends MaterialParameters {
    color?: ColorRepresentation | undefined;
    roughness?: number | undefined;
    metalness?: number | undefined;
    map?: Texture | null | undefined;
    lightMap?: Texture | null | undefined;
    lightMapIntensity?: number | undefined;
    aoMap?: Texture | null | undefined;
    aoMapIntensity?: number | undefined;
    emissive?: ColorRepresentation | undefined;
    emissiveIntensity?: number | undefined;
    emissiveMap?: Texture | null | undefined;
    bumpMap?: Texture | null | undefined;
    bumpScale?: number | undefined;
    normalMap?: Texture | null | undefined;
    normalMapType?: NormalMapTypes | undefined;
    normalScale?: Vector2 | undefined;
    displacementMap?: Texture | null | undefined;
    displacementScale?: number | undefined;
    displacementBias?: number | undefined;
    roughnessMap?: Texture | null | undefined;
    metalnessMap?: Texture | null | undefined;
    alphaMap?: Texture | null | undefined;
    envMap?: Texture | null | undefined;
    envMapIntensity?: number | undefined;
    wireframe?: boolean | undefined;
    wireframeLinewidth?: number | undefined;
    fog?: boolean | undefined;
    flatShading?: boolean | undefined;
}

declare class MeshStandardMaterial extends Material {
    constructor(parameters?: MeshStandardMaterialParameters);

    /**
     * @default 'MeshStandardMaterial'
     */
    type: string;

    /**
     * @default { 'STANDARD': '' }
     */
    defines: { [key: string]: any };

    /**
     * @default new THREE.Color( 0xffffff )
     */
    color: Color;

    /**
     * @default 1
     */
    roughness: number;

    /**
     * @default 0
     */
    metalness: number;

    /**
     * @default null
     */
    map: Texture | null;

    /**
     * @default null
     */
    lightMap: Texture | null;

    /**
     * @default 1
     */
    lightMapIntensity: number;

    /**
     * @default null
     */
    aoMap: Texture | null;

    /**
     * @default 1
     */
    aoMapIntensity: number;

    /**
     * @default new THREE.Color( 0x000000 )
     */
    emissive: Color;

    /**
     * @default 1
     */
    emissiveIntensity: number;

    /**
     * @default null
     */
    emissiveMap: Texture | null;

    /**
     * @default null
     */
    bumpMap: Texture | null;

    /**
     * @default 1
     */
    bumpScale: number;

    /**
     * @default null
     */
    normalMap: Texture | null;

    /**
     * @default THREE.TangentSpaceNormalMap
     */
    normalMapType: NormalMapTypes;

    /**
     * @default new THREE.Vector2( 1, 1 )
     */
    normalScale: Vector2;

    /**
     * @default null
     */
    displacementMap: Texture | null;

    /**
     * @default 1
     */
    displacementScale: number;

    /**
     * @default 0
     */
    displacementBias: number;

    /**
     * @default null
     */
    roughnessMap: Texture | null;

    /**
     * @default null
     */
    metalnessMap: Texture | null;

    /**
     * @default null
     */
    alphaMap: Texture | null;

    /**
     * @default null
     */
    envMap: Texture | null;

    /**
     * @default 1
     */
    envMapIntensity: number;

    /**
     * @default false
     */
    wireframe: boolean;

    /**
     * @default 1
     */
    wireframeLinewidth: number;

    /**
     * @default 'round'
     */
    wireframeLinecap: string;

    /**
     * @default 'round'
     */
    wireframeLinejoin: string;

    /**
     * Define whether the material is rendered with flat shading. Default is false.
     * @default false
     */
    flatShading: boolean;

    /**
     * Whether the material is affected by fog. Default is true.
     * @default fog
     */
    fog: boolean;

    isMeshStandardMaterial: boolean;

    setValues(parameters: MeshStandardMaterialParameters): void;
}

interface MeshPhysicalMaterialParameters extends MeshStandardMaterialParameters {
    clearcoat?: number | undefined;
    clearcoatMap?: Texture | null | undefined;
    clearcoatRoughness?: number | undefined;
    clearcoatRoughnessMap?: Texture | null | undefined;
    clearcoatNormalScale?: Vector2 | undefined;
    clearcoatNormalMap?: Texture | null | undefined;

    reflectivity?: number | undefined;
    ior?: number | undefined;

    sheen?: number | undefined;
    sheenColor?: Color | undefined;
    sheenRoughness?: number | undefined;

    transmission?: number | undefined;
    transmissionMap?: Texture | null | undefined;
    attenuationDistance?: number | undefined;
    attenuationColor?: Color | undefined;

    specularIntensity?: number | undefined;
    specularColor?: Color | undefined;
    specularIntensityMap?: Texture | null | undefined;
    specularColorMap?: Texture | null | undefined;
}

declare class MeshPhysicalMaterial extends MeshStandardMaterial {
    constructor(parameters?: MeshPhysicalMaterialParameters);

    /**
     * @default 'MeshPhysicalMaterial'
     */
    type: string;

    /**
     * @default { 'STANDARD': '', 'PHYSICAL': '' }
     */
    defines: { [key: string]: any };

    /**
     * @default 0
     */
    clearcoat: number;

    /**
     * @default null
     */
    clearcoatMap: Texture | null;

    /**
     * @default 0
     */
    clearcoatRoughness: number;

    /**
     * @default null
     */
    clearcoatRoughnessMap: Texture | null;

    /**
     * @default new THREE.Vector2( 1, 1 )
     */
    clearcoatNormalScale: Vector2;

    /**
     * @default null
     */
    clearcoatNormalMap: Texture | null;

    /**
     * @default 0.5
     */
    reflectivity: number;

    /**
     * @default 1.5
     */
    ior: number;

    /**
     * @default 0.0
     */
    sheen: number;

    /**
     * @default Color( 0x000000 )
     */
    sheenColor: Color;

    /**
     * @default null
     */
    sheenColorMap: Texture | null;

    /**
     * @default 1.0
     */
    sheenRoughness: number;

    /**
     * @default null
     */
    sheenRoughnessMap: Texture | null;

    /**
     * @default 0
     */
    transmission: number;

    /**
     * @default null
     */
    transmissionMap: Texture | null;

    /**
     * @default 0.01
     */
    thickness: number;

    /**
     * @default null
     */
    thicknessMap: Texture | null;

    /**
     * @default 0.0
     */
    attenuationDistance: number;

    /**
     * @default Color( 1, 1, 1 )
     */
    attenuationColor: Color;

    /**
     * @default 1.0
     */
    specularIntensity: number;

    /**
     * @default Color(1, 1, 1)
     */
    specularColor: Color;

    /**
     * @default null
     */
    specularIntensityMap: Texture | null;

    /**
     * @default null
     */
    specularColorMap: Texture | null;

    /**
     * @default null
     */
    iridescenceMap: Texture | null;

    /**
     * @default 1.3
     */
    iridescenceIOR: number;

    /**
     * @default [100, 400]
     */
    iridescenceThicknessRange: number[];

    /**
     * @default null
     */
    iridescenceThicknessMap: Texture | null;
}

interface MeshPhongMaterialParameters extends MaterialParameters {
    /** geometry color in hexadecimal. Default is 0xffffff. */
    color?: ColorRepresentation | undefined;
    specular?: ColorRepresentation | undefined;
    shininess?: number | undefined;
    opacity?: number | undefined;
    map?: Texture | null | undefined;
    lightMap?: Texture | null | undefined;
    lightMapIntensity?: number | undefined;
    aoMap?: Texture | null | undefined;
    aoMapIntensity?: number | undefined;
    emissive?: ColorRepresentation | undefined;
    emissiveIntensity?: number | undefined;
    emissiveMap?: Texture | null | undefined;
    bumpMap?: Texture | null | undefined;
    bumpScale?: number | undefined;
    normalMap?: Texture | null | undefined;
    normalMapType?: NormalMapTypes | undefined;
    normalScale?: Vector2 | undefined;
    displacementMap?: Texture | null | undefined;
    displacementScale?: number | undefined;
    displacementBias?: number | undefined;
    specularMap?: Texture | null | undefined;
    alphaMap?: Texture | null | undefined;
    envMap?: Texture | null | undefined;
    combine?: Combine | undefined;
    reflectivity?: number | undefined;
    refractionRatio?: number | undefined;
    wireframe?: boolean | undefined;
    wireframeLinewidth?: number | undefined;
    wireframeLinecap?: string | undefined;
    wireframeLinejoin?: string | undefined;
    fog?: boolean | undefined;
    flatShading?: boolean | undefined;
}

declare class MeshPhongMaterial extends Material {
    constructor(parameters?: MeshPhongMaterialParameters);

    /**
     * @default 'MeshNormalMaterial'
     */
    type: string;

    /**
     * @default new THREE.Color( 0xffffff )
     */
    color: Color;

    /**
     * @default new THREE.Color( 0x111111 )
     */
    specular: Color;

    /**
     * @default 30
     */
    shininess: number;

    /**
     * @default null
     */
    map: Texture | null;

    /**
     * @default null
     */
    lightMap: Texture | null;

    /**
     * @default null
     */
    lightMapIntensity: number;

    /**
     * @default null
     */
    aoMap: Texture | null;

    /**
     * @default null
     */
    aoMapIntensity: number;

    /**
     * @default new THREE.Color( 0x000000 )
     */
    emissive: Color;

    /**
     * @default 1
     */
    emissiveIntensity: number;

    /**
     * @default null
     */
    emissiveMap: Texture | null;

    /**
     * @default null
     */
    bumpMap: Texture | null;

    /**
     * @default 1
     */
    bumpScale: number;

    /**
     * @default null
     */
    normalMap: Texture | null;

    /**
     * @default THREE.TangentSpaceNormalMap
     */
    normalMapType: NormalMapTypes;

    /**
     * @default new Vector2( 1, 1 )
     */
    normalScale: Vector2;

    /**
     * @default null
     */
    displacementMap: Texture | null;

    /**
     * @default 1
     */
    displacementScale: number;

    /**
     * @default 0
     */
    displacementBias: number;

    /**
     * @default null
     */
    specularMap: Texture | null;

    /**
     * @default null
     */
    alphaMap: Texture | null;

    /**
     * @default null
     */
    envMap: Texture | null;

    /**
     * @default THREE.MultiplyOperation
     */
    combine: Combine;

    /**
     * @default 1
     */
    reflectivity: number;

    /**
     * @default 0.98
     */
    refractionRatio: number;

    /**
     * @default false
     */
    wireframe: boolean;

    /**
     * @default 1
     */
    wireframeLinewidth: number;

    /**
     * @default 'round'
     */
    wireframeLinecap: string;

    /**
     * @default 'round'
     */
    wireframeLinejoin: string;

    /**
     * Define whether the material is rendered with flat shading. Default is false.
     * @default false
     */
    flatShading: boolean;

    /**
     * @deprecated Use {@link MeshStandardMaterial THREE.MeshStandardMaterial} instead.
     */
    metal: boolean;

    /**
     * Whether the material is affected by fog. Default is true.
     * @default fog
     */
    fog: boolean;

    setValues(parameters: MeshPhongMaterialParameters): void;
}

interface MeshToonMaterialParameters extends MaterialParameters {
    /** geometry color in hexadecimal. Default is 0xffffff. */
    color?: ColorRepresentation | undefined;
    opacity?: number | undefined;
    gradientMap?: Texture | null | undefined;
    map?: Texture | null | undefined;
    lightMap?: Texture | null | undefined;
    lightMapIntensity?: number | undefined;
    aoMap?: Texture | null | undefined;
    aoMapIntensity?: number | undefined;
    emissive?: ColorRepresentation | undefined;
    emissiveIntensity?: number | undefined;
    emissiveMap?: Texture | null | undefined;
    bumpMap?: Texture | null | undefined;
    bumpScale?: number | undefined;
    normalMap?: Texture | null | undefined;
    normalMapType?: NormalMapTypes | undefined;
    normalScale?: Vector2 | undefined;
    displacementMap?: Texture | null | undefined;
    displacementScale?: number | undefined;
    displacementBias?: number | undefined;
    alphaMap?: Texture | null | undefined;
    wireframe?: boolean | undefined;
    wireframeLinewidth?: number | undefined;
    wireframeLinecap?: string | undefined;
    wireframeLinejoin?: string | undefined;
    fog?: boolean | undefined;
}

declare class MeshToonMaterial extends Material {
    constructor(parameters?: MeshToonMaterialParameters);

    /**
     * @default 'MeshToonMaterial'
     */
    type: string;

    /**
     * @default { 'TOON': '' }
     */
    defines: { [key: string]: any };

    /**
     * @default new THREE.Color( 0xffffff )
     */
    color: Color;

    /**
     * @default null
     */
    gradientMap: Texture | null;

    /**
     * @default null
     */
    map: Texture | null;

    /**
     * @default null
     */
    lightMap: Texture | null;

    /**
     * @default 1
     */
    lightMapIntensity: number;

    /**
     * @default null
     */
    aoMap: Texture | null;

    /**
     * @default 1
     */
    aoMapIntensity: number;

    /**
     * @default new THREE.Color( 0x000000 )
     */
    emissive: Color;

    /**
     * @default 1
     */
    emissiveIntensity: number;

    /**
     * @default null
     */
    emissiveMap: Texture | null;

    /**
     * @default null
     */
    bumpMap: Texture | null;

    /**
     * @default 1
     */
    bumpScale: number;

    /**
     * @default null
     */
    normalMap: Texture | null;

    /**
     * @default THREE.TangentSpaceNormalMap
     */
    normalMapType: NormalMapTypes;

    /**
     * @default new THREE.Vector2( 1, 1 )
     */
    normalScale: Vector2;

    /**
     * @default null
     */
    displacementMap: Texture | null;

    /**
     * @default 1
     */
    displacementScale: number;

    /**
     * @default 0
     */
    displacementBias: number;

    /**
     * @default null
     */
    alphaMap: Texture | null;

    /**
     * @default false
     */
    wireframe: boolean;

    /**
     * @default 1
     */
    wireframeLinewidth: number;

    /**
     * @default 'round'
     */
    wireframeLinecap: string;

    /**
     * @default 'round'
     */
    wireframeLinejoin: string;

    /**
     * Whether the material is affected by fog. Default is true.
     * @default fog
     */
    fog: boolean;

    setValues(parameters: MeshToonMaterialParameters): void;
}

interface MeshNormalMaterialParameters extends MaterialParameters {
    bumpMap?: Texture | null | undefined;
    bumpScale?: number | undefined;
    normalMap?: Texture | null | undefined;
    normalMapType?: NormalMapTypes | undefined;
    normalScale?: Vector2 | undefined;
    displacementMap?: Texture | null | undefined;
    displacementScale?: number | undefined;
    displacementBias?: number | undefined;
    wireframe?: boolean | undefined;
    wireframeLinewidth?: number | undefined;

    flatShading?: boolean | undefined;
}

declare class MeshNormalMaterial extends Material {
    constructor(parameters?: MeshNormalMaterialParameters);

    /**
     * @default 'MeshNormalMaterial'
     */
    type: string;

    /**
     * @default null
     */
    bumpMap: Texture | null;

    /**
     * @default 1
     */
    bumpScale: number;

    /**
     * @default null
     */
    normalMap: Texture | null;

    /**
     * @default THREE.TangentSpaceNormalMap
     */
    normalMapType: NormalMapTypes;

    /**
     * @default new THREE.Vector2( 1, 1 )
     */
    normalScale: Vector2;

    /**
     * @default null
     */
    displacementMap: Texture | null;

    /**
     * @default 1
     */
    displacementScale: number;

    /**
     * @default 0
     */
    displacementBias: number;

    /**
     * @default false
     */
    wireframe: boolean;

    /**
     * @default 1
     */
    wireframeLinewidth: number;

    /**
     * Define whether the material is rendered with flat shading. Default is false.
     * @default false
     */
    flatShading: boolean;

    setValues(parameters: MeshNormalMaterialParameters): void;
}

interface MeshLambertMaterialParameters extends MaterialParameters {
    color?: ColorRepresentation | undefined;
    emissive?: ColorRepresentation | undefined;
    emissiveIntensity?: number | undefined;
    emissiveMap?: Texture | null | undefined;
    map?: Texture | null | undefined;
    lightMap?: Texture | null | undefined;
    lightMapIntensity?: number | undefined;
    aoMap?: Texture | null | undefined;
    aoMapIntensity?: number | undefined;
    specularMap?: Texture | null | undefined;
    alphaMap?: Texture | null | undefined;
    envMap?: Texture | null | undefined;
    combine?: Combine | undefined;
    reflectivity?: number | undefined;
    refractionRatio?: number | undefined;
    wireframe?: boolean | undefined;
    wireframeLinewidth?: number | undefined;
    wireframeLinecap?: string | undefined;
    wireframeLinejoin?: string | undefined;
    fog?: boolean | undefined;
}

declare class MeshLambertMaterial extends Material {
    constructor(parameters?: MeshLambertMaterialParameters);

    /**
     * @default 'MeshLambertMaterial'
     */
    type: string;

    /**
     * @default new THREE.Color( 0xffffff )
     */
    color: Color;

    /**
     * @default new THREE.Color( 0x000000 )
     */
    emissive: Color;

    /**
     * @default 1
     */
    emissiveIntensity: number;

    /**
     * @default null
     */
    emissiveMap: Texture | null;

    /**
     * @default null
     */
    map: Texture | null;

    /**
     * @default null
     */
    lightMap: Texture | null;

    /**
     * @default 1
     */
    lightMapIntensity: number;

    /**
     * @default null
     */
    aoMap: Texture | null;

    /**
     * @default 1
     */
    aoMapIntensity: number;

    /**
     * @default null
     */
    specularMap: Texture | null;

    /**
     * @default null
     */
    alphaMap: Texture | null;

    /**
     * @default null
     */
    envMap: Texture | null;

    /**
     * @default THREE.MultiplyOperation
     */
    combine: Combine;

    /**
     * @default 1
     */
    reflectivity: number;

    /**
     * @default 0.98
     */
    refractionRatio: number;

    /**
     * @default false
     */
    wireframe: boolean;

    /**
     * @default 1
     */
    wireframeLinewidth: number;

    /**
     * @default 'round'
     */
    wireframeLinecap: string;

    /**
     * @default 'round'
     */
    wireframeLinejoin: string;

    /**
     * Whether the material is affected by fog. Default is true.
     * @default fog
     */
    fog: boolean;

    setValues(parameters: MeshLambertMaterialParameters): void;
}

interface MeshDepthMaterialParameters extends MaterialParameters {
    map?: Texture | null | undefined;
    alphaMap?: Texture | null | undefined;
    depthPacking?: DepthPackingStrategies | undefined;
    displacementMap?: Texture | null | undefined;
    displacementScale?: number | undefined;
    displacementBias?: number | undefined;
    wireframe?: boolean | undefined;
    wireframeLinewidth?: number | undefined;
}

declare class MeshDepthMaterial extends Material {
    constructor(parameters?: MeshDepthMaterialParameters);

    /**
     * @default 'MeshDepthMaterial'
     */
    type: string;

    /**
     * @default null
     */
    map: Texture | null;

    /**
     * @default null
     */
    alphaMap: Texture | null;

    /**
     * @default THREE.BasicDepthPacking
     */
    depthPacking: DepthPackingStrategies;

    /**
     * @default null
     */
    displacementMap: Texture | null;

    /**
     * @default 1
     */
    displacementScale: number;

    /**
     * @default 0
     */
    displacementBias: number;

    /**
     * @default false
     */
    wireframe: boolean;

    /**
     * @default 1
     */
    wireframeLinewidth: number;

    /**
     * @default false
     */
    fog: boolean;

    setValues(parameters: MeshDepthMaterialParameters): void;
}

interface MeshDistanceMaterialParameters extends MaterialParameters {
    map?: Texture | null | undefined;
    alphaMap?: Texture | null | undefined;
    displacementMap?: Texture | null | undefined;
    displacementScale?: number | undefined;
    displacementBias?: number | undefined;
    farDistance?: number | undefined;
    nearDistance?: number | undefined;
    referencePosition?: Vector3 | undefined;
}

declare class MeshDistanceMaterial extends Material {
    constructor(parameters?: MeshDistanceMaterialParameters);

    /**
     * @default 'MeshDistanceMaterial'
     */
    type: string;

    /**
     * @default null
     */
    map: Texture | null;

    /**
     * @default null
     */
    alphaMap: Texture | null;

    /**
     * @default null
     */
    displacementMap: Texture | null;

    /**
     * @default 1
     */
    displacementScale: number;

    /**
     * @default 0
     */
    displacementBias: number;

    /**
     * @default 1000
     */
    farDistance: number;

    /**
     * @default 1
     */
    nearDistance: number;

    /**
     * @default new THREE.Vector3()
     */
    referencePosition: Vector3;

    /**
     * @default false
     */
    fog: boolean;

    setValues(parameters: MeshDistanceMaterialParameters): void;
}

interface MeshMatcapMaterialParameters extends MaterialParameters {
    color?: ColorRepresentation | undefined;
    matcap?: Texture | null | undefined;
    map?: Texture | null | undefined;
    bumpMap?: Texture | null | undefined;
    bumpScale?: number | undefined;
    normalMap?: Texture | null | undefined;
    normalMapType?: NormalMapTypes | undefined;
    normalScale?: Vector2 | undefined;
    displacementMap?: Texture | null | undefined;
    displacementScale?: number | undefined;
    displacementBias?: number | undefined;
    alphaMap?: Texture | null | undefined;
    fog?: boolean | undefined;
    flatShading?: boolean | undefined;
}

declare class MeshMatcapMaterial extends Material {
    constructor(parameters?: MeshMatcapMaterialParameters);

    /**
     * @default 'MeshMatcapMaterial'
     */
    type: string;

    /**
     * @default { 'MATCAP': '' }
     */
    defines: { [key: string]: any };

    /**
     * @default new THREE.Color( 0xffffff )
     */
    color: Color;

    /**
     * @default null
     */
    matcap: Texture | null;

    /**
     * @default null
     */
    map: Texture | null;

    /**
     * @default null
     */
    bumpMap: Texture | null;

    /**
     * @default 1
     */
    bumpScale: number;

    /**
     * @default null
     */
    normalMap: Texture | null;

    /**
     * @default THREE.TangentSpaceNormalMap
     */
    normalMapType: NormalMapTypes;

    /**
     * @default new Vector2( 1, 1 )
     */
    normalScale: Vector2;

    /**
     * @default null
     */
    displacementMap: Texture | null;

    /**
     * @default 1
     */
    displacementScale: number;

    /**
     * @default 0
     */
    displacementBias: number;

    /**
     * @default null
     */
    alphaMap: Texture | null;

    /**
     * Define whether the material is rendered with flat shading. Default is false.
     * @default false
     */
    flatShading: boolean;

    /**
     * Whether the material is affected by fog. Default is true.
     * @default fog
     */
    fog: boolean;

    setValues(parameters: MeshMatcapMaterialParameters): void;
}

interface LineBasicMaterialParameters extends MaterialParameters {
    color?: ColorRepresentation | undefined;
    linewidth?: number | undefined;
    linecap?: string | undefined;
    linejoin?: string | undefined;
}

declare class LineBasicMaterial extends Material {
    constructor(parameters?: LineBasicMaterialParameters);

    /**
     * @default 'LineBasicMaterial'
     */
    type: string;

    /**
     * @default 0xffffff
     */
    color: Color;

    /**
     * @default 1
     */
    linewidth: number;

    /**
     * @default 'round'
     */
    linecap: string;

    /**
     * @default 'round'
     */
    linejoin: string;

    setValues(parameters: LineBasicMaterialParameters): void;
}

interface LineDashedMaterialParameters extends LineBasicMaterialParameters {
    scale?: number | undefined;
    dashSize?: number | undefined;
    gapSize?: number | undefined;
}

declare class LineDashedMaterial extends LineBasicMaterial {
    constructor(parameters?: LineDashedMaterialParameters);

    /**
     * @default 'LineDashedMaterial'
     */
    type: string;

    /**
     * @default 1
     */
    scale: number;

    /**
     * @default 1
     */
    dashSize: number;

    /**
     * @default 1
     */
    gapSize: number;
    readonly isLineDashedMaterial: true;

    setValues(parameters: LineDashedMaterialParameters): void;
}

declare class QuaternionLinearInterpolant extends Interpolant {
    constructor(parameterPositions: any, samplesValues: any, sampleSize: number, resultBuffer?: any);

    interpolate_(i1: number, t0: number, t: number, t1: number): any;
}

declare class Sprite extends Object3D {
    constructor(material?: SpriteMaterial);

    type: 'Sprite';
    readonly isSprite: true;

    geometry: BufferGeometry;
    material: SpriteMaterial;
    center: Vector2;

    raycast(raycaster: Raycaster, intersects: Intersection[]): void;
    copy(source: this): this;
}

/**
 * Frustums are used to determine what is inside the camera's field of view. They help speed up the rendering process.
 */
declare class Frustum {
    constructor(p0?: Plane, p1?: Plane, p2?: Plane, p3?: Plane, p4?: Plane, p5?: Plane);

    /**
     * Array of 6 vectors.
     */
    planes: Plane[];

    set(p0: Plane, p1: Plane, p2: Plane, p3: Plane, p4: Plane, p5: Plane): Frustum;
    clone(): this;
    copy(frustum: Frustum): this;
    setFromProjectionMatrix(m: Matrix4): this;
    intersectsObject(object: Object3D): boolean;
    intersectsSprite(sprite: Sprite): boolean;
    intersectsSphere(sphere: Sphere): boolean;
    intersectsBox(box: Box3): boolean;
    containsPoint(point: Vector3): boolean;
}

// Math //////////////////////////////////////////////////////////////////////////////////

declare class Box2 {
    constructor(min?: Vector2, max?: Vector2);

    /**
     * @default new THREE.Vector2( + Infinity, + Infinity )
     */
    min: Vector2;

    /**
     * @default new THREE.Vector2( - Infinity, - Infinity )
     */
    max: Vector2;

    set(min: Vector2, max: Vector2): Box2;
    setFromPoints(points: Vector2[]): Box2;
    setFromCenterAndSize(center: Vector2, size: Vector2): Box2;
    clone(): this;
    copy(box: Box2): this;
    makeEmpty(): Box2;
    isEmpty(): boolean;
    getCenter(target: Vector2): Vector2;
    getSize(target: Vector2): Vector2;
    expandByPoint(point: Vector2): Box2;
    expandByVector(vector: Vector2): Box2;
    expandByScalar(scalar: number): Box2;
    containsPoint(point: Vector2): boolean;
    containsBox(box: Box2): boolean;
    getParameter(point: Vector2, target: Vector2): Vector2;
    intersectsBox(box: Box2): boolean;
    clampPoint(point: Vector2, target: Vector2): Vector2;
    distanceToPoint(point: Vector2): number;
    intersect(box: Box2): Box2;
    union(box: Box2): Box2;
    translate(offset: Vector2): Box2;
    equals(box: Box2): boolean;
    /**
     * @deprecated Use {@link Box2#isEmpty .isEmpty()} instead.
     */
    empty(): any;
    /**
     * @deprecated Use {@link Box2#intersectsBox .intersectsBox()} instead.
     */
    isIntersectionBox(b: any): any;
}

/**
 * @see {@link https://github.com/mrdoob/three.js/blob/master/src/math/MathUtils.js|src/math/MathUtils.js}
 */

declare const DEG2RAD: number;
declare const RAD2DEG: number;

declare function generateUUID(): string;

/**
 * Clamps the x to be between a and b.
 *
 * @param value Value to be clamped.
 * @param min Minimum value
 * @param max Maximum value.
 */
declare function clamp(value: number, min: number, max: number): number;
declare function euclideanModulo(n: number, m: number): number;

/**
 * Linear mapping of x from range [a1, a2] to range [b1, b2].
 *
 * @param x Value to be mapped.
 * @param a1 Minimum value for range A.
 * @param a2 Maximum value for range A.
 * @param b1 Minimum value for range B.
 * @param b2 Maximum value for range B.
 */
declare function mapLinear(x: number, a1: number, a2: number, b1: number, b2: number): number;

declare function smoothstep(x: number, min: number, max: number): number;

declare function smootherstep(x: number, min: number, max: number): number;

/**
 * Random float from 0 to 1 with 16 bits of randomness.
 * Standard Math.random() creates repetitive patterns when applied over larger space.
 *
 * @deprecated Use {@link Math#random Math.random()}
 */
declare function random16(): number;

/**
 * Random integer from low to high interval.
 */
declare function randInt(low: number, high: number): number;

/**
 * Random float from low to high interval.
 */
declare function randFloat(low: number, high: number): number;

/**
 * Random float from - range / 2 to range / 2 interval.
 */
declare function randFloatSpread(range: number): number;

/**
 * Deterministic pseudo-random float in the interval [ 0, 1 ].
 */
declare function seededRandom(seed?: number): number;

declare function degToRad(degrees: number): number;

declare function radToDeg(radians: number): number;

declare function isPowerOfTwo(value: number): boolean;

declare function inverseLerp(x: number, y: number, t: number): number;

/**
 * Returns a value linearly interpolated from two known points based
 * on the given interval - t = 0 will return x and t = 1 will return y.
 *
 * @param x Start point.
 * @param y End point.
 * @param t interpolation factor in the closed interval [0, 1]
 */
declare function lerp(x: number, y: number, t: number): number;

/**
 * Smoothly interpolate a number from x toward y in a spring-like
 * manner using the dt to maintain frame rate independent movement.
 *
 * @param x Current point.
 * @param y Target point.
 * @param lambda A higher lambda value will make the movement more sudden, and a lower value will make the movement more gradual.
 * @param dt Delta time in seconds.
 */
declare function damp(x: number, y: number, lambda: number, dt: number): number;

/**
 * Returns a value that alternates between 0 and length.
 *
 * @param x The value to pingpong.
 * @param length The positive value the export function will pingpong to. Default is 1.
 */
declare function pingpong(x: number, length?: number): number;

/**
 * @deprecated Use {@link Math#floorPowerOfTwo .floorPowerOfTwo()}
 */
declare function nearestPowerOfTwo(value: number): number;

/**
 * @deprecated Use {@link Math#ceilPowerOfTwo .ceilPowerOfTwo()}
 */
declare function nextPowerOfTwo(value: number): number;

declare function floorPowerOfTwo(value: number): number;

declare function ceilPowerOfTwo(value: number): number;

declare function setQuaternionFromProperEuler(q: Quaternion, a: number, b: number, c: number, order: string): void;

declare const MathUtils_d_DEG2RAD: typeof DEG2RAD;
declare const MathUtils_d_RAD2DEG: typeof RAD2DEG;
declare const MathUtils_d_generateUUID: typeof generateUUID;
declare const MathUtils_d_clamp: typeof clamp;
declare const MathUtils_d_euclideanModulo: typeof euclideanModulo;
declare const MathUtils_d_mapLinear: typeof mapLinear;
declare const MathUtils_d_smoothstep: typeof smoothstep;
declare const MathUtils_d_smootherstep: typeof smootherstep;
declare const MathUtils_d_random16: typeof random16;
declare const MathUtils_d_randInt: typeof randInt;
declare const MathUtils_d_randFloat: typeof randFloat;
declare const MathUtils_d_randFloatSpread: typeof randFloatSpread;
declare const MathUtils_d_seededRandom: typeof seededRandom;
declare const MathUtils_d_degToRad: typeof degToRad;
declare const MathUtils_d_radToDeg: typeof radToDeg;
declare const MathUtils_d_isPowerOfTwo: typeof isPowerOfTwo;
declare const MathUtils_d_inverseLerp: typeof inverseLerp;
declare const MathUtils_d_lerp: typeof lerp;
declare const MathUtils_d_damp: typeof damp;
declare const MathUtils_d_pingpong: typeof pingpong;
declare const MathUtils_d_nearestPowerOfTwo: typeof nearestPowerOfTwo;
declare const MathUtils_d_nextPowerOfTwo: typeof nextPowerOfTwo;
declare const MathUtils_d_floorPowerOfTwo: typeof floorPowerOfTwo;
declare const MathUtils_d_ceilPowerOfTwo: typeof ceilPowerOfTwo;
declare const MathUtils_d_setQuaternionFromProperEuler: typeof setQuaternionFromProperEuler;
declare namespace MathUtils_d {
  export {
    MathUtils_d_DEG2RAD as DEG2RAD,
    MathUtils_d_RAD2DEG as RAD2DEG,
    MathUtils_d_generateUUID as generateUUID,
    MathUtils_d_clamp as clamp,
    MathUtils_d_euclideanModulo as euclideanModulo,
    MathUtils_d_mapLinear as mapLinear,
    MathUtils_d_smoothstep as smoothstep,
    MathUtils_d_smootherstep as smootherstep,
    MathUtils_d_random16 as random16,
    MathUtils_d_randInt as randInt,
    MathUtils_d_randFloat as randFloat,
    MathUtils_d_randFloatSpread as randFloatSpread,
    MathUtils_d_seededRandom as seededRandom,
    MathUtils_d_degToRad as degToRad,
    MathUtils_d_radToDeg as radToDeg,
    MathUtils_d_isPowerOfTwo as isPowerOfTwo,
    MathUtils_d_inverseLerp as inverseLerp,
    MathUtils_d_lerp as lerp,
    MathUtils_d_damp as damp,
    MathUtils_d_pingpong as pingpong,
    MathUtils_d_nearestPowerOfTwo as nearestPowerOfTwo,
    MathUtils_d_nextPowerOfTwo as nextPowerOfTwo,
    MathUtils_d_floorPowerOfTwo as floorPowerOfTwo,
    MathUtils_d_ceilPowerOfTwo as ceilPowerOfTwo,
    MathUtils_d_setQuaternionFromProperEuler as setQuaternionFromProperEuler,
  };
}

declare class LOD extends Object3D {
    constructor();

    type: 'LOD';

    levels: Array<{ distance: number; object: Object3D }>;
    autoUpdate: boolean;
    readonly isLOD: true;

    addLevel(object: Object3D, distance?: number): this;
    getCurrentLevel(): number;
    getObjectForDistance(distance: number): Object3D | null;
    raycast(raycaster: Raycaster, intersects: Intersection[]): void;
    update(camera: Camera): void;
    toJSON(meta: any): any;

    /**
     * @deprecated Use {@link LOD#levels .levels} instead.
     */
    objects: any[];
}

declare class InstancedMesh<
    TGeometry extends BufferGeometry = BufferGeometry,
    TMaterial extends Material | Material[] = Material | Material[],
> extends Mesh<TGeometry, TMaterial> {
    constructor(geometry: TGeometry | undefined, material: TMaterial | undefined, count: number);

    count: number;
    instanceColor: null | InstancedBufferAttribute;
    instanceMatrix: InstancedBufferAttribute;
    readonly isInstancedMesh: true;

    getColorAt(index: number, color: Color): void;
    getMatrixAt(index: number, matrix: Matrix4): void;
    setColorAt(index: number, color: Color): void;
    setMatrixAt(index: number, matrix: Matrix4): void;
    dispose(): void;
}

declare class Skeleton {
    constructor(bones: Bone[], boneInverses?: Matrix4[]);

    uuid: string;
    bones: Bone[];
    boneInverses: Matrix4[];
    boneMatrices: Float32Array;
    boneTexture: null | DataTexture;
    boneTextureSize: number;
    frame: number;

    init(): void;
    calculateInverses(): void;
    computeBoneTexture(): this;
    pose(): void;
    update(): void;
    clone(): Skeleton;
    getBoneByName(name: string): undefined | Bone;
    dispose(): void;

    /**
     * @deprecated This property has been removed completely.
     */
    useVertexTexture: boolean;
}

declare class SkinnedMesh<
    TGeometry extends BufferGeometry = BufferGeometry,
    TMaterial extends Material | Material[] = Material | Material[],
> extends Mesh<TGeometry, TMaterial> {
    constructor(geometry?: TGeometry, material?: TMaterial, useVertexTexture?: boolean);

    bindMode: string;
    bindMatrix: Matrix4;
    bindMatrixInverse: Matrix4;
    skeleton: Skeleton;
    readonly isSkinnedMesh: true;

    bind(skeleton: Skeleton, bindMatrix?: Matrix4): void;
    pose(): void;
    normalizeSkinWeights(): void;
    updateMatrixWorld(force?: boolean): void;
    boneTransform(index: number, target: Vector3): Vector3;
}

declare class LineLoop<
    TGeometry extends BufferGeometry = BufferGeometry,
    TMaterial extends Material | Material[] = Material | Material[],
> extends Line<TGeometry, TMaterial> {
    constructor(geometry?: TGeometry, material?: TMaterial);

    type: 'LineLoop';
    readonly isLineLoop: true;
}

/**
 * A class for displaying points. The points are rendered by the WebGLRenderer using gl.POINTS.
 */
declare class Points<
    TGeometry extends BufferGeometry = BufferGeometry,
    TMaterial extends Material | Material[] = Material | Material[],
> extends Object3D {
    /**
     * @param geometry An instance of BufferGeometry.
     * @param material An instance of Material (optional).
     */
    constructor(geometry?: TGeometry, material?: TMaterial);

    type: 'Points';
    morphTargetInfluences?: number[] | undefined;
    morphTargetDictionary?: { [key: string]: number } | undefined;
    readonly isPoints: true;

    /**
     * An instance of BufferGeometry, where each vertex designates the position of a particle in the system.
     */
    geometry: TGeometry;

    /**
     * An instance of Material, defining the object's appearance. Default is a PointsMaterial with randomised colour.
     */
    material: TMaterial;

    raycast(raycaster: Raycaster, intersects: Intersection[]): void;
    updateMorphTargets(): void;
}

/**
 * @deprecated THREE.WebGLMultisampleRenderTarget has been removed. Use a normal {@link WebGLRenderTarget render target} and set the "samples" property to greater 0 to enable multisampling.
 */
declare class WebGLMultisampleRenderTarget extends WebGLRenderTarget {}

declare class WebGL1Renderer extends WebGLRenderer {
    constructor(parameters?: WebGLRendererParameters);
    readonly isWebGL1Renderer: true;
}

/**
 * Represents a three-dimensional render target.
 */
declare class WebGL3DRenderTarget extends WebGLRenderTarget {
    /**
     * Creates a new WebGL3DRenderTarget.
     *
     * @param width the width of the render target, in pixels.
     * @param height the height of the render target, in pixels.
     * @param depth the depth of the render target.
     */
    constructor(width: number, height: number, depth: number);

    /**
     * The depth of the render target.
     */
    depth: number;

    /**
     * The texture property is overwritten with an instance of {@link Data3DTexture}.
     */
    texture: Data3DTexture;

    readonly isWebGL3DRenderTarget: true;
}

/**
 * This type of render target represents an array of textures.
 */
declare class WebGLArrayRenderTarget extends WebGLRenderTarget {
    /**
     * Creates a new WebGLArrayRenderTarget.
     *
     * @param width the width of the render target, in pixels.
     * @param height the height of the render target, in pixels.
     * @param depth the depth/layer count of the render target.
     */
    constructor(width: number, height: number, depth: number);

    /**
     * The depth of the render target.
     */
    depth: number;

    /**
     * The texture property is overwritten with an instance of {@link DataArrayTexture}.
     */
    texture: DataArrayTexture;

    readonly isWebGLArrayRenderTarget: true;
}

declare function cloneUniforms(uniforms_src: any): any;
declare function mergeUniforms(uniforms: any[]): any;

declare namespace UniformsUtils {
    export { mergeUniforms as merge, cloneUniforms as clone };
}

// Renderers / Shaders /////////////////////////////////////////////////////////////////////
declare let ShaderChunk: {
    [name: string]: string;

    alphamap_fragment: string;
    alphamap_pars_fragment: string;
    alphatest_fragment: string;
    aomap_fragment: string;
    aomap_pars_fragment: string;
    begin_vertex: string;
    beginnormal_vertex: string;
    bsdfs: string;
    bumpmap_pars_fragment: string;
    clipping_planes_fragment: string;
    clipping_planes_pars_fragment: string;
    clipping_planes_pars_vertex: string;
    clipping_planes_vertex: string;
    color_fragment: string;
    color_pars_fragment: string;
    color_pars_vertex: string;
    color_vertex: string;
    common: string;
    cube_frag: string;
    cube_vert: string;
    cube_uv_reflection_fragment: string;
    defaultnormal_vertex: string;
    depth_frag: string;
    depth_vert: string;
    distanceRGBA_frag: string;
    distanceRGBA_vert: string;
    displacementmap_vertex: string;
    displacementmap_pars_vertex: string;
    emissivemap_fragment: string;
    emissivemap_pars_fragment: string;
    encodings_pars_fragment: string;
    encodings_fragment: string;
    envmap_fragment: string;
    envmap_common_pars_fragment: string;
    envmap_pars_fragment: string;
    envmap_pars_vertex: string;
    envmap_vertex: string;
    equirect_frag: string;
    equirect_vert: string;
    fog_fragment: string;
    fog_pars_fragment: string;
    linedashed_frag: string;
    linedashed_vert: string;
    lightmap_fragment: string;
    lightmap_pars_fragment: string;
    lights_lambert_vertex: string;
    lights_pars_begin: string;
    envmap_physical_pars_fragment: string;
    lights_pars_map: string;
    lights_phong_fragment: string;
    lights_phong_pars_fragment: string;
    lights_physical_fragment: string;
    lights_physical_pars_fragment: string;
    lights_fragment_begin: string;
    lights_fragment_maps: string;
    lights_fragment_end: string;
    logdepthbuf_fragment: string;
    logdepthbuf_pars_fragment: string;
    logdepthbuf_pars_vertex: string;
    logdepthbuf_vertex: string;
    map_fragment: string;
    map_pars_fragment: string;
    map_particle_fragment: string;
    map_particle_pars_fragment: string;
    meshbasic_frag: string;
    meshbasic_vert: string;
    meshlambert_frag: string;
    meshlambert_vert: string;
    meshphong_frag: string;
    meshphong_vert: string;
    meshphysical_frag: string;
    meshphysical_vert: string;
    metalnessmap_fragment: string;
    metalnessmap_pars_fragment: string;
    morphnormal_vertex: string;
    morphtarget_pars_vertex: string;
    morphtarget_vertex: string;
    normal_flip: string;
    normal_frag: string;
    normal_fragment_begin: string;
    normal_fragment_maps: string;
    normal_vert: string;
    normalmap_pars_fragment: string;
    clearcoat_normal_fragment_begin: string;
    clearcoat_normal_fragment_maps: string;
    clearcoat_pars_fragment: string;
    packing: string;
    points_frag: string;
    points_vert: string;
    shadow_frag: string;
    shadow_vert: string;

    premultiplied_alpha_fragment: string;
    project_vertex: string;
    roughnessmap_fragment: string;
    roughnessmap_pars_fragment: string;
    shadowmap_pars_fragment: string;
    shadowmap_pars_vertex: string;
    shadowmap_vertex: string;
    shadowmask_pars_fragment: string;
    skinbase_vertex: string;
    skinning_pars_vertex: string;
    skinning_vertex: string;
    skinnormal_vertex: string;
    specularmap_fragment: string;
    specularmap_pars_fragment: string;
    tonemapping_fragment: string;
    tonemapping_pars_fragment: string;
    uv2_pars_fragment: string;
    uv2_pars_vertex: string;
    uv2_vertex: string;
    uv_pars_fragment: string;
    uv_pars_vertex: string;
    uv_vertex: string;
    worldpos_vertex: string;
};

// Renderers / WebGL /////////////////////////////////////////////////////////////////////


declare class WebGLBufferRenderer {
    constructor(
        gl: WebGLRenderingContext,
        extensions: WebGLExtensions,
        info: WebGLInfo,
        capabilities: WebGLCapabilities,
    );

    setMode(value: any): void;
    render(start: any, count: number): void;
    renderInstances(start: any, count: number, primcount: number): void;
}

declare class WebGLClipping {
    constructor(properties: WebGLProperties);

    uniform: { value: any; needsUpdate: boolean };

    /**
     * @default 0
     */
    numPlanes: number;

    /**
     * @default 0
     */
    numIntersection: number;

    init(planes: any[], enableLocalClipping: boolean, camera: Camera): boolean;
    beginShadows(): void;
    endShadows(): void;
    setState(material: Material, camera: Camera, useCache: boolean): void;
}

declare class WebGLCubeUVMaps {
    constructor(renderer: WebGLRenderer);

    get<T>(texture: T): T extends Texture ? Texture : T;
    dispose(): void;
}

declare class WebGLAttributes {
    constructor(gl: WebGLRenderingContext | WebGL2RenderingContext, capabilities: WebGLCapabilities);

    get(attribute: BufferAttribute | InterleavedBufferAttribute): {
        buffer: WebGLBuffer;
        type: number;
        bytesPerElement: number;
        version: number;
    };

    remove(attribute: BufferAttribute | InterleavedBufferAttribute): void;

    update(attribute: BufferAttribute | InterleavedBufferAttribute, bufferType: number): void;
}

declare class WebGLGeometries {
    constructor(gl: WebGLRenderingContext, attributes: WebGLAttributes, info: WebGLInfo);

    get(object: Object3D, geometry: BufferGeometry): BufferGeometry;
    update(geometry: BufferGeometry): void;
    getWireframeAttribute(geometry: BufferGeometry): BufferAttribute;
}

declare class WebGLIndexedBufferRenderer {
    constructor(gl: WebGLRenderingContext, extensions: any, info: any, capabilities: any);

    setMode(value: any): void;
    setIndex(index: any): void;
    render(start: any, count: number): void;
    renderInstances(start: any, count: number, primcount: number): void;
}

declare class WebGLLights {
    constructor(extensions: WebGLExtensions, capabilities: WebGLCapabilities);

    state: {
        version: number;

        hash: {
            directionalLength: number;
            pointLength: number;
            spotLength: number;
            rectAreaLength: number;
            hemiLength: number;

            numDirectionalShadows: number;
            numPointShadows: number;
            numSpotShadows: number;
        };

        ambient: number[];
        probe: any[];
        directional: any[];
        directionalShadow: any[];
        directionalShadowMap: any[];
        directionalShadowMatrix: any[];
        spot: any[];
        spotShadow: any[];
        spotShadowMap: any[];
        spotShadowMatrix: any[];
        rectArea: any[];
        point: any[];
        pointShadow: any[];
        pointShadowMap: any[];
        pointShadowMatrix: any[];
        hemi: any[];
    };

    get(light: any): any;
    setup(lights: any): void;
    setupView(lights: any, camera: any): void;
}

declare class WebGLCubeMaps {
    constructor(renderer: WebGLRenderer);

    get(texture: any): any;
    dispose(): void;
}

declare class WebGLBindingStates {
    constructor(
        gl: WebGLRenderingContext,
        extensions: WebGLExtensions,
        attributes: WebGLAttributes,
        capabilities: WebGLCapabilities,
    );

    setup(
        object: Object3D,
        material: Material,
        program: WebGLProgram,
        geometry: BufferGeometry,
        index: BufferAttribute,
    ): void;
    reset(): void;
    resetDefaultState(): void;
    dispose(): void;
    releaseStatesOfGeometry(): void;
    releaseStatesOfProgram(): void;
    initAttributes(): void;
    enableAttribute(attribute: number): void;
    disableUnusedAttributes(): void;
}

declare class WebGLPrograms {
    constructor(
        renderer: WebGLRenderer,
        cubemaps: WebGLCubeMaps,
        extensions: WebGLExtensions,
        capabilities: WebGLCapabilities,
        bindingStates: WebGLBindingStates,
        clipping: WebGLClipping,
    );

    programs: WebGLProgram[];

    getParameters(material: Material, lights: any, shadows: object[], scene: Scene$1, object: any): any;
    getProgramCacheKey(parameters: any): string;
    getUniforms(material: Material): object;
    acquireProgram(parameters: any, cacheKey: string): WebGLProgram;
    releaseProgram(program: WebGLProgram): void;
}

/**
 * This class contains the parameters that define linear fog, i.e., that grows exponentially denser with the distance.
 */
declare class FogExp2 implements FogBase {
    constructor(hex: number | string, density?: number);

    /**
     * @default ''
     */
    name: string;

    color: Color;

    /**
     * Defines how fast the fog will grow dense.
     * @default 0.00025
     */
    density: number;

    readonly isFogExp2: true;

    clone(): FogExp2;
    toJSON(): any;
}

declare class VideoTexture extends Texture {
    /**
     * @param video
     * @param [mapping=THREE.Texture.DEFAULT_MAPPING]
     * @param [wrapS=THREE.ClampToEdgeWrapping]
     * @param [wrapT=THREE.ClampToEdgeWrapping]
     * @param [magFilter=THREE.LinearFilter]
     * @param [minFilter=THREE.LinearFilter]
     * @param [format=THREE.RGBAFormat]
     * @param [type=THREE.UnsignedByteType]
     * @param [anisotropy=1]
     */
    constructor(
        video: HTMLVideoElement,
        mapping?: Mapping,
        wrapS?: Wrapping,
        wrapT?: Wrapping,
        magFilter?: TextureFilter,
        minFilter?: TextureFilter,
        format?: PixelFormat,
        type?: TextureDataType,
        anisotropy?: number,
    );

    readonly isVideoTexture: true;

    /**
     * @default false
     */
    generateMipmaps: boolean;
}

/**
 * @deprecated THREE.DataTexture2DArray has been renamed to DataArrayTexture.
 */
declare class DataTexture2DArray extends DataArrayTexture {}

/**
 * @deprecated THREE.DataTexture3D has been renamed to Data3DTexture.
 */
declare class DataTexture3D extends Data3DTexture {}

declare class CanvasTexture extends Texture {
    /**
     * @param canvas
     * @param [format=THREE.RGBAFormat]
     * @param [type=THREE.UnsignedByteType]
     * @param [mapping=THREE.Texture.DEFAULT_MAPPING]
     * @param [wrapS=THREE.ClampToEdgeWrapping]
     * @param [wrapT=THREE.ClampToEdgeWrapping]
     * @param [magFilter=THREE.LinearFilter]
     * @param [minFilter=THREE.LinearMipmapLinearFilter]
     * @param [anisotropy=1]
     * @param [encoding=THREE.LinearEncoding]
     */
    constructor(
        canvas: HTMLImageElement | HTMLCanvasElement | HTMLVideoElement | ImageBitmap,
        mapping?: Mapping,
        wrapS?: Wrapping,
        wrapT?: Wrapping,
        magFilter?: TextureFilter,
        minFilter?: TextureFilter,
        format?: PixelFormat,
        type?: TextureDataType,
        anisotropy?: number,
    );

    readonly isCanvasTexture: true;
}

declare class FramebufferTexture extends Texture {
    readonly isFramebufferTexture: true;

    constructor(width: number, height: number, format: PixelFormat);
}

// Type definitions for three 0.142

type THREE_WebGLUtils = WebGLUtils;
declare const THREE_WebGLUtils: typeof WebGLUtils;
declare const THREE_REVISION: typeof REVISION;
type THREE_MOUSE = MOUSE;
declare const THREE_MOUSE: typeof MOUSE;
type THREE_TOUCH = TOUCH;
declare const THREE_TOUCH: typeof TOUCH;
type THREE_CullFace = CullFace;
declare const THREE_CullFace: typeof CullFace;
declare const THREE_CullFaceNone: typeof CullFaceNone;
declare const THREE_CullFaceBack: typeof CullFaceBack;
declare const THREE_CullFaceFront: typeof CullFaceFront;
declare const THREE_CullFaceFrontBack: typeof CullFaceFrontBack;
type THREE_ShadowMapType = ShadowMapType;
declare const THREE_ShadowMapType: typeof ShadowMapType;
declare const THREE_BasicShadowMap: typeof BasicShadowMap;
declare const THREE_PCFShadowMap: typeof PCFShadowMap;
declare const THREE_PCFSoftShadowMap: typeof PCFSoftShadowMap;
declare const THREE_VSMShadowMap: typeof VSMShadowMap;
type THREE_Side = Side;
declare const THREE_Side: typeof Side;
declare const THREE_FrontSide: typeof FrontSide;
declare const THREE_BackSide: typeof BackSide;
declare const THREE_DoubleSide: typeof DoubleSide;
type THREE_Shading = Shading;
declare const THREE_Shading: typeof Shading;
declare const THREE_FlatShading: typeof FlatShading;
declare const THREE_SmoothShading: typeof SmoothShading;
type THREE_Blending = Blending;
declare const THREE_Blending: typeof Blending;
declare const THREE_NoBlending: typeof NoBlending;
declare const THREE_NormalBlending: typeof NormalBlending;
declare const THREE_AdditiveBlending: typeof AdditiveBlending;
declare const THREE_SubtractiveBlending: typeof SubtractiveBlending;
declare const THREE_MultiplyBlending: typeof MultiplyBlending;
declare const THREE_CustomBlending: typeof CustomBlending;
type THREE_BlendingEquation = BlendingEquation;
declare const THREE_BlendingEquation: typeof BlendingEquation;
declare const THREE_AddEquation: typeof AddEquation;
declare const THREE_SubtractEquation: typeof SubtractEquation;
declare const THREE_ReverseSubtractEquation: typeof ReverseSubtractEquation;
declare const THREE_MinEquation: typeof MinEquation;
declare const THREE_MaxEquation: typeof MaxEquation;
type THREE_BlendingDstFactor = BlendingDstFactor;
declare const THREE_BlendingDstFactor: typeof BlendingDstFactor;
declare const THREE_ZeroFactor: typeof ZeroFactor;
declare const THREE_OneFactor: typeof OneFactor;
declare const THREE_SrcColorFactor: typeof SrcColorFactor;
declare const THREE_OneMinusSrcColorFactor: typeof OneMinusSrcColorFactor;
declare const THREE_SrcAlphaFactor: typeof SrcAlphaFactor;
declare const THREE_OneMinusSrcAlphaFactor: typeof OneMinusSrcAlphaFactor;
declare const THREE_DstAlphaFactor: typeof DstAlphaFactor;
declare const THREE_OneMinusDstAlphaFactor: typeof OneMinusDstAlphaFactor;
declare const THREE_DstColorFactor: typeof DstColorFactor;
declare const THREE_OneMinusDstColorFactor: typeof OneMinusDstColorFactor;
type THREE_BlendingSrcFactor = BlendingSrcFactor;
declare const THREE_BlendingSrcFactor: typeof BlendingSrcFactor;
declare const THREE_SrcAlphaSaturateFactor: typeof SrcAlphaSaturateFactor;
type THREE_DepthModes = DepthModes;
declare const THREE_DepthModes: typeof DepthModes;
declare const THREE_NeverDepth: typeof NeverDepth;
declare const THREE_AlwaysDepth: typeof AlwaysDepth;
declare const THREE_LessDepth: typeof LessDepth;
declare const THREE_LessEqualDepth: typeof LessEqualDepth;
declare const THREE_EqualDepth: typeof EqualDepth;
declare const THREE_GreaterEqualDepth: typeof GreaterEqualDepth;
declare const THREE_GreaterDepth: typeof GreaterDepth;
declare const THREE_NotEqualDepth: typeof NotEqualDepth;
type THREE_Combine = Combine;
declare const THREE_Combine: typeof Combine;
declare const THREE_MultiplyOperation: typeof MultiplyOperation;
declare const THREE_MixOperation: typeof MixOperation;
declare const THREE_AddOperation: typeof AddOperation;
type THREE_ToneMapping = ToneMapping;
declare const THREE_ToneMapping: typeof ToneMapping;
declare const THREE_NoToneMapping: typeof NoToneMapping;
declare const THREE_LinearToneMapping: typeof LinearToneMapping;
declare const THREE_ReinhardToneMapping: typeof ReinhardToneMapping;
declare const THREE_CineonToneMapping: typeof CineonToneMapping;
declare const THREE_ACESFilmicToneMapping: typeof ACESFilmicToneMapping;
declare const THREE_CustomToneMapping: typeof CustomToneMapping;
type THREE_Mapping = Mapping;
declare const THREE_Mapping: typeof Mapping;
declare const THREE_UVMapping: typeof UVMapping;
declare const THREE_CubeReflectionMapping: typeof CubeReflectionMapping;
declare const THREE_CubeRefractionMapping: typeof CubeRefractionMapping;
declare const THREE_EquirectangularReflectionMapping: typeof EquirectangularReflectionMapping;
declare const THREE_EquirectangularRefractionMapping: typeof EquirectangularRefractionMapping;
declare const THREE_CubeUVReflectionMapping: typeof CubeUVReflectionMapping;
type THREE_Wrapping = Wrapping;
declare const THREE_Wrapping: typeof Wrapping;
declare const THREE_RepeatWrapping: typeof RepeatWrapping;
declare const THREE_ClampToEdgeWrapping: typeof ClampToEdgeWrapping;
declare const THREE_MirroredRepeatWrapping: typeof MirroredRepeatWrapping;
type THREE_TextureFilter = TextureFilter;
declare const THREE_TextureFilter: typeof TextureFilter;
declare const THREE_NearestFilter: typeof NearestFilter;
declare const THREE_NearestMipmapNearestFilter: typeof NearestMipmapNearestFilter;
declare const THREE_NearestMipMapNearestFilter: typeof NearestMipMapNearestFilter;
declare const THREE_NearestMipmapLinearFilter: typeof NearestMipmapLinearFilter;
declare const THREE_NearestMipMapLinearFilter: typeof NearestMipMapLinearFilter;
declare const THREE_LinearFilter: typeof LinearFilter;
declare const THREE_LinearMipmapNearestFilter: typeof LinearMipmapNearestFilter;
declare const THREE_LinearMipMapNearestFilter: typeof LinearMipMapNearestFilter;
declare const THREE_LinearMipmapLinearFilter: typeof LinearMipmapLinearFilter;
declare const THREE_LinearMipMapLinearFilter: typeof LinearMipMapLinearFilter;
type THREE_TextureDataType = TextureDataType;
declare const THREE_TextureDataType: typeof TextureDataType;
declare const THREE_UnsignedByteType: typeof UnsignedByteType;
declare const THREE_ByteType: typeof ByteType;
declare const THREE_ShortType: typeof ShortType;
declare const THREE_UnsignedShortType: typeof UnsignedShortType;
declare const THREE_IntType: typeof IntType;
declare const THREE_UnsignedIntType: typeof UnsignedIntType;
declare const THREE_FloatType: typeof FloatType;
declare const THREE_HalfFloatType: typeof HalfFloatType;
declare const THREE_UnsignedShort4444Type: typeof UnsignedShort4444Type;
declare const THREE_UnsignedShort5551Type: typeof UnsignedShort5551Type;
declare const THREE_UnsignedInt248Type: typeof UnsignedInt248Type;
type THREE_PixelFormat = PixelFormat;
declare const THREE_PixelFormat: typeof PixelFormat;
declare const THREE_AlphaFormat: typeof AlphaFormat;
declare const THREE_RGBFormat: typeof RGBFormat;
declare const THREE_RGBAFormat: typeof RGBAFormat;
declare const THREE_LuminanceFormat: typeof LuminanceFormat;
declare const THREE_LuminanceAlphaFormat: typeof LuminanceAlphaFormat;
declare const THREE_DepthFormat: typeof DepthFormat;
declare const THREE_DepthStencilFormat: typeof DepthStencilFormat;
declare const THREE_RedFormat: typeof RedFormat;
declare const THREE_RedIntegerFormat: typeof RedIntegerFormat;
declare const THREE_RGFormat: typeof RGFormat;
declare const THREE_RGIntegerFormat: typeof RGIntegerFormat;
declare const THREE_RGBAIntegerFormat: typeof RGBAIntegerFormat;
declare const THREE__SRGBFormat: typeof _SRGBFormat;
declare const THREE__SRGBAFormat: typeof _SRGBAFormat;
type THREE_PixelFormatGPU = PixelFormatGPU;
type THREE_CompressedPixelFormat = CompressedPixelFormat;
declare const THREE_CompressedPixelFormat: typeof CompressedPixelFormat;
declare const THREE_RGB_S3TC_DXT1_Format: typeof RGB_S3TC_DXT1_Format;
declare const THREE_RGBA_S3TC_DXT1_Format: typeof RGBA_S3TC_DXT1_Format;
declare const THREE_RGBA_S3TC_DXT3_Format: typeof RGBA_S3TC_DXT3_Format;
declare const THREE_RGBA_S3TC_DXT5_Format: typeof RGBA_S3TC_DXT5_Format;
declare const THREE_RGB_PVRTC_4BPPV1_Format: typeof RGB_PVRTC_4BPPV1_Format;
declare const THREE_RGB_PVRTC_2BPPV1_Format: typeof RGB_PVRTC_2BPPV1_Format;
declare const THREE_RGBA_PVRTC_4BPPV1_Format: typeof RGBA_PVRTC_4BPPV1_Format;
declare const THREE_RGBA_PVRTC_2BPPV1_Format: typeof RGBA_PVRTC_2BPPV1_Format;
declare const THREE_RGB_ETC1_Format: typeof RGB_ETC1_Format;
declare const THREE_RGB_ETC2_Format: typeof RGB_ETC2_Format;
declare const THREE_RGBA_ETC2_EAC_Format: typeof RGBA_ETC2_EAC_Format;
declare const THREE_RGBA_ASTC_4x4_Format: typeof RGBA_ASTC_4x4_Format;
declare const THREE_RGBA_ASTC_5x4_Format: typeof RGBA_ASTC_5x4_Format;
declare const THREE_RGBA_ASTC_5x5_Format: typeof RGBA_ASTC_5x5_Format;
declare const THREE_RGBA_ASTC_6x5_Format: typeof RGBA_ASTC_6x5_Format;
declare const THREE_RGBA_ASTC_6x6_Format: typeof RGBA_ASTC_6x6_Format;
declare const THREE_RGBA_ASTC_8x5_Format: typeof RGBA_ASTC_8x5_Format;
declare const THREE_RGBA_ASTC_8x6_Format: typeof RGBA_ASTC_8x6_Format;
declare const THREE_RGBA_ASTC_8x8_Format: typeof RGBA_ASTC_8x8_Format;
declare const THREE_RGBA_ASTC_10x5_Format: typeof RGBA_ASTC_10x5_Format;
declare const THREE_RGBA_ASTC_10x6_Format: typeof RGBA_ASTC_10x6_Format;
declare const THREE_RGBA_ASTC_10x8_Format: typeof RGBA_ASTC_10x8_Format;
declare const THREE_RGBA_ASTC_10x10_Format: typeof RGBA_ASTC_10x10_Format;
declare const THREE_RGBA_ASTC_12x10_Format: typeof RGBA_ASTC_12x10_Format;
declare const THREE_RGBA_ASTC_12x12_Format: typeof RGBA_ASTC_12x12_Format;
declare const THREE_RGBA_BPTC_Format: typeof RGBA_BPTC_Format;
type THREE_AnimationActionLoopStyles = AnimationActionLoopStyles;
declare const THREE_AnimationActionLoopStyles: typeof AnimationActionLoopStyles;
declare const THREE_LoopOnce: typeof LoopOnce;
declare const THREE_LoopRepeat: typeof LoopRepeat;
declare const THREE_LoopPingPong: typeof LoopPingPong;
type THREE_InterpolationModes = InterpolationModes;
declare const THREE_InterpolationModes: typeof InterpolationModes;
declare const THREE_InterpolateDiscrete: typeof InterpolateDiscrete;
declare const THREE_InterpolateLinear: typeof InterpolateLinear;
declare const THREE_InterpolateSmooth: typeof InterpolateSmooth;
type THREE_InterpolationEndingModes = InterpolationEndingModes;
declare const THREE_InterpolationEndingModes: typeof InterpolationEndingModes;
declare const THREE_ZeroCurvatureEnding: typeof ZeroCurvatureEnding;
declare const THREE_ZeroSlopeEnding: typeof ZeroSlopeEnding;
declare const THREE_WrapAroundEnding: typeof WrapAroundEnding;
type THREE_AnimationBlendMode = AnimationBlendMode;
declare const THREE_AnimationBlendMode: typeof AnimationBlendMode;
declare const THREE_NormalAnimationBlendMode: typeof NormalAnimationBlendMode;
declare const THREE_AdditiveAnimationBlendMode: typeof AdditiveAnimationBlendMode;
type THREE_TrianglesDrawModes = TrianglesDrawModes;
declare const THREE_TrianglesDrawModes: typeof TrianglesDrawModes;
declare const THREE_TrianglesDrawMode: typeof TrianglesDrawMode;
declare const THREE_TriangleStripDrawMode: typeof TriangleStripDrawMode;
declare const THREE_TriangleFanDrawMode: typeof TriangleFanDrawMode;
type THREE_TextureEncoding = TextureEncoding;
declare const THREE_TextureEncoding: typeof TextureEncoding;
declare const THREE_LinearEncoding: typeof LinearEncoding;
declare const THREE_sRGBEncoding: typeof sRGBEncoding;
type THREE_DepthPackingStrategies = DepthPackingStrategies;
declare const THREE_DepthPackingStrategies: typeof DepthPackingStrategies;
declare const THREE_BasicDepthPacking: typeof BasicDepthPacking;
declare const THREE_RGBADepthPacking: typeof RGBADepthPacking;
type THREE_NormalMapTypes = NormalMapTypes;
declare const THREE_NormalMapTypes: typeof NormalMapTypes;
declare const THREE_TangentSpaceNormalMap: typeof TangentSpaceNormalMap;
declare const THREE_ObjectSpaceNormalMap: typeof ObjectSpaceNormalMap;
type THREE_ColorSpace = ColorSpace;
type THREE_NoColorSpace = NoColorSpace;
type THREE_SRGBColorSpace = SRGBColorSpace;
type THREE_LinearSRGBColorSpace = LinearSRGBColorSpace;
type THREE_StencilOp = StencilOp;
declare const THREE_StencilOp: typeof StencilOp;
declare const THREE_ZeroStencilOp: typeof ZeroStencilOp;
declare const THREE_KeepStencilOp: typeof KeepStencilOp;
declare const THREE_ReplaceStencilOp: typeof ReplaceStencilOp;
declare const THREE_IncrementStencilOp: typeof IncrementStencilOp;
declare const THREE_DecrementStencilOp: typeof DecrementStencilOp;
declare const THREE_IncrementWrapStencilOp: typeof IncrementWrapStencilOp;
declare const THREE_DecrementWrapStencilOp: typeof DecrementWrapStencilOp;
declare const THREE_InvertStencilOp: typeof InvertStencilOp;
type THREE_StencilFunc = StencilFunc;
declare const THREE_StencilFunc: typeof StencilFunc;
declare const THREE_NeverStencilFunc: typeof NeverStencilFunc;
declare const THREE_LessStencilFunc: typeof LessStencilFunc;
declare const THREE_EqualStencilFunc: typeof EqualStencilFunc;
declare const THREE_LessEqualStencilFunc: typeof LessEqualStencilFunc;
declare const THREE_GreaterStencilFunc: typeof GreaterStencilFunc;
declare const THREE_NotEqualStencilFunc: typeof NotEqualStencilFunc;
declare const THREE_GreaterEqualStencilFunc: typeof GreaterEqualStencilFunc;
declare const THREE_AlwaysStencilFunc: typeof AlwaysStencilFunc;
type THREE_Usage = Usage;
declare const THREE_Usage: typeof Usage;
declare const THREE_StaticDrawUsage: typeof StaticDrawUsage;
declare const THREE_DynamicDrawUsage: typeof DynamicDrawUsage;
declare const THREE_StreamDrawUsage: typeof StreamDrawUsage;
declare const THREE_StaticReadUsage: typeof StaticReadUsage;
declare const THREE_DynamicReadUsage: typeof DynamicReadUsage;
declare const THREE_StreamReadUsage: typeof StreamReadUsage;
declare const THREE_StaticCopyUsage: typeof StaticCopyUsage;
declare const THREE_DynamicCopyUsage: typeof DynamicCopyUsage;
declare const THREE_StreamCopyUsage: typeof StreamCopyUsage;
type THREE_GLSLVersion = GLSLVersion;
declare const THREE_GLSLVersion: typeof GLSLVersion;
declare const THREE_GLSL1: typeof GLSL1;
declare const THREE_GLSL3: typeof GLSL3;
type THREE_BuiltinShaderAttributeName = BuiltinShaderAttributeName;
type THREE_ColorRepresentation = ColorRepresentation;
type THREE_VectorKeyframeTrack = VectorKeyframeTrack;
declare const THREE_VectorKeyframeTrack: typeof VectorKeyframeTrack;
type THREE_StringKeyframeTrack = StringKeyframeTrack;
declare const THREE_StringKeyframeTrack: typeof StringKeyframeTrack;
type THREE_QuaternionKeyframeTrack = QuaternionKeyframeTrack;
declare const THREE_QuaternionKeyframeTrack: typeof QuaternionKeyframeTrack;
type THREE_NumberKeyframeTrack = NumberKeyframeTrack;
declare const THREE_NumberKeyframeTrack: typeof NumberKeyframeTrack;
type THREE_ColorKeyframeTrack = ColorKeyframeTrack;
declare const THREE_ColorKeyframeTrack: typeof ColorKeyframeTrack;
type THREE_BooleanKeyframeTrack = BooleanKeyframeTrack;
declare const THREE_BooleanKeyframeTrack: typeof BooleanKeyframeTrack;
type THREE_PropertyMixer = PropertyMixer;
declare const THREE_PropertyMixer: typeof PropertyMixer;
type THREE_ParseTrackNameResults = ParseTrackNameResults;
declare const THREE_PropertyBinding: typeof PropertyBinding;
type THREE_KeyframeTrack = KeyframeTrack;
declare const THREE_KeyframeTrack: typeof KeyframeTrack;
declare const THREE_AnimationUtils: typeof AnimationUtils;
type THREE_AnimationObjectGroup = AnimationObjectGroup;
declare const THREE_AnimationObjectGroup: typeof AnimationObjectGroup;
type THREE_AnimationMixer = AnimationMixer;
declare const THREE_AnimationMixer: typeof AnimationMixer;
type THREE_MorphTarget = MorphTarget;
type THREE_AnimationClip = AnimationClip;
declare const THREE_AnimationClip: typeof AnimationClip;
type THREE_AnimationAction = AnimationAction;
declare const THREE_AnimationAction: typeof AnimationAction;
type THREE_AudioListener = AudioListener;
declare const THREE_AudioListener: typeof AudioListener;
type THREE_PositionalAudio = PositionalAudio;
declare const THREE_PositionalAudio: typeof PositionalAudio;
declare const THREE_AudioContext: typeof AudioContext;
type THREE_AudioAnalyser = AudioAnalyser;
declare const THREE_AudioAnalyser: typeof AudioAnalyser;
type THREE_Audio<NodeType extends AudioNode = GainNode> = Audio<NodeType>;
declare const THREE_Audio: typeof Audio;
type THREE_StereoCamera = StereoCamera;
declare const THREE_StereoCamera: typeof StereoCamera;
type THREE_PerspectiveCamera = PerspectiveCamera;
declare const THREE_PerspectiveCamera: typeof PerspectiveCamera;
type THREE_OrthographicCamera = OrthographicCamera;
declare const THREE_OrthographicCamera: typeof OrthographicCamera;
type THREE_CubeCamera = CubeCamera;
declare const THREE_CubeCamera: typeof CubeCamera;
type THREE_ArrayCamera = ArrayCamera;
declare const THREE_ArrayCamera: typeof ArrayCamera;
type THREE_Camera = Camera;
declare const THREE_Camera: typeof Camera;
type THREE_Uniform = Uniform;
declare const THREE_Uniform: typeof Uniform;
type THREE_InstancedBufferGeometry = InstancedBufferGeometry;
declare const THREE_InstancedBufferGeometry: typeof InstancedBufferGeometry;
type THREE_BufferGeometry = BufferGeometry;
declare const THREE_BufferGeometry: typeof BufferGeometry;
type THREE_InterleavedBufferAttribute = InterleavedBufferAttribute;
declare const THREE_InterleavedBufferAttribute: typeof InterleavedBufferAttribute;
type THREE_InstancedInterleavedBuffer = InstancedInterleavedBuffer;
declare const THREE_InstancedInterleavedBuffer: typeof InstancedInterleavedBuffer;
type THREE_InterleavedBuffer = InterleavedBuffer;
declare const THREE_InterleavedBuffer: typeof InterleavedBuffer;
declare const THREE_BufferGeometryUtils: typeof BufferGeometryUtils;
declare const THREE_GeometryUtils: typeof GeometryUtils;
type THREE_InstancedBufferAttribute = InstancedBufferAttribute;
declare const THREE_InstancedBufferAttribute: typeof InstancedBufferAttribute;
type THREE_GLBufferAttribute = GLBufferAttribute;
declare const THREE_GLBufferAttribute: typeof GLBufferAttribute;
type THREE_BufferAttribute = BufferAttribute;
declare const THREE_BufferAttribute: typeof BufferAttribute;
type THREE_Int8Attribute = Int8Attribute;
declare const THREE_Int8Attribute: typeof Int8Attribute;
type THREE_Uint8Attribute = Uint8Attribute;
declare const THREE_Uint8Attribute: typeof Uint8Attribute;
type THREE_Uint8ClampedAttribute = Uint8ClampedAttribute;
declare const THREE_Uint8ClampedAttribute: typeof Uint8ClampedAttribute;
type THREE_Int16Attribute = Int16Attribute;
declare const THREE_Int16Attribute: typeof Int16Attribute;
type THREE_Uint16Attribute = Uint16Attribute;
declare const THREE_Uint16Attribute: typeof Uint16Attribute;
type THREE_Int32Attribute = Int32Attribute;
declare const THREE_Int32Attribute: typeof Int32Attribute;
type THREE_Uint32Attribute = Uint32Attribute;
declare const THREE_Uint32Attribute: typeof Uint32Attribute;
type THREE_Float32Attribute = Float32Attribute;
declare const THREE_Float32Attribute: typeof Float32Attribute;
type THREE_Float64Attribute = Float64Attribute;
declare const THREE_Float64Attribute: typeof Float64Attribute;
type THREE_Int8BufferAttribute = Int8BufferAttribute;
declare const THREE_Int8BufferAttribute: typeof Int8BufferAttribute;
type THREE_Uint8BufferAttribute = Uint8BufferAttribute;
declare const THREE_Uint8BufferAttribute: typeof Uint8BufferAttribute;
type THREE_Uint8ClampedBufferAttribute = Uint8ClampedBufferAttribute;
declare const THREE_Uint8ClampedBufferAttribute: typeof Uint8ClampedBufferAttribute;
type THREE_Int16BufferAttribute = Int16BufferAttribute;
declare const THREE_Int16BufferAttribute: typeof Int16BufferAttribute;
type THREE_Uint16BufferAttribute = Uint16BufferAttribute;
declare const THREE_Uint16BufferAttribute: typeof Uint16BufferAttribute;
type THREE_Int32BufferAttribute = Int32BufferAttribute;
declare const THREE_Int32BufferAttribute: typeof Int32BufferAttribute;
type THREE_Uint32BufferAttribute = Uint32BufferAttribute;
declare const THREE_Uint32BufferAttribute: typeof Uint32BufferAttribute;
type THREE_Float16BufferAttribute = Float16BufferAttribute;
declare const THREE_Float16BufferAttribute: typeof Float16BufferAttribute;
type THREE_Float32BufferAttribute = Float32BufferAttribute;
declare const THREE_Float32BufferAttribute: typeof Float32BufferAttribute;
type THREE_Float64BufferAttribute = Float64BufferAttribute;
declare const THREE_Float64BufferAttribute: typeof Float64BufferAttribute;
type THREE_Object3D<E extends BaseEvent = Event$1> = Object3D<E>;
declare const THREE_Object3D: typeof Object3D;
type THREE_Face = Face;
type THREE_Intersection<TIntersected extends Object3D = Object3D> = Intersection<TIntersected>;
type THREE_RaycasterParameters = RaycasterParameters;
type THREE_Raycaster = Raycaster;
declare const THREE_Raycaster: typeof Raycaster;
type THREE_Layers = Layers;
declare const THREE_Layers: typeof Layers;
type THREE_BaseEvent = BaseEvent;
type THREE_EventListener<E, T, U> = EventListener<E, T, U>;
type THREE_EventDispatcher<E extends BaseEvent = Event$1> = EventDispatcher<E>;
declare const THREE_EventDispatcher: typeof EventDispatcher;
type THREE_Clock = Clock;
declare const THREE_Clock: typeof Clock;
type THREE_ArcCurve = ArcCurve;
declare const THREE_ArcCurve: typeof ArcCurve;
declare const THREE_CurveUtils: typeof CurveUtils;
type THREE_CatmullRomCurve3 = CatmullRomCurve3;
declare const THREE_CatmullRomCurve3: typeof CatmullRomCurve3;
type THREE_CubicBezierCurve = CubicBezierCurve;
declare const THREE_CubicBezierCurve: typeof CubicBezierCurve;
type THREE_CubicBezierCurve3 = CubicBezierCurve3;
declare const THREE_CubicBezierCurve3: typeof CubicBezierCurve3;
type THREE_EllipseCurve = EllipseCurve;
declare const THREE_EllipseCurve: typeof EllipseCurve;
type THREE_LineCurve = LineCurve;
declare const THREE_LineCurve: typeof LineCurve;
type THREE_LineCurve3 = LineCurve3;
declare const THREE_LineCurve3: typeof LineCurve3;
type THREE_QuadraticBezierCurve = QuadraticBezierCurve;
declare const THREE_QuadraticBezierCurve: typeof QuadraticBezierCurve;
type THREE_QuadraticBezierCurve3 = QuadraticBezierCurve3;
declare const THREE_QuadraticBezierCurve3: typeof QuadraticBezierCurve3;
type THREE_SplineCurve = SplineCurve;
declare const THREE_SplineCurve: typeof SplineCurve;
type THREE_Shape = Shape;
declare const THREE_Shape: typeof Shape;
type THREE_Path = Path;
declare const THREE_Path: typeof Path;
type THREE_ShapePath = ShapePath;
declare const THREE_ShapePath: typeof ShapePath;
type THREE_CurvePath<T extends Vector> = CurvePath<T>;
declare const THREE_CurvePath: typeof CurvePath;
type THREE_Curve<T extends Vector> = Curve<T>;
declare const THREE_Curve: typeof Curve;
declare const THREE_toHalfFloat: typeof toHalfFloat;
declare const THREE_fromHalfFloat: typeof fromHalfFloat;
declare const THREE_Earcut: typeof Earcut;
declare const THREE_ImageUtils: typeof ImageUtils;
type THREE_Vec2 = Vec2;
declare const THREE_ShapeUtils: typeof ShapeUtils;
type THREE_PMREMGenerator = PMREMGenerator;
declare const THREE_PMREMGenerator: typeof PMREMGenerator;
type THREE_BoxGeometry = BoxGeometry;
declare const THREE_BoxGeometry: typeof BoxGeometry;
type THREE_CapsuleGeometry = CapsuleGeometry;
declare const THREE_CapsuleGeometry: typeof CapsuleGeometry;
type THREE_CircleGeometry = CircleGeometry;
declare const THREE_CircleGeometry: typeof CircleGeometry;
type THREE_ConeGeometry = ConeGeometry;
declare const THREE_ConeGeometry: typeof ConeGeometry;
type THREE_CylinderGeometry = CylinderGeometry;
declare const THREE_CylinderGeometry: typeof CylinderGeometry;
type THREE_DodecahedronGeometry = DodecahedronGeometry;
declare const THREE_DodecahedronGeometry: typeof DodecahedronGeometry;
type THREE_EdgesGeometry<TBufferGeometry extends BufferGeometry = BufferGeometry> = EdgesGeometry<TBufferGeometry>;
declare const THREE_EdgesGeometry: typeof EdgesGeometry;
type THREE_ExtrudeGeometryOptions = ExtrudeGeometryOptions;
type THREE_UVGenerator = UVGenerator;
type THREE_ExtrudeGeometry = ExtrudeGeometry;
declare const THREE_ExtrudeGeometry: typeof ExtrudeGeometry;
type THREE_IcosahedronGeometry = IcosahedronGeometry;
declare const THREE_IcosahedronGeometry: typeof IcosahedronGeometry;
type THREE_LatheGeometry = LatheGeometry;
declare const THREE_LatheGeometry: typeof LatheGeometry;
type THREE_OctahedronGeometry = OctahedronGeometry;
declare const THREE_OctahedronGeometry: typeof OctahedronGeometry;
type THREE_PlaneGeometry = PlaneGeometry;
declare const THREE_PlaneGeometry: typeof PlaneGeometry;
type THREE_PolyhedronGeometry = PolyhedronGeometry;
declare const THREE_PolyhedronGeometry: typeof PolyhedronGeometry;
type THREE_RingGeometry = RingGeometry;
declare const THREE_RingGeometry: typeof RingGeometry;
type THREE_ShapeGeometry = ShapeGeometry;
declare const THREE_ShapeGeometry: typeof ShapeGeometry;
type THREE_SphereGeometry = SphereGeometry;
declare const THREE_SphereGeometry: typeof SphereGeometry;
type THREE_TetrahedronGeometry = TetrahedronGeometry;
declare const THREE_TetrahedronGeometry: typeof TetrahedronGeometry;
type THREE_TorusGeometry = TorusGeometry;
declare const THREE_TorusGeometry: typeof TorusGeometry;
type THREE_TorusKnotGeometry = TorusKnotGeometry;
declare const THREE_TorusKnotGeometry: typeof TorusKnotGeometry;
type THREE_TubeGeometry = TubeGeometry;
declare const THREE_TubeGeometry: typeof TubeGeometry;
type THREE_WireframeGeometry<TBufferGeometry extends BufferGeometry = BufferGeometry> = WireframeGeometry<TBufferGeometry>;
declare const THREE_WireframeGeometry: typeof WireframeGeometry;
type THREE_SpotLightHelper = SpotLightHelper;
declare const THREE_SpotLightHelper: typeof SpotLightHelper;
type THREE_SkeletonHelper = SkeletonHelper;
declare const THREE_SkeletonHelper: typeof SkeletonHelper;
type THREE_PointLightHelper = PointLightHelper;
declare const THREE_PointLightHelper: typeof PointLightHelper;
type THREE_HemisphereLightHelper = HemisphereLightHelper;
declare const THREE_HemisphereLightHelper: typeof HemisphereLightHelper;
type THREE_GridHelper = GridHelper;
declare const THREE_GridHelper: typeof GridHelper;
type THREE_PolarGridHelper = PolarGridHelper;
declare const THREE_PolarGridHelper: typeof PolarGridHelper;
type THREE_DirectionalLightHelper = DirectionalLightHelper;
declare const THREE_DirectionalLightHelper: typeof DirectionalLightHelper;
type THREE_CameraHelper = CameraHelper;
declare const THREE_CameraHelper: typeof CameraHelper;
type THREE_BoxHelper = BoxHelper;
declare const THREE_BoxHelper: typeof BoxHelper;
type THREE_Box3Helper = Box3Helper;
declare const THREE_Box3Helper: typeof Box3Helper;
type THREE_PlaneHelper = PlaneHelper;
declare const THREE_PlaneHelper: typeof PlaneHelper;
type THREE_ArrowHelper = ArrowHelper;
declare const THREE_ArrowHelper: typeof ArrowHelper;
type THREE_AxesHelper = AxesHelper;
declare const THREE_AxesHelper: typeof AxesHelper;
type THREE_SpotLightShadow = SpotLightShadow;
declare const THREE_SpotLightShadow: typeof SpotLightShadow;
type THREE_SpotLight = SpotLight;
declare const THREE_SpotLight: typeof SpotLight;
type THREE_PointLight = PointLight;
declare const THREE_PointLight: typeof PointLight;
type THREE_PointLightShadow = PointLightShadow;
declare const THREE_PointLightShadow: typeof PointLightShadow;
type THREE_RectAreaLight = RectAreaLight;
declare const THREE_RectAreaLight: typeof RectAreaLight;
type THREE_HemisphereLight = HemisphereLight;
declare const THREE_HemisphereLight: typeof HemisphereLight;
type THREE_DirectionalLightShadow = DirectionalLightShadow;
declare const THREE_DirectionalLightShadow: typeof DirectionalLightShadow;
type THREE_DirectionalLight = DirectionalLight;
declare const THREE_DirectionalLight: typeof DirectionalLight;
type THREE_AmbientLight = AmbientLight;
declare const THREE_AmbientLight: typeof AmbientLight;
type THREE_LightShadow = LightShadow;
declare const THREE_LightShadow: typeof LightShadow;
type THREE_Light = Light;
declare const THREE_Light: typeof Light;
type THREE_AmbientLightProbe = AmbientLightProbe;
declare const THREE_AmbientLightProbe: typeof AmbientLightProbe;
type THREE_HemisphereLightProbe = HemisphereLightProbe;
declare const THREE_HemisphereLightProbe: typeof HemisphereLightProbe;
type THREE_LightProbe = LightProbe;
declare const THREE_LightProbe: typeof LightProbe;
type THREE_AnimationLoader = AnimationLoader;
declare const THREE_AnimationLoader: typeof AnimationLoader;
type THREE_CompressedTextureLoader = CompressedTextureLoader;
declare const THREE_CompressedTextureLoader: typeof CompressedTextureLoader;
type THREE_DataTextureLoader = DataTextureLoader;
declare const THREE_DataTextureLoader: typeof DataTextureLoader;
type THREE_CubeTextureLoader = CubeTextureLoader;
declare const THREE_CubeTextureLoader: typeof CubeTextureLoader;
type THREE_TextureLoader = TextureLoader;
declare const THREE_TextureLoader: typeof TextureLoader;
type THREE_ObjectLoader = ObjectLoader;
declare const THREE_ObjectLoader: typeof ObjectLoader;
type THREE_MaterialLoader = MaterialLoader;
declare const THREE_MaterialLoader: typeof MaterialLoader;
type THREE_BufferGeometryLoader = BufferGeometryLoader;
declare const THREE_BufferGeometryLoader: typeof BufferGeometryLoader;
declare const THREE_DefaultLoadingManager: typeof DefaultLoadingManager;
type THREE_LoadingManager = LoadingManager;
declare const THREE_LoadingManager: typeof LoadingManager;
type THREE_ImageLoader = ImageLoader;
declare const THREE_ImageLoader: typeof ImageLoader;
type THREE_ImageBitmapLoader = ImageBitmapLoader;
declare const THREE_ImageBitmapLoader: typeof ImageBitmapLoader;
type THREE_FileLoader = FileLoader;
declare const THREE_FileLoader: typeof FileLoader;
type THREE_Loader = Loader;
declare const THREE_Loader: typeof Loader;
declare const THREE_LoaderUtils: typeof LoaderUtils;
declare const THREE_Cache: typeof Cache;
type THREE_AudioLoader = AudioLoader;
declare const THREE_AudioLoader: typeof AudioLoader;
type THREE_ShadowMaterialParameters = ShadowMaterialParameters;
type THREE_ShadowMaterial = ShadowMaterial;
declare const THREE_ShadowMaterial: typeof ShadowMaterial;
type THREE_SpriteMaterialParameters = SpriteMaterialParameters;
type THREE_SpriteMaterial = SpriteMaterial;
declare const THREE_SpriteMaterial: typeof SpriteMaterial;
type THREE_RawShaderMaterial = RawShaderMaterial;
declare const THREE_RawShaderMaterial: typeof RawShaderMaterial;
type THREE_ShaderMaterialParameters = ShaderMaterialParameters;
type THREE_ShaderMaterial = ShaderMaterial;
declare const THREE_ShaderMaterial: typeof ShaderMaterial;
type THREE_PointsMaterialParameters = PointsMaterialParameters;
type THREE_PointsMaterial = PointsMaterial;
declare const THREE_PointsMaterial: typeof PointsMaterial;
type THREE_MeshPhysicalMaterialParameters = MeshPhysicalMaterialParameters;
type THREE_MeshPhysicalMaterial = MeshPhysicalMaterial;
declare const THREE_MeshPhysicalMaterial: typeof MeshPhysicalMaterial;
type THREE_MeshStandardMaterialParameters = MeshStandardMaterialParameters;
type THREE_MeshStandardMaterial = MeshStandardMaterial;
declare const THREE_MeshStandardMaterial: typeof MeshStandardMaterial;
type THREE_MeshPhongMaterialParameters = MeshPhongMaterialParameters;
type THREE_MeshPhongMaterial = MeshPhongMaterial;
declare const THREE_MeshPhongMaterial: typeof MeshPhongMaterial;
type THREE_MeshToonMaterialParameters = MeshToonMaterialParameters;
type THREE_MeshToonMaterial = MeshToonMaterial;
declare const THREE_MeshToonMaterial: typeof MeshToonMaterial;
type THREE_MeshNormalMaterialParameters = MeshNormalMaterialParameters;
type THREE_MeshNormalMaterial = MeshNormalMaterial;
declare const THREE_MeshNormalMaterial: typeof MeshNormalMaterial;
type THREE_MeshLambertMaterialParameters = MeshLambertMaterialParameters;
type THREE_MeshLambertMaterial = MeshLambertMaterial;
declare const THREE_MeshLambertMaterial: typeof MeshLambertMaterial;
type THREE_MeshDepthMaterialParameters = MeshDepthMaterialParameters;
type THREE_MeshDepthMaterial = MeshDepthMaterial;
declare const THREE_MeshDepthMaterial: typeof MeshDepthMaterial;
type THREE_MeshDistanceMaterialParameters = MeshDistanceMaterialParameters;
type THREE_MeshDistanceMaterial = MeshDistanceMaterial;
declare const THREE_MeshDistanceMaterial: typeof MeshDistanceMaterial;
type THREE_MeshBasicMaterialParameters = MeshBasicMaterialParameters;
type THREE_MeshBasicMaterial = MeshBasicMaterial;
declare const THREE_MeshBasicMaterial: typeof MeshBasicMaterial;
type THREE_MeshMatcapMaterialParameters = MeshMatcapMaterialParameters;
type THREE_MeshMatcapMaterial = MeshMatcapMaterial;
declare const THREE_MeshMatcapMaterial: typeof MeshMatcapMaterial;
type THREE_LineDashedMaterialParameters = LineDashedMaterialParameters;
type THREE_LineDashedMaterial = LineDashedMaterial;
declare const THREE_LineDashedMaterial: typeof LineDashedMaterial;
type THREE_LineBasicMaterialParameters = LineBasicMaterialParameters;
type THREE_LineBasicMaterial = LineBasicMaterial;
declare const THREE_LineBasicMaterial: typeof LineBasicMaterial;
type THREE_MaterialParameters = MaterialParameters;
type THREE_Material = Material;
declare const THREE_Material: typeof Material;
type THREE_QuaternionLinearInterpolant = QuaternionLinearInterpolant;
declare const THREE_QuaternionLinearInterpolant: typeof QuaternionLinearInterpolant;
type THREE_LinearInterpolant = LinearInterpolant;
declare const THREE_LinearInterpolant: typeof LinearInterpolant;
type THREE_DiscreteInterpolant = DiscreteInterpolant;
declare const THREE_DiscreteInterpolant: typeof DiscreteInterpolant;
type THREE_CubicInterpolant = CubicInterpolant;
declare const THREE_CubicInterpolant: typeof CubicInterpolant;
type THREE_Interpolant = Interpolant;
declare const THREE_Interpolant: typeof Interpolant;
type THREE_Triangle = Triangle;
declare const THREE_Triangle: typeof Triangle;
type THREE_Spherical = Spherical;
declare const THREE_Spherical: typeof Spherical;
type THREE_Cylindrical = Cylindrical;
declare const THREE_Cylindrical: typeof Cylindrical;
type THREE_Plane = Plane;
declare const THREE_Plane: typeof Plane;
type THREE_Frustum = Frustum;
declare const THREE_Frustum: typeof Frustum;
type THREE_Sphere = Sphere;
declare const THREE_Sphere: typeof Sphere;
type THREE_Ray = Ray;
declare const THREE_Ray: typeof Ray;
type THREE_Matrix4Tuple = Matrix4Tuple;
type THREE_Matrix4 = Matrix4;
declare const THREE_Matrix4: typeof Matrix4;
type THREE_Matrix3Tuple = Matrix3Tuple;
type THREE_Matrix = Matrix;
type THREE_Matrix3 = Matrix3;
declare const THREE_Matrix3: typeof Matrix3;
type THREE_Box3 = Box3;
declare const THREE_Box3: typeof Box3;
type THREE_Box2 = Box2;
declare const THREE_Box2: typeof Box2;
type THREE_Line3 = Line3;
declare const THREE_Line3: typeof Line3;
type THREE_Euler = Euler;
declare const THREE_Euler: typeof Euler;
type THREE_Vector4Tuple = Vector4Tuple;
type THREE_Vector4 = Vector4;
declare const THREE_Vector4: typeof Vector4;
type THREE_Vector3Tuple = Vector3Tuple;
type THREE_Vector3 = Vector3;
declare const THREE_Vector3: typeof Vector3;
type THREE_Vector2Tuple = Vector2Tuple;
type THREE_Vector = Vector;
type THREE_Vector2 = Vector2;
declare const THREE_Vector2: typeof Vector2;
type THREE_Quaternion = Quaternion;
declare const THREE_Quaternion: typeof Quaternion;
declare const THREE_SRGBToLinear: typeof SRGBToLinear;
type THREE_HSL = HSL;
type THREE_RGB = RGB;
type THREE_Color = Color;
declare const THREE_Color: typeof Color;
type THREE_SphericalHarmonics3 = SphericalHarmonics3;
declare const THREE_SphericalHarmonics3: typeof SphericalHarmonics3;
type THREE_Sprite = Sprite;
declare const THREE_Sprite: typeof Sprite;
type THREE_LOD = LOD;
declare const THREE_LOD: typeof LOD;
type THREE_InstancedMesh<TGeometry extends BufferGeometry = BufferGeometry, TMaterial extends Material | Material[] = Material | Material[]> = InstancedMesh<TGeometry, TMaterial>;
declare const THREE_InstancedMesh: typeof InstancedMesh;
type THREE_SkinnedMesh<TGeometry extends BufferGeometry = BufferGeometry, TMaterial extends Material | Material[] = Material | Material[]> = SkinnedMesh<TGeometry, TMaterial>;
declare const THREE_SkinnedMesh: typeof SkinnedMesh;
type THREE_Skeleton = Skeleton;
declare const THREE_Skeleton: typeof Skeleton;
type THREE_Bone = Bone;
declare const THREE_Bone: typeof Bone;
type THREE_Mesh<TGeometry extends BufferGeometry = BufferGeometry, TMaterial extends Material | Material[] = Material | Material[]> = Mesh<TGeometry, TMaterial>;
declare const THREE_Mesh: typeof Mesh;
declare const THREE_LineStrip: typeof LineStrip;
declare const THREE_LinePieces: typeof LinePieces;
type THREE_LineSegments<TGeometry extends BufferGeometry = BufferGeometry, TMaterial extends Material | Material[] = Material | Material[]> = LineSegments<TGeometry, TMaterial>;
declare const THREE_LineSegments: typeof LineSegments;
type THREE_LineLoop<TGeometry extends BufferGeometry = BufferGeometry, TMaterial extends Material | Material[] = Material | Material[]> = LineLoop<TGeometry, TMaterial>;
declare const THREE_LineLoop: typeof LineLoop;
type THREE_Line<TGeometry extends BufferGeometry = BufferGeometry, TMaterial extends Material | Material[] = Material | Material[]> = Line<TGeometry, TMaterial>;
declare const THREE_Line: typeof Line;
type THREE_Points<TGeometry extends BufferGeometry = BufferGeometry, TMaterial extends Material | Material[] = Material | Material[]> = Points<TGeometry, TMaterial>;
declare const THREE_Points: typeof Points;
type THREE_Group = Group;
declare const THREE_Group: typeof Group;
type THREE_WebGLMultisampleRenderTarget = WebGLMultisampleRenderTarget;
declare const THREE_WebGLMultisampleRenderTarget: typeof WebGLMultisampleRenderTarget;
type THREE_WebGLCubeRenderTarget = WebGLCubeRenderTarget;
declare const THREE_WebGLCubeRenderTarget: typeof WebGLCubeRenderTarget;
type THREE_WebGLMultipleRenderTargets = WebGLMultipleRenderTargets;
declare const THREE_WebGLMultipleRenderTargets: typeof WebGLMultipleRenderTargets;
type THREE_WebGLRenderTargetOptions = WebGLRenderTargetOptions;
type THREE_WebGLRenderTarget = WebGLRenderTarget;
declare const THREE_WebGLRenderTarget: typeof WebGLRenderTarget;
type THREE_Renderer = Renderer;
type THREE_OffscreenCanvas = OffscreenCanvas;
type THREE_WebGLRendererParameters = WebGLRendererParameters;
type THREE_WebGLDebug = WebGLDebug;
type THREE_WebGLRenderer = WebGLRenderer;
declare const THREE_WebGLRenderer: typeof WebGLRenderer;
type THREE_WebGL1Renderer = WebGL1Renderer;
declare const THREE_WebGL1Renderer: typeof WebGL1Renderer;
type THREE_WebGL3DRenderTarget = WebGL3DRenderTarget;
declare const THREE_WebGL3DRenderTarget: typeof WebGL3DRenderTarget;
type THREE_WebGLArrayRenderTarget = WebGLArrayRenderTarget;
declare const THREE_WebGLArrayRenderTarget: typeof WebGLArrayRenderTarget;
type THREE_Shader = Shader;
declare const THREE_ShaderLib: typeof ShaderLib;
type THREE_IUniform<TValue = any> = IUniform<TValue>;
declare const THREE_UniformsLib: typeof UniformsLib;
declare const THREE_cloneUniforms: typeof cloneUniforms;
declare const THREE_mergeUniforms: typeof mergeUniforms;
declare const THREE_UniformsUtils: typeof UniformsUtils;
declare const THREE_ShaderChunk: typeof ShaderChunk;
type THREE_WebGLBufferRenderer = WebGLBufferRenderer;
declare const THREE_WebGLBufferRenderer: typeof WebGLBufferRenderer;
type THREE_WebGLCapabilitiesParameters = WebGLCapabilitiesParameters;
type THREE_WebGLCapabilities = WebGLCapabilities;
declare const THREE_WebGLCapabilities: typeof WebGLCapabilities;
type THREE_WebGLClipping = WebGLClipping;
declare const THREE_WebGLClipping: typeof WebGLClipping;
type THREE_WebGLCubeUVMaps = WebGLCubeUVMaps;
declare const THREE_WebGLCubeUVMaps: typeof WebGLCubeUVMaps;
type THREE_WebGLExtensions = WebGLExtensions;
declare const THREE_WebGLExtensions: typeof WebGLExtensions;
type THREE_WebGLGeometries = WebGLGeometries;
declare const THREE_WebGLGeometries: typeof WebGLGeometries;
type THREE_WebGLIndexedBufferRenderer = WebGLIndexedBufferRenderer;
declare const THREE_WebGLIndexedBufferRenderer: typeof WebGLIndexedBufferRenderer;
type THREE_WebGLInfo = WebGLInfo;
declare const THREE_WebGLInfo: typeof WebGLInfo;
type THREE_WebGLLights = WebGLLights;
declare const THREE_WebGLLights: typeof WebGLLights;
type THREE_WebGLObjects = WebGLObjects;
declare const THREE_WebGLObjects: typeof WebGLObjects;
type THREE_WebGLProgram = WebGLProgram;
declare const THREE_WebGLProgram: typeof WebGLProgram;
type THREE_WebGLPrograms = WebGLPrograms;
declare const THREE_WebGLPrograms: typeof WebGLPrograms;
type THREE_WebGLProperties = WebGLProperties;
declare const THREE_WebGLProperties: typeof WebGLProperties;
type THREE_RenderItem = RenderItem;
type THREE_WebGLRenderList = WebGLRenderList;
declare const THREE_WebGLRenderList: typeof WebGLRenderList;
type THREE_WebGLRenderLists = WebGLRenderLists;
declare const THREE_WebGLRenderLists: typeof WebGLRenderLists;
declare const THREE_WebGLShader: typeof WebGLShader;
type THREE_WebGLShadowMap = WebGLShadowMap;
declare const THREE_WebGLShadowMap: typeof WebGLShadowMap;
type THREE_WebGLColorBuffer = WebGLColorBuffer;
declare const THREE_WebGLColorBuffer: typeof WebGLColorBuffer;
type THREE_WebGLDepthBuffer = WebGLDepthBuffer;
declare const THREE_WebGLDepthBuffer: typeof WebGLDepthBuffer;
type THREE_WebGLStencilBuffer = WebGLStencilBuffer;
declare const THREE_WebGLStencilBuffer: typeof WebGLStencilBuffer;
type THREE_WebGLState = WebGLState;
declare const THREE_WebGLState: typeof WebGLState;
type THREE_WebGLTextures = WebGLTextures;
declare const THREE_WebGLTextures: typeof WebGLTextures;
type THREE_WebGLUniforms = WebGLUniforms;
declare const THREE_WebGLUniforms: typeof WebGLUniforms;
type THREE_XRControllerEventType = XRControllerEventType;
type THREE_XRHandJoints = XRHandJoints;
type THREE_XRHandInputState = XRHandInputState;
type THREE_XRHandSpace = XRHandSpace;
declare const THREE_XRHandSpace: typeof XRHandSpace;
type THREE_XRTargetRaySpace = XRTargetRaySpace;
declare const THREE_XRTargetRaySpace: typeof XRTargetRaySpace;
type THREE_XRGripSpace = XRGripSpace;
declare const THREE_XRGripSpace: typeof XRGripSpace;
type THREE_WebXRController = WebXRController;
declare const THREE_WebXRController: typeof WebXRController;
type THREE_WebXRCamera = WebXRCamera;
type THREE_WebXRArrayCamera = WebXRArrayCamera;
type THREE_WebXRManager = WebXRManager;
declare const THREE_WebXRManager: typeof WebXRManager;
type THREE_FogExp2 = FogExp2;
declare const THREE_FogExp2: typeof FogExp2;
type THREE_FogBase = FogBase;
type THREE_Fog = Fog;
declare const THREE_Fog: typeof Fog;
type THREE_VideoTexture = VideoTexture;
declare const THREE_VideoTexture: typeof VideoTexture;
type THREE_DataTexture = DataTexture;
declare const THREE_DataTexture: typeof DataTexture;
type THREE_DataTexture2DArray = DataTexture2DArray;
declare const THREE_DataTexture2DArray: typeof DataTexture2DArray;
type THREE_DataTexture3D = DataTexture3D;
declare const THREE_DataTexture3D: typeof DataTexture3D;
type THREE_CompressedTexture = CompressedTexture;
declare const THREE_CompressedTexture: typeof CompressedTexture;
type THREE_CubeTexture = CubeTexture;
declare const THREE_CubeTexture: typeof CubeTexture;
type THREE_Data3DTexture = Data3DTexture;
declare const THREE_Data3DTexture: typeof Data3DTexture;
type THREE_DataArrayTexture = DataArrayTexture;
declare const THREE_DataArrayTexture: typeof DataArrayTexture;
type THREE_CanvasTexture = CanvasTexture;
declare const THREE_CanvasTexture: typeof CanvasTexture;
type THREE_DepthTexture = DepthTexture;
declare const THREE_DepthTexture: typeof DepthTexture;
type THREE_FramebufferTexture = FramebufferTexture;
declare const THREE_FramebufferTexture: typeof FramebufferTexture;
type THREE_Source = Source;
declare const THREE_Source: typeof Source;
type THREE_Texture = Texture;
declare const THREE_Texture: typeof Texture;
declare namespace THREE {
  export {
    THREE_WebGLUtils as WebGLUtils,
    THREE_REVISION as REVISION,
    THREE_MOUSE as MOUSE,
    THREE_TOUCH as TOUCH,
    THREE_CullFace as CullFace,
    THREE_CullFaceNone as CullFaceNone,
    THREE_CullFaceBack as CullFaceBack,
    THREE_CullFaceFront as CullFaceFront,
    THREE_CullFaceFrontBack as CullFaceFrontBack,
    THREE_ShadowMapType as ShadowMapType,
    THREE_BasicShadowMap as BasicShadowMap,
    THREE_PCFShadowMap as PCFShadowMap,
    THREE_PCFSoftShadowMap as PCFSoftShadowMap,
    THREE_VSMShadowMap as VSMShadowMap,
    THREE_Side as Side,
    THREE_FrontSide as FrontSide,
    THREE_BackSide as BackSide,
    THREE_DoubleSide as DoubleSide,
    THREE_Shading as Shading,
    THREE_FlatShading as FlatShading,
    THREE_SmoothShading as SmoothShading,
    THREE_Blending as Blending,
    THREE_NoBlending as NoBlending,
    THREE_NormalBlending as NormalBlending,
    THREE_AdditiveBlending as AdditiveBlending,
    THREE_SubtractiveBlending as SubtractiveBlending,
    THREE_MultiplyBlending as MultiplyBlending,
    THREE_CustomBlending as CustomBlending,
    THREE_BlendingEquation as BlendingEquation,
    THREE_AddEquation as AddEquation,
    THREE_SubtractEquation as SubtractEquation,
    THREE_ReverseSubtractEquation as ReverseSubtractEquation,
    THREE_MinEquation as MinEquation,
    THREE_MaxEquation as MaxEquation,
    THREE_BlendingDstFactor as BlendingDstFactor,
    THREE_ZeroFactor as ZeroFactor,
    THREE_OneFactor as OneFactor,
    THREE_SrcColorFactor as SrcColorFactor,
    THREE_OneMinusSrcColorFactor as OneMinusSrcColorFactor,
    THREE_SrcAlphaFactor as SrcAlphaFactor,
    THREE_OneMinusSrcAlphaFactor as OneMinusSrcAlphaFactor,
    THREE_DstAlphaFactor as DstAlphaFactor,
    THREE_OneMinusDstAlphaFactor as OneMinusDstAlphaFactor,
    THREE_DstColorFactor as DstColorFactor,
    THREE_OneMinusDstColorFactor as OneMinusDstColorFactor,
    THREE_BlendingSrcFactor as BlendingSrcFactor,
    THREE_SrcAlphaSaturateFactor as SrcAlphaSaturateFactor,
    THREE_DepthModes as DepthModes,
    THREE_NeverDepth as NeverDepth,
    THREE_AlwaysDepth as AlwaysDepth,
    THREE_LessDepth as LessDepth,
    THREE_LessEqualDepth as LessEqualDepth,
    THREE_EqualDepth as EqualDepth,
    THREE_GreaterEqualDepth as GreaterEqualDepth,
    THREE_GreaterDepth as GreaterDepth,
    THREE_NotEqualDepth as NotEqualDepth,
    THREE_Combine as Combine,
    THREE_MultiplyOperation as MultiplyOperation,
    THREE_MixOperation as MixOperation,
    THREE_AddOperation as AddOperation,
    THREE_ToneMapping as ToneMapping,
    THREE_NoToneMapping as NoToneMapping,
    THREE_LinearToneMapping as LinearToneMapping,
    THREE_ReinhardToneMapping as ReinhardToneMapping,
    THREE_CineonToneMapping as CineonToneMapping,
    THREE_ACESFilmicToneMapping as ACESFilmicToneMapping,
    THREE_CustomToneMapping as CustomToneMapping,
    THREE_Mapping as Mapping,
    THREE_UVMapping as UVMapping,
    THREE_CubeReflectionMapping as CubeReflectionMapping,
    THREE_CubeRefractionMapping as CubeRefractionMapping,
    THREE_EquirectangularReflectionMapping as EquirectangularReflectionMapping,
    THREE_EquirectangularRefractionMapping as EquirectangularRefractionMapping,
    THREE_CubeUVReflectionMapping as CubeUVReflectionMapping,
    THREE_Wrapping as Wrapping,
    THREE_RepeatWrapping as RepeatWrapping,
    THREE_ClampToEdgeWrapping as ClampToEdgeWrapping,
    THREE_MirroredRepeatWrapping as MirroredRepeatWrapping,
    THREE_TextureFilter as TextureFilter,
    THREE_NearestFilter as NearestFilter,
    THREE_NearestMipmapNearestFilter as NearestMipmapNearestFilter,
    THREE_NearestMipMapNearestFilter as NearestMipMapNearestFilter,
    THREE_NearestMipmapLinearFilter as NearestMipmapLinearFilter,
    THREE_NearestMipMapLinearFilter as NearestMipMapLinearFilter,
    THREE_LinearFilter as LinearFilter,
    THREE_LinearMipmapNearestFilter as LinearMipmapNearestFilter,
    THREE_LinearMipMapNearestFilter as LinearMipMapNearestFilter,
    THREE_LinearMipmapLinearFilter as LinearMipmapLinearFilter,
    THREE_LinearMipMapLinearFilter as LinearMipMapLinearFilter,
    THREE_TextureDataType as TextureDataType,
    THREE_UnsignedByteType as UnsignedByteType,
    THREE_ByteType as ByteType,
    THREE_ShortType as ShortType,
    THREE_UnsignedShortType as UnsignedShortType,
    THREE_IntType as IntType,
    THREE_UnsignedIntType as UnsignedIntType,
    THREE_FloatType as FloatType,
    THREE_HalfFloatType as HalfFloatType,
    THREE_UnsignedShort4444Type as UnsignedShort4444Type,
    THREE_UnsignedShort5551Type as UnsignedShort5551Type,
    THREE_UnsignedInt248Type as UnsignedInt248Type,
    THREE_PixelFormat as PixelFormat,
    THREE_AlphaFormat as AlphaFormat,
    THREE_RGBFormat as RGBFormat,
    THREE_RGBAFormat as RGBAFormat,
    THREE_LuminanceFormat as LuminanceFormat,
    THREE_LuminanceAlphaFormat as LuminanceAlphaFormat,
    THREE_DepthFormat as DepthFormat,
    THREE_DepthStencilFormat as DepthStencilFormat,
    THREE_RedFormat as RedFormat,
    THREE_RedIntegerFormat as RedIntegerFormat,
    THREE_RGFormat as RGFormat,
    THREE_RGIntegerFormat as RGIntegerFormat,
    THREE_RGBAIntegerFormat as RGBAIntegerFormat,
    THREE__SRGBFormat as _SRGBFormat,
    THREE__SRGBAFormat as _SRGBAFormat,
    THREE_PixelFormatGPU as PixelFormatGPU,
    THREE_CompressedPixelFormat as CompressedPixelFormat,
    THREE_RGB_S3TC_DXT1_Format as RGB_S3TC_DXT1_Format,
    THREE_RGBA_S3TC_DXT1_Format as RGBA_S3TC_DXT1_Format,
    THREE_RGBA_S3TC_DXT3_Format as RGBA_S3TC_DXT3_Format,
    THREE_RGBA_S3TC_DXT5_Format as RGBA_S3TC_DXT5_Format,
    THREE_RGB_PVRTC_4BPPV1_Format as RGB_PVRTC_4BPPV1_Format,
    THREE_RGB_PVRTC_2BPPV1_Format as RGB_PVRTC_2BPPV1_Format,
    THREE_RGBA_PVRTC_4BPPV1_Format as RGBA_PVRTC_4BPPV1_Format,
    THREE_RGBA_PVRTC_2BPPV1_Format as RGBA_PVRTC_2BPPV1_Format,
    THREE_RGB_ETC1_Format as RGB_ETC1_Format,
    THREE_RGB_ETC2_Format as RGB_ETC2_Format,
    THREE_RGBA_ETC2_EAC_Format as RGBA_ETC2_EAC_Format,
    THREE_RGBA_ASTC_4x4_Format as RGBA_ASTC_4x4_Format,
    THREE_RGBA_ASTC_5x4_Format as RGBA_ASTC_5x4_Format,
    THREE_RGBA_ASTC_5x5_Format as RGBA_ASTC_5x5_Format,
    THREE_RGBA_ASTC_6x5_Format as RGBA_ASTC_6x5_Format,
    THREE_RGBA_ASTC_6x6_Format as RGBA_ASTC_6x6_Format,
    THREE_RGBA_ASTC_8x5_Format as RGBA_ASTC_8x5_Format,
    THREE_RGBA_ASTC_8x6_Format as RGBA_ASTC_8x6_Format,
    THREE_RGBA_ASTC_8x8_Format as RGBA_ASTC_8x8_Format,
    THREE_RGBA_ASTC_10x5_Format as RGBA_ASTC_10x5_Format,
    THREE_RGBA_ASTC_10x6_Format as RGBA_ASTC_10x6_Format,
    THREE_RGBA_ASTC_10x8_Format as RGBA_ASTC_10x8_Format,
    THREE_RGBA_ASTC_10x10_Format as RGBA_ASTC_10x10_Format,
    THREE_RGBA_ASTC_12x10_Format as RGBA_ASTC_12x10_Format,
    THREE_RGBA_ASTC_12x12_Format as RGBA_ASTC_12x12_Format,
    THREE_RGBA_BPTC_Format as RGBA_BPTC_Format,
    THREE_AnimationActionLoopStyles as AnimationActionLoopStyles,
    THREE_LoopOnce as LoopOnce,
    THREE_LoopRepeat as LoopRepeat,
    THREE_LoopPingPong as LoopPingPong,
    THREE_InterpolationModes as InterpolationModes,
    THREE_InterpolateDiscrete as InterpolateDiscrete,
    THREE_InterpolateLinear as InterpolateLinear,
    THREE_InterpolateSmooth as InterpolateSmooth,
    THREE_InterpolationEndingModes as InterpolationEndingModes,
    THREE_ZeroCurvatureEnding as ZeroCurvatureEnding,
    THREE_ZeroSlopeEnding as ZeroSlopeEnding,
    THREE_WrapAroundEnding as WrapAroundEnding,
    THREE_AnimationBlendMode as AnimationBlendMode,
    THREE_NormalAnimationBlendMode as NormalAnimationBlendMode,
    THREE_AdditiveAnimationBlendMode as AdditiveAnimationBlendMode,
    THREE_TrianglesDrawModes as TrianglesDrawModes,
    THREE_TrianglesDrawMode as TrianglesDrawMode,
    THREE_TriangleStripDrawMode as TriangleStripDrawMode,
    THREE_TriangleFanDrawMode as TriangleFanDrawMode,
    THREE_TextureEncoding as TextureEncoding,
    THREE_LinearEncoding as LinearEncoding,
    THREE_sRGBEncoding as sRGBEncoding,
    THREE_DepthPackingStrategies as DepthPackingStrategies,
    THREE_BasicDepthPacking as BasicDepthPacking,
    THREE_RGBADepthPacking as RGBADepthPacking,
    THREE_NormalMapTypes as NormalMapTypes,
    THREE_TangentSpaceNormalMap as TangentSpaceNormalMap,
    THREE_ObjectSpaceNormalMap as ObjectSpaceNormalMap,
    THREE_ColorSpace as ColorSpace,
    THREE_NoColorSpace as NoColorSpace,
    THREE_SRGBColorSpace as SRGBColorSpace,
    THREE_LinearSRGBColorSpace as LinearSRGBColorSpace,
    THREE_StencilOp as StencilOp,
    THREE_ZeroStencilOp as ZeroStencilOp,
    THREE_KeepStencilOp as KeepStencilOp,
    THREE_ReplaceStencilOp as ReplaceStencilOp,
    THREE_IncrementStencilOp as IncrementStencilOp,
    THREE_DecrementStencilOp as DecrementStencilOp,
    THREE_IncrementWrapStencilOp as IncrementWrapStencilOp,
    THREE_DecrementWrapStencilOp as DecrementWrapStencilOp,
    THREE_InvertStencilOp as InvertStencilOp,
    THREE_StencilFunc as StencilFunc,
    THREE_NeverStencilFunc as NeverStencilFunc,
    THREE_LessStencilFunc as LessStencilFunc,
    THREE_EqualStencilFunc as EqualStencilFunc,
    THREE_LessEqualStencilFunc as LessEqualStencilFunc,
    THREE_GreaterStencilFunc as GreaterStencilFunc,
    THREE_NotEqualStencilFunc as NotEqualStencilFunc,
    THREE_GreaterEqualStencilFunc as GreaterEqualStencilFunc,
    THREE_AlwaysStencilFunc as AlwaysStencilFunc,
    THREE_Usage as Usage,
    THREE_StaticDrawUsage as StaticDrawUsage,
    THREE_DynamicDrawUsage as DynamicDrawUsage,
    THREE_StreamDrawUsage as StreamDrawUsage,
    THREE_StaticReadUsage as StaticReadUsage,
    THREE_DynamicReadUsage as DynamicReadUsage,
    THREE_StreamReadUsage as StreamReadUsage,
    THREE_StaticCopyUsage as StaticCopyUsage,
    THREE_DynamicCopyUsage as DynamicCopyUsage,
    THREE_StreamCopyUsage as StreamCopyUsage,
    THREE_GLSLVersion as GLSLVersion,
    THREE_GLSL1 as GLSL1,
    THREE_GLSL3 as GLSL3,
    THREE_BuiltinShaderAttributeName as BuiltinShaderAttributeName,
    THREE_ColorRepresentation as ColorRepresentation,
    THREE_VectorKeyframeTrack as VectorKeyframeTrack,
    THREE_StringKeyframeTrack as StringKeyframeTrack,
    THREE_QuaternionKeyframeTrack as QuaternionKeyframeTrack,
    THREE_NumberKeyframeTrack as NumberKeyframeTrack,
    THREE_ColorKeyframeTrack as ColorKeyframeTrack,
    THREE_BooleanKeyframeTrack as BooleanKeyframeTrack,
    THREE_PropertyMixer as PropertyMixer,
    THREE_ParseTrackNameResults as ParseTrackNameResults,
    THREE_PropertyBinding as PropertyBinding,
    THREE_KeyframeTrack as KeyframeTrack,
    THREE_AnimationUtils as AnimationUtils,
    THREE_AnimationObjectGroup as AnimationObjectGroup,
    THREE_AnimationMixer as AnimationMixer,
    THREE_MorphTarget as MorphTarget,
    THREE_AnimationClip as AnimationClip,
    THREE_AnimationAction as AnimationAction,
    THREE_AudioListener as AudioListener,
    THREE_PositionalAudio as PositionalAudio,
    THREE_AudioContext as AudioContext,
    THREE_AudioAnalyser as AudioAnalyser,
    THREE_Audio as Audio,
    THREE_StereoCamera as StereoCamera,
    THREE_PerspectiveCamera as PerspectiveCamera,
    THREE_OrthographicCamera as OrthographicCamera,
    THREE_CubeCamera as CubeCamera,
    THREE_ArrayCamera as ArrayCamera,
    THREE_Camera as Camera,
    THREE_Uniform as Uniform,
    THREE_InstancedBufferGeometry as InstancedBufferGeometry,
    THREE_BufferGeometry as BufferGeometry,
    THREE_InterleavedBufferAttribute as InterleavedBufferAttribute,
    THREE_InstancedInterleavedBuffer as InstancedInterleavedBuffer,
    THREE_InterleavedBuffer as InterleavedBuffer,
    THREE_BufferGeometryUtils as BufferGeometryUtils,
    THREE_GeometryUtils as GeometryUtils,
    THREE_InstancedBufferAttribute as InstancedBufferAttribute,
    THREE_GLBufferAttribute as GLBufferAttribute,
    THREE_BufferAttribute as BufferAttribute,
    THREE_Int8Attribute as Int8Attribute,
    THREE_Uint8Attribute as Uint8Attribute,
    THREE_Uint8ClampedAttribute as Uint8ClampedAttribute,
    THREE_Int16Attribute as Int16Attribute,
    THREE_Uint16Attribute as Uint16Attribute,
    THREE_Int32Attribute as Int32Attribute,
    THREE_Uint32Attribute as Uint32Attribute,
    THREE_Float32Attribute as Float32Attribute,
    THREE_Float64Attribute as Float64Attribute,
    THREE_Int8BufferAttribute as Int8BufferAttribute,
    THREE_Uint8BufferAttribute as Uint8BufferAttribute,
    THREE_Uint8ClampedBufferAttribute as Uint8ClampedBufferAttribute,
    THREE_Int16BufferAttribute as Int16BufferAttribute,
    THREE_Uint16BufferAttribute as Uint16BufferAttribute,
    THREE_Int32BufferAttribute as Int32BufferAttribute,
    THREE_Uint32BufferAttribute as Uint32BufferAttribute,
    THREE_Float16BufferAttribute as Float16BufferAttribute,
    THREE_Float32BufferAttribute as Float32BufferAttribute,
    THREE_Float64BufferAttribute as Float64BufferAttribute,
    THREE_Object3D as Object3D,
    THREE_Face as Face,
    THREE_Intersection as Intersection,
    THREE_RaycasterParameters as RaycasterParameters,
    THREE_Raycaster as Raycaster,
    THREE_Layers as Layers,
    THREE_BaseEvent as BaseEvent,
    Event$1 as Event,
    THREE_EventListener as EventListener,
    THREE_EventDispatcher as EventDispatcher,
    THREE_Clock as Clock,
    THREE_ArcCurve as ArcCurve,
    THREE_CurveUtils as CurveUtils,
    THREE_CatmullRomCurve3 as CatmullRomCurve3,
    THREE_CubicBezierCurve as CubicBezierCurve,
    THREE_CubicBezierCurve3 as CubicBezierCurve3,
    THREE_EllipseCurve as EllipseCurve,
    THREE_LineCurve as LineCurve,
    THREE_LineCurve3 as LineCurve3,
    THREE_QuadraticBezierCurve as QuadraticBezierCurve,
    THREE_QuadraticBezierCurve3 as QuadraticBezierCurve3,
    THREE_SplineCurve as SplineCurve,
    THREE_Shape as Shape,
    THREE_Path as Path,
    THREE_ShapePath as ShapePath,
    THREE_CurvePath as CurvePath,
    THREE_Curve as Curve,
    THREE_toHalfFloat as toHalfFloat,
    THREE_fromHalfFloat as fromHalfFloat,
    THREE_Earcut as Earcut,
    THREE_ImageUtils as ImageUtils,
    THREE_Vec2 as Vec2,
    THREE_ShapeUtils as ShapeUtils,
    THREE_PMREMGenerator as PMREMGenerator,
    BoxGeometry as BoxBufferGeometry,
    THREE_BoxGeometry as BoxGeometry,
    CapsuleGeometry as CapsuleBufferGeometry,
    THREE_CapsuleGeometry as CapsuleGeometry,
    CircleGeometry as CircleBufferGeometry,
    THREE_CircleGeometry as CircleGeometry,
    ConeGeometry as ConeBufferGeometry,
    THREE_ConeGeometry as ConeGeometry,
    CylinderGeometry as CylinderBufferGeometry,
    THREE_CylinderGeometry as CylinderGeometry,
    DodecahedronGeometry as DodecahedronBufferGeometry,
    THREE_DodecahedronGeometry as DodecahedronGeometry,
    THREE_EdgesGeometry as EdgesGeometry,
    ExtrudeGeometry as ExtrudeBufferGeometry,
    THREE_ExtrudeGeometryOptions as ExtrudeGeometryOptions,
    THREE_UVGenerator as UVGenerator,
    THREE_ExtrudeGeometry as ExtrudeGeometry,
    IcosahedronGeometry as IcosahedronBufferGeometry,
    THREE_IcosahedronGeometry as IcosahedronGeometry,
    LatheGeometry as LatheBufferGeometry,
    THREE_LatheGeometry as LatheGeometry,
    OctahedronGeometry as OctahedronBufferGeometry,
    THREE_OctahedronGeometry as OctahedronGeometry,
    PlaneGeometry as PlaneBufferGeometry,
    THREE_PlaneGeometry as PlaneGeometry,
    PolyhedronGeometry as PolyhedronBufferGeometry,
    THREE_PolyhedronGeometry as PolyhedronGeometry,
    RingGeometry as RingBufferGeometry,
    THREE_RingGeometry as RingGeometry,
    ShapeGeometry as ShapeBufferGeometry,
    THREE_ShapeGeometry as ShapeGeometry,
    SphereGeometry as SphereBufferGeometry,
    THREE_SphereGeometry as SphereGeometry,
    TetrahedronGeometry as TetrahedronBufferGeometry,
    THREE_TetrahedronGeometry as TetrahedronGeometry,
    TorusGeometry as TorusBufferGeometry,
    THREE_TorusGeometry as TorusGeometry,
    TorusKnotGeometry as TorusKnotBufferGeometry,
    THREE_TorusKnotGeometry as TorusKnotGeometry,
    TubeGeometry as TubeBufferGeometry,
    THREE_TubeGeometry as TubeGeometry,
    THREE_WireframeGeometry as WireframeGeometry,
    THREE_SpotLightHelper as SpotLightHelper,
    THREE_SkeletonHelper as SkeletonHelper,
    THREE_PointLightHelper as PointLightHelper,
    THREE_HemisphereLightHelper as HemisphereLightHelper,
    THREE_GridHelper as GridHelper,
    THREE_PolarGridHelper as PolarGridHelper,
    THREE_DirectionalLightHelper as DirectionalLightHelper,
    THREE_CameraHelper as CameraHelper,
    THREE_BoxHelper as BoxHelper,
    THREE_Box3Helper as Box3Helper,
    THREE_PlaneHelper as PlaneHelper,
    THREE_ArrowHelper as ArrowHelper,
    THREE_AxesHelper as AxesHelper,
    THREE_SpotLightShadow as SpotLightShadow,
    THREE_SpotLight as SpotLight,
    THREE_PointLight as PointLight,
    THREE_PointLightShadow as PointLightShadow,
    THREE_RectAreaLight as RectAreaLight,
    THREE_HemisphereLight as HemisphereLight,
    THREE_DirectionalLightShadow as DirectionalLightShadow,
    THREE_DirectionalLight as DirectionalLight,
    THREE_AmbientLight as AmbientLight,
    THREE_LightShadow as LightShadow,
    THREE_Light as Light,
    THREE_AmbientLightProbe as AmbientLightProbe,
    THREE_HemisphereLightProbe as HemisphereLightProbe,
    THREE_LightProbe as LightProbe,
    THREE_AnimationLoader as AnimationLoader,
    THREE_CompressedTextureLoader as CompressedTextureLoader,
    THREE_DataTextureLoader as DataTextureLoader,
    THREE_CubeTextureLoader as CubeTextureLoader,
    THREE_TextureLoader as TextureLoader,
    THREE_ObjectLoader as ObjectLoader,
    THREE_MaterialLoader as MaterialLoader,
    THREE_BufferGeometryLoader as BufferGeometryLoader,
    THREE_DefaultLoadingManager as DefaultLoadingManager,
    THREE_LoadingManager as LoadingManager,
    THREE_ImageLoader as ImageLoader,
    THREE_ImageBitmapLoader as ImageBitmapLoader,
    THREE_FileLoader as FileLoader,
    THREE_Loader as Loader,
    THREE_LoaderUtils as LoaderUtils,
    THREE_Cache as Cache,
    THREE_AudioLoader as AudioLoader,
    THREE_ShadowMaterialParameters as ShadowMaterialParameters,
    THREE_ShadowMaterial as ShadowMaterial,
    THREE_SpriteMaterialParameters as SpriteMaterialParameters,
    THREE_SpriteMaterial as SpriteMaterial,
    THREE_RawShaderMaterial as RawShaderMaterial,
    THREE_ShaderMaterialParameters as ShaderMaterialParameters,
    THREE_ShaderMaterial as ShaderMaterial,
    THREE_PointsMaterialParameters as PointsMaterialParameters,
    THREE_PointsMaterial as PointsMaterial,
    THREE_MeshPhysicalMaterialParameters as MeshPhysicalMaterialParameters,
    THREE_MeshPhysicalMaterial as MeshPhysicalMaterial,
    THREE_MeshStandardMaterialParameters as MeshStandardMaterialParameters,
    THREE_MeshStandardMaterial as MeshStandardMaterial,
    THREE_MeshPhongMaterialParameters as MeshPhongMaterialParameters,
    THREE_MeshPhongMaterial as MeshPhongMaterial,
    THREE_MeshToonMaterialParameters as MeshToonMaterialParameters,
    THREE_MeshToonMaterial as MeshToonMaterial,
    THREE_MeshNormalMaterialParameters as MeshNormalMaterialParameters,
    THREE_MeshNormalMaterial as MeshNormalMaterial,
    THREE_MeshLambertMaterialParameters as MeshLambertMaterialParameters,
    THREE_MeshLambertMaterial as MeshLambertMaterial,
    THREE_MeshDepthMaterialParameters as MeshDepthMaterialParameters,
    THREE_MeshDepthMaterial as MeshDepthMaterial,
    THREE_MeshDistanceMaterialParameters as MeshDistanceMaterialParameters,
    THREE_MeshDistanceMaterial as MeshDistanceMaterial,
    THREE_MeshBasicMaterialParameters as MeshBasicMaterialParameters,
    THREE_MeshBasicMaterial as MeshBasicMaterial,
    THREE_MeshMatcapMaterialParameters as MeshMatcapMaterialParameters,
    THREE_MeshMatcapMaterial as MeshMatcapMaterial,
    THREE_LineDashedMaterialParameters as LineDashedMaterialParameters,
    THREE_LineDashedMaterial as LineDashedMaterial,
    THREE_LineBasicMaterialParameters as LineBasicMaterialParameters,
    THREE_LineBasicMaterial as LineBasicMaterial,
    THREE_MaterialParameters as MaterialParameters,
    THREE_Material as Material,
    THREE_QuaternionLinearInterpolant as QuaternionLinearInterpolant,
    THREE_LinearInterpolant as LinearInterpolant,
    THREE_DiscreteInterpolant as DiscreteInterpolant,
    THREE_CubicInterpolant as CubicInterpolant,
    THREE_Interpolant as Interpolant,
    THREE_Triangle as Triangle,
    THREE_Spherical as Spherical,
    THREE_Cylindrical as Cylindrical,
    THREE_Plane as Plane,
    THREE_Frustum as Frustum,
    THREE_Sphere as Sphere,
    THREE_Ray as Ray,
    THREE_Matrix4Tuple as Matrix4Tuple,
    THREE_Matrix4 as Matrix4,
    THREE_Matrix3Tuple as Matrix3Tuple,
    THREE_Matrix as Matrix,
    THREE_Matrix3 as Matrix3,
    THREE_Box3 as Box3,
    THREE_Box2 as Box2,
    THREE_Line3 as Line3,
    THREE_Euler as Euler,
    THREE_Vector4Tuple as Vector4Tuple,
    THREE_Vector4 as Vector4,
    THREE_Vector3Tuple as Vector3Tuple,
    THREE_Vector3 as Vector3,
    THREE_Vector2Tuple as Vector2Tuple,
    THREE_Vector as Vector,
    THREE_Vector2 as Vector2,
    THREE_Quaternion as Quaternion,
    THREE_SRGBToLinear as SRGBToLinear,
    THREE_HSL as HSL,
    THREE_RGB as RGB,
    THREE_Color as Color,
    THREE_SphericalHarmonics3 as SphericalHarmonics3,
    THREE_Sprite as Sprite,
    THREE_LOD as LOD,
    THREE_InstancedMesh as InstancedMesh,
    THREE_SkinnedMesh as SkinnedMesh,
    THREE_Skeleton as Skeleton,
    THREE_Bone as Bone,
    THREE_Mesh as Mesh,
    THREE_LineStrip as LineStrip,
    THREE_LinePieces as LinePieces,
    THREE_LineSegments as LineSegments,
    THREE_LineLoop as LineLoop,
    THREE_Line as Line,
    THREE_Points as Points,
    THREE_Group as Group,
    THREE_WebGLMultisampleRenderTarget as WebGLMultisampleRenderTarget,
    THREE_WebGLCubeRenderTarget as WebGLCubeRenderTarget,
    THREE_WebGLMultipleRenderTargets as WebGLMultipleRenderTargets,
    THREE_WebGLRenderTargetOptions as WebGLRenderTargetOptions,
    THREE_WebGLRenderTarget as WebGLRenderTarget,
    THREE_Renderer as Renderer,
    THREE_OffscreenCanvas as OffscreenCanvas,
    THREE_WebGLRendererParameters as WebGLRendererParameters,
    THREE_WebGLDebug as WebGLDebug,
    THREE_WebGLRenderer as WebGLRenderer,
    THREE_WebGL1Renderer as WebGL1Renderer,
    THREE_WebGL3DRenderTarget as WebGL3DRenderTarget,
    THREE_WebGLArrayRenderTarget as WebGLArrayRenderTarget,
    THREE_Shader as Shader,
    THREE_ShaderLib as ShaderLib,
    THREE_IUniform as IUniform,
    THREE_UniformsLib as UniformsLib,
    THREE_cloneUniforms as cloneUniforms,
    THREE_mergeUniforms as mergeUniforms,
    THREE_UniformsUtils as UniformsUtils,
    THREE_ShaderChunk as ShaderChunk,
    THREE_WebGLBufferRenderer as WebGLBufferRenderer,
    THREE_WebGLCapabilitiesParameters as WebGLCapabilitiesParameters,
    THREE_WebGLCapabilities as WebGLCapabilities,
    THREE_WebGLClipping as WebGLClipping,
    THREE_WebGLCubeUVMaps as WebGLCubeUVMaps,
    THREE_WebGLExtensions as WebGLExtensions,
    THREE_WebGLGeometries as WebGLGeometries,
    THREE_WebGLIndexedBufferRenderer as WebGLIndexedBufferRenderer,
    THREE_WebGLInfo as WebGLInfo,
    THREE_WebGLLights as WebGLLights,
    THREE_WebGLObjects as WebGLObjects,
    THREE_WebGLProgram as WebGLProgram,
    THREE_WebGLPrograms as WebGLPrograms,
    THREE_WebGLProperties as WebGLProperties,
    THREE_RenderItem as RenderItem,
    THREE_WebGLRenderList as WebGLRenderList,
    THREE_WebGLRenderLists as WebGLRenderLists,
    THREE_WebGLShader as WebGLShader,
    THREE_WebGLShadowMap as WebGLShadowMap,
    THREE_WebGLColorBuffer as WebGLColorBuffer,
    THREE_WebGLDepthBuffer as WebGLDepthBuffer,
    THREE_WebGLStencilBuffer as WebGLStencilBuffer,
    THREE_WebGLState as WebGLState,
    THREE_WebGLTextures as WebGLTextures,
    THREE_WebGLUniforms as WebGLUniforms,
    THREE_XRControllerEventType as XRControllerEventType,
    THREE_XRHandJoints as XRHandJoints,
    THREE_XRHandInputState as XRHandInputState,
    THREE_XRHandSpace as XRHandSpace,
    THREE_XRTargetRaySpace as XRTargetRaySpace,
    THREE_XRGripSpace as XRGripSpace,
    THREE_WebXRController as WebXRController,
    THREE_WebXRCamera as WebXRCamera,
    THREE_WebXRArrayCamera as WebXRArrayCamera,
    THREE_WebXRManager as WebXRManager,
    THREE_FogExp2 as FogExp2,
    THREE_FogBase as FogBase,
    THREE_Fog as Fog,
    Scene$1 as Scene,
    THREE_VideoTexture as VideoTexture,
    THREE_DataTexture as DataTexture,
    THREE_DataTexture2DArray as DataTexture2DArray,
    THREE_DataTexture3D as DataTexture3D,
    THREE_CompressedTexture as CompressedTexture,
    THREE_CubeTexture as CubeTexture,
    THREE_Data3DTexture as Data3DTexture,
    THREE_DataArrayTexture as DataArrayTexture,
    THREE_CanvasTexture as CanvasTexture,
    THREE_DepthTexture as DepthTexture,
    THREE_FramebufferTexture as FramebufferTexture,
    THREE_Source as Source,
    THREE_Texture as Texture,
    MathUtils_d as MathUtils,
  };
}

declare class OrbitControls {
    constructor(object: Camera, domElement?: HTMLElement);

    object: Camera;
    domElement: HTMLElement | Document;

    // API
    enabled: boolean;
    target: Vector3;

    // deprecated
    center: Vector3;

    minDistance: number;
    maxDistance: number;

    minZoom: number;
    maxZoom: number;

    minPolarAngle: number;
    maxPolarAngle: number;

    minAzimuthAngle: number;
    maxAzimuthAngle: number;

    enableDamping: boolean;
    dampingFactor: number;

    enableZoom: boolean;
    zoomSpeed: number;

    enableRotate: boolean;
    rotateSpeed: number;

    enablePan: boolean;
    panSpeed: number;
    screenSpacePanning: boolean;
    keyPanSpeed: number;

    autoRotate: boolean;
    autoRotateSpeed: number;

    keys: { LEFT: string; UP: string; RIGHT: string; BOTTOM: string };
    mouseButtons: Partial<{ LEFT: MOUSE; MIDDLE: MOUSE; RIGHT: MOUSE }>;
    touches: Partial<{ ONE: TOUCH; TWO: TOUCH }>;

    target0: Vector3;
    position0: Vector3;
    zoomO: number;

    update(): boolean;

    listenToKeyEvents(domElement: HTMLElement | Window): void;

    saveState(): void;

    reset(): void;

    dispose(): void;

    getPolarAngle(): number;

    getAzimuthalAngle(): number;

    getDistance(): number;

    // EventDispatcher mixins
    addEventListener(type: string, listener: (event: any) => void): void;

    hasEventListener(type: string, listener: (event: any) => void): boolean;

    removeEventListener(type: string, listener: (event: any) => void): void;

    dispatchEvent(event: { type: string; target: any }): void;
}

declare class DRACOLoader extends Loader {
    constructor(manager?: LoadingManager);

    load(
        url: string,
        onLoad: (geometry: BufferGeometry) => void,
        onProgress?: (event: ProgressEvent) => void,
        onError?: (event: ErrorEvent) => void,
    ): void;
    loadAsync(url: string, onProgress?: (event: ProgressEvent) => void): Promise<BufferGeometry>;
    setDecoderPath(path: string): DRACOLoader;
    setDecoderConfig(config: object): DRACOLoader;
    setWorkerLimit(workerLimit: number): DRACOLoader;
    preload(): DRACOLoader;
    dispose(): DRACOLoader;
}

declare class KTX2Loader extends CompressedTextureLoader {
    constructor(manager?: LoadingManager);

    setTranscoderPath(path: string): KTX2Loader;
    setWorkerLimit(limit: number): KTX2Loader;
    detectSupport(renderer: WebGLRenderer): KTX2Loader;
    dispose(): KTX2Loader;

    parse(
        buffer: ArrayBuffer,
        onLoad: (texture: CompressedTexture) => void,
        onError?: (event: ErrorEvent) => void,
    ): KTX2Loader;
}

interface GLTF {
    animations: AnimationClip[];
    scene: Group;
    scenes: Group[];
    cameras: Camera[];
    asset: {
        copyright?: string | undefined;
        generator?: string | undefined;
        version?: string | undefined;
        minVersion?: string | undefined;
        extensions?: any;
        extras?: any;
    };
    parser: GLTFParser;
    userData: any;
}

declare class GLTFLoader extends Loader {
    constructor(manager?: LoadingManager);
    dracoLoader: DRACOLoader | null;

    load(
        url: string,
        onLoad: (gltf: GLTF) => void,
        onProgress?: (event: ProgressEvent) => void,
        onError?: (event: ErrorEvent) => void,
    ): void;
    loadAsync(url: string, onProgress?: (event: ProgressEvent) => void): Promise<GLTF>;

    setDRACOLoader(dracoLoader: DRACOLoader): GLTFLoader;

    register(callback: (parser: GLTFParser) => GLTFLoaderPlugin): GLTFLoader;
    unregister(callback: (parser: GLTFParser) => GLTFLoaderPlugin): GLTFLoader;

    setKTX2Loader(ktx2Loader: KTX2Loader): GLTFLoader;
    setMeshoptDecoder(meshoptDecoder: /* MeshoptDecoder */ any): GLTFLoader;

    parse(
        data: ArrayBuffer | string,
        path: string,
        onLoad: (gltf: GLTF) => void,
        onError?: (event: ErrorEvent) => void,
    ): void;

    parseAsync(data: ArrayBuffer | string, path: string): Promise<GLTF>;
}

interface GLTFReference {
    materials?: number;
    nodes?: number;
    textures?: number;
    meshes?: number;
}

declare class GLTFParser {
    json: any;

    options: {
        path: string;
        manager: LoadingManager;
        ktx2Loader: KTX2Loader;
        meshoptDecoder: /* MeshoptDecoder */ any;
        crossOrigin: string;
        requestHeader: { [header: string]: string };
    };

    fileLoader: FileLoader;
    textureLoader: TextureLoader | ImageBitmapLoader;
    plugins: { [name: string]: GLTFLoaderPlugin };
    extensions: { [name: string]: any };
    associations: Map<Object3D | Material | Texture, GLTFReference>;

    setExtensions(extensions: { [name: string]: any }): void;
    setPlugins(plugins: { [name: string]: GLTFLoaderPlugin }): void;

    parse(onLoad: (gltf: GLTF) => void, onError?: (event: ErrorEvent) => void): void;

    getDependency: (type: string, index: number) => Promise<any>;
    getDependencies: (type: string) => Promise<any[]>;
    loadBuffer: (bufferIndex: number) => Promise<ArrayBuffer>;
    loadBufferView: (bufferViewIndex: number) => Promise<ArrayBuffer>;
    loadAccessor: (accessorIndex: number) => Promise<BufferAttribute | InterleavedBufferAttribute>;
    loadTexture: (textureIndex: number) => Promise<Texture>;
    loadTextureImage: (textureIndex: number, sourceIndex: number, loader: Loader) => Promise<Texture>;
    loadImageSource: (sourceIndex: number, loader: Loader) => Promise<Texture>;
    assignTexture: (
        materialParams: { [key: string]: any },
        mapName: string,
        mapDef: {
            index: number;
            texCoord?: number | undefined;
            extensions?: any;
        },
    ) => Promise<void>;
    assignFinalMaterial: (object: Mesh) => void;
    getMaterialType: () => typeof MeshStandardMaterial;
    loadMaterial: (materialIndex: number) => Promise<Material>;
    createUniqueName: (originalName: string) => string;
    createNodeMesh: (nodeIndex: number) => Promise<Group | Mesh | SkinnedMesh>;
    loadGeometries: (
        /**
         * GLTF.Primitive[]
         * See: https://github.com/KhronosGroup/glTF/blob/master/specification/2.0/schema/mesh.primitive.schema.json
         */
        primitives: Array<{ [key: string]: any }>,
    ) => Promise<BufferGeometry[]>;
    loadMesh: (meshIndex: number) => Promise<Group | Mesh | SkinnedMesh>;
    loadCamera: (cameraIndex: number) => Promise<Camera>;
    loadSkin: (skinIndex: number) => Promise<{
        joints: number[];
        inverseBindMatrices?: BufferAttribute | InterleavedBufferAttribute | undefined;
    }>;
    loadAnimation: (animationIndex: number) => Promise<AnimationClip>;
    loadNode: (nodeIndex: number) => Promise<Object3D>;
    loadScene: () => Promise<Group>;
}

interface GLTFLoaderPlugin {
    beforeRoot?: (() => Promise<void> | null) | undefined;
    afterRoot?: ((result: GLTF) => Promise<void> | null) | undefined;
    loadMesh?: ((meshIndex: number) => Promise<Group | Mesh | SkinnedMesh> | null) | undefined;
    loadBufferView?: ((bufferViewIndex: number) => Promise<ArrayBuffer> | null) | undefined;
    loadMaterial?: ((materialIndex: number) => Promise<Material> | null) | undefined;
    loadTexture?: ((textureIndex: number) => Promise<Texture> | null) | undefined;
    getMaterialType?: ((materialIndex: number) => typeof Material | null) | undefined;
    extendMaterialParams?:
        | ((materialIndex: number, materialParams: { [key: string]: any }) => Promise<any> | null)
        | undefined;
    createNodeMesh?: ((nodeIndex: number) => Promise<Group | Mesh | SkinnedMesh> | null) | undefined;
    createNodeAttachment?: ((nodeIndex: number) => Promise<Object3D> | null) | undefined;
}

// Type definitions for dat.GUI 0.7


declare class GUIController {
    domElement: HTMLElement;
    object: Object;
    property: string;

    constructor(object: Object, property: string);

    options(option: any): GUIController;
    name(name: string): GUIController;

    listen(): GUIController;
    remove(): GUIController;

    onChange(fnc: (value?: any) => void): GUIController;
    onFinishChange(fnc: (value?: any) => void): GUIController;

    setValue(value: any): GUIController;
    getValue(): any;

    updateDisplay(): GUIController;
    isModified(): boolean;

    // NumberController
    min(n: number): GUIController;
    max(n: number): GUIController;
    step(n: number): GUIController;

    // FunctionController
    fire(): GUIController;
}

declare class Font {
    constructor(jsondata: any);

    /**
     * @default 'Font'
     */
    type: string;

    data: string;

    generateShapes(text: string, size: number): Shape[];
}

interface TextGeometryParameters {
    font: Font;
    size?: number | undefined;
    height?: number | undefined;
    curveSegments?: number | undefined;
    bevelEnabled?: boolean | undefined;
    bevelThickness?: number | undefined;
    bevelSize?: number | undefined;
    bevelOffset?: number | undefined;
    bevelSegments?: number | undefined;
}

declare class TextGeometry extends ExtrudeGeometry {
    /**
     * @default 'TextGeometry'
     */
    type: string;

    constructor(text: string, parameters: TextGeometryParameters);

    parameters: {
        font: Font;
        size: number;
        height: number;
        curveSegments: number;
        bevelEnabled: boolean;
        bevelThickness: number;
        bevelSize: number;
        bevelOffset: number;
        bevelSegments: number;
    };
}

declare type T = typeof THREE;
declare type K = keyof WebGLRenderer;
interface AnimateOptions {
    camera: PerspectiveCamera;
    elapsedTime: number;
    timestamp: number;
}
interface MiddlewareOptions {
    OrbitControls: OrbitControls;
    camera: PerspectiveCamera;
}
interface FnNameMap {
    cc: 'CubeCamera';
    v3: 'Vector3';
    v2: 'Vector2';
    v4: 'Vector4';
    oc: 'OrthographicCamera';
    pc: 'PerspectiveCamera';
    sc: 'StereoCamera';
    bg: 'BoxGeometry';
    ba: 'BufferAttribute';
    bufferg: 'BufferGeometry';
    capsuleg: 'CapsuleGeometry';
    coneg: 'ConeGeometry';
    cylinderg: 'CylinderGeometry';
    dg: 'DodecahedronGeometry';
    edgeg: 'EdgesGeometry';
    extrudeg: 'ExtrudeGeometry';
    ig: 'IcosahedronGeometry';
    lg: 'LatheGeometry';
    og: 'OctahedronGeometry';
    pg: 'PlaneGeometry';
    polyg: 'PolyhedronGeometry';
    rg: 'RingGeometry';
    sg: 'SphereGeometry';
    tetrag: 'TetrahedronGeometry';
    tg: 'TorusGeometry';
    tkg: 'TorusKnotGeometry';
    tubeg: 'TubeGeometry';
    wfg: 'WireframeGeometry';
    cg: 'CircleGeometry';
    ac: 'ArcCurve';
    crc3: 'CatmullRomCurve3';
    cbc: 'CubicBezierCurve';
    cbc3: 'CubicBezierCurve3';
    ec: 'EllipseCurve';
    lc: 'LineCurve';
    lc3: 'LineCurve3';
    qbc: 'QuadraticBezierCurve';
    qbc3: 'QuadraticBezierCurve3';
    splinec: 'SplineCurve';
    arrowh: 'ArrowHelper';
    ah: 'AxesHelper';
    bh: 'BoxHelper';
    b3h: 'Box3Helper';
    ch: 'CameraHelper';
    dlh: 'DirectionalLightHelper';
    gh: 'GridHelper';
    pgh: 'PolarGridHelper';
    hlh: 'HemisphereLightHelper';
    ph: 'PlaneHelper';
    plh: 'PointLightHelper';
    sh: 'SkeletonHelper';
    slh: 'SpotLightHelper';
    animationl: 'AnimationLoader';
    audiol: 'AudioLoader';
    bgl: 'BufferGeometryLoader';
    cache: 'Cache';
    c: 'Color';
    compressedtl: 'CompressedTextureLoader';
    ctl: 'CubeTextureLoader';
    dtl: 'DataTextureLoader';
    filel: 'FileLoader';
    fl: 'FontLoader';
    svgl: 'SVGLoader';
    il: 'ImageLoader';
    ibl: 'ImageBitmapLoader';
    l: 'Loader';
    lu: 'LoaderUtils';
    ml: 'MaterialLoader';
    ol: 'ObjectLoader';
    tl: 'TextureLoader';
    g: 'Group';
    lbm: 'LineBasicMaterial';
    ldm: 'LineDashedMaterial';
    m: 'Mesh';
    mbm: 'MeshBasicMaterial';
    mdepthm: 'MeshDepthMaterial';
    mdistancem: 'MeshDistanceMaterial';
    mlm: 'MeshLambertMaterial';
    mmm: 'MeshMatcapMaterial';
    mnm: 'MeshNormalMaterial';
    mpm: 'MeshPhongMaterial';
    mphysicalm: 'MeshPhysicalMaterial';
    msm: 'MeshStandardMaterial';
    mtm: 'MeshToonMaterial';
    p: 'Points';
    pm: 'PointsMaterial';
    rsm: 'RawShaderMaterial';
    sm: 'ShaderMaterial';
    shadowm: 'ShadowMaterial';
    spritem: 'SpriteMaterial';
    line: 'Line';
    lp: 'LineLoop';
    ls: 'LineSegments';
    al: 'AmbientLight';
    alp: 'AmbientLightProbe';
    dl: 'DirectionalLight';
    hl: 'HemisphereLight';
    hlp: 'HemisphereLightProbe';
    pl: 'PointLight';
    ral: 'RectAreaLight';
    sl: 'SpotLight';
    pls: 'PointLightShadow';
    dls: 'DirectionalLightShadow';
    sls: 'SpotLightShadow';
    lph: 'LightProbeHelper';
    ralh: 'RectAreaLightHelper';
    f: 'Fog';
    aa: 'AnimationAction';
    anc: 'AnimationClip';
    am: 'AnimationMixer';
    aog: 'AnimationObjectGroup';
    au: 'AnimationUtils';
    a: 'Animation';
    anl: 'AnimationLoader';
}
declare type ShadowType = 'BasicShadowMap' | 'PCFShadowMap' | 'PCFSoftShadowMap' | 'VSMShadowMap';
interface SThreeOptions extends Record<string, any> {
    createMesh: () => void;
    createCamera: () => PerspectiveCamera;
    animate?: (animationOptions: AnimateOptions) => void | PerspectiveCamera;
    middleware?: (middlewareOptions: MiddlewareOptions) => any;
    mousemove?: (e: Event) => void;
    mousedown?: (e: Event) => void;
    mouseup?: (e: Event) => void;
    debug?: boolean;
    alias?: Record<string, string>;
    shadowType?: ShadowType;
}
interface Scene extends Object3D {
    _add?: (...args: any[]) => void;
}
interface ReturnType {
    c: (fnName: keyof FnNameMap | keyof T, ...args: any[]) => any;
    cf: (url: string, text: string, options: TextGeometryParameters) => Promise<TextGeometry>;
    track: (...args: [target: Object, propName: string, min?: number, max?: number, step?: number]) => GUIController;
    setUV: (target: Mesh, size?: number) => void;
    glTFLoader: (url: string, dracoLoader?: DRACOLoader, callback?: (gltf: GLTFLoader) => void) => Promise<GLTFLoader>;
    draCOLoader: (decoderPath: string) => DRACOLoader;
    animationArray: Mesh[];
    THREE: T;
    scene: Scene;
    renderer: WebGLRenderer;
    dom: HTMLCanvasElement;
    setRendererAttributes: (options: Record<K, any>) => void;
}
declare function sThree(container: HTMLElement | string, options: SThreeOptions): ReturnType;

export { Props, sThree };
